//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   3 Nov 06  Brian Frank  Creation
//

**
** BuildPod is the base class for build scripts used to manage
** building a Fan source code and resources into a Fan pod.
**
abstract class BuildPod : BuildScript
{

//////////////////////////////////////////////////////////////////////////
// Pod Meta-Data
//////////////////////////////////////////////////////////////////////////

  **
  ** Programatic name of the pod.  Required.
  **
  Str podName := null

  **
  ** Short one sentence description of the pod.  Optional.
  **
  Str description := null

  **
  ** Version of the pod - see BuildScript.globalVersion.  Required.
  **
  Version version := null

  **
  ** Dependencies of the pod formatted as a list of sys::Depends.  Required.
  **
  Str[] depends := null

  **
  ** List of Uris relative to scriptDir of directories containing
  ** the Fan source files to compile.  Required.
  **
  Uri[] srcDirs := null

  **
  ** List of Uris relative to scriptDir of directories of resources
  ** files to package into pod zip file.  Optional.
  **
  Uri[] resDirs := null

  **
  ** List of Uris relative to scriptDir of directories containing
  ** the Java source files to compile for Java native jar.
  **
  Uri[] javaDirs := null

  **
  ** List of Uris relative to scriptDir of Java jar files which
  ** are automatically included in the classpath when compiling the
  ** javaDirs.
  **
  Uri[] javaLibs := null

//////////////////////////////////////////////////////////////////////////
// Setup
//////////////////////////////////////////////////////////////////////////

  **
  ** Validate subclass constructor setup required meta-data.
  **
  internal override Void validate()
  {
    ok := true
    ok &= validateReqField("podName")
    ok &= validateReqField("version")
    ok &= validateReqField("depends")
    if (!ok) throw FatalBuildErr.make
  }

//////////////////////////////////////////////////////////////////////////
// Targets
//////////////////////////////////////////////////////////////////////////

  **
  ** BuildPod publishes a default set of targets used
  ** for managing building pods.
  **
  override virtual Target[] targets()
  {
    mine :=
    [
      Target.make(this, "compile", "compile fan source into pod") |,| { compile },
      Target.make(this, "clean",   "delete all intermediate and target files") |,| { clean },
      Target.make(this, "native",  "build native Java jar file") |,| { natives },
      Target.make(this, "full",    "clean+compile+native (with doc+src)") |,| { full },
      Target.make(this, "test",    "run fant for specified pod") |,| { test },
    ]
    return mine.addAll(super.targets)
  }

//////////////////////////////////////////////////////////////////////////
// Compile
//////////////////////////////////////////////////////////////////////////

  Void compile(Bool full := false)
  {
    log.message("compile", podName)
    log.indent
    fanc := CompileFan.make(this)
      fanc.includeDoc = full
      fanc.includeSrc = full
    fanc.run
    log.unindent
  }

  internal Depend[] parseDepends()
  {
    return (Depend[])depends.map(Depend[,]) |Str s->Depend| { return Depend.parse(s) }
  }

//////////////////////////////////////////////////////////////////////////
// Clean
//////////////////////////////////////////////////////////////////////////

  Void clean()
  {
    log.message("clean", podName)
    log.indent
    Delete.make(this, libFanDir+"${podName}.pod".toUri).run
    Delete.make(this, libJavaDir+"${podName}.jar".toUri).run
    Delete.make(this, scriptDir+"temp-java/".toUri).run
    log.unindent
  }

//////////////////////////////////////////////////////////////////////////
// Native
//////////////////////////////////////////////////////////////////////////

  Void natives()
  {
    if (javaDirs == null) return

    log.message("native", podName)
    log.indent

    // env
    jtemp    := scriptDir + `temp-java/`
    jstub    := jtemp + "${podName}.jar".toUri
    jdk      := JdkTask.make(this)
    javaExe  := jdk.javaExe
    jarExe   := jdk.jarExe
    curPod   := libFanDir + "${podName}.pod".toUri
    curJar   := libJavaDir + "${podName}.jar".toUri
    javaDirs := resolveDirs(javaDirs)
    javaLibs := resolveFiles(javaLibs)
    depends  := parseDepends

    // start with a clean directory
    Delete.make(this, jtemp).run
    CreateDir.make(this, jtemp).run

    // stub the pods fan classes into Java classfiles
    // by calling the JStub tool in the jsys runtime
    Exec.make(this, [javaExe.osPath,
                     "-cp", "${libJavaDir}sys.jar",
                     "fanx.tools.Jstub",
                     "-d", jtemp.osPath,
                     podName]).run

    // compile
    javac := CompileJava.make(this)
    javac.outDir = jtemp
    javac.cp.add(jtemp+"${podName}.jar".toUri).addAll(javaLibs)
    depends.each |Depend d| { javac.cp.add(libJavaDir+(d.name+".jar").toUri) }
    javac.src = javaDirs
    javac.run

    // extract stub jar into the temp directory
    Exec.make(this, [jarExe.osPath, "-xf", jstub.osPath], jtemp).run

    // now we can nuke the stub jar (and manifest)
    Delete.make(this, jstub).run
    Delete.make(this, jtemp + `meta-inf/`).run

    // jar everything back up to lib/java/{pod}.jar
    Exec.make(this, [jarExe.osPath, "cf", curJar.osPath, "-C", jtemp.osPath, "."], jtemp).run

    // append files to the pod zip (we use java's jar tool)
    Exec.make(this, [jarExe.osPath, "-fu", curPod.osPath, "-C", jtemp.osPath, "."], jtemp).run

    // cleanup temp
    Delete.make(this, jtemp).run

    log.unindent
  }

//////////////////////////////////////////////////////////////////////////
// Full
//////////////////////////////////////////////////////////////////////////

  Void full()
  {
    clean
    compile(true)
    natives
  }

//////////////////////////////////////////////////////////////////////////
// Test
//////////////////////////////////////////////////////////////////////////

  Void test()
  {
    log.message("test", podName)
    log.indent

    fant := binDir + `fant.exe`
    Exec.make(this, [fant.osPath, podName]).run

    log.unindent
  }

}