//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   3 Nov 06  Brian Frank  Creation
//

using compiler

**
** BuildScript is the base class for build scripts - it manages
** the command line interface, argument parsing, environment, and
** target execution.
**
abstract class BuildScript
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  **
  ** Construct a new build script.
  **
  new make()
  {
    log = BuildLog.make
    initEnv
    try
    {
      setup
      validate
    }
    catch (Err err)
    {
      log.error("Error initializing script", scriptFile.osPath)
      throw err
    }
  }

//////////////////////////////////////////////////////////////////////////
// Env
//////////////////////////////////////////////////////////////////////////

  ** The source file of this script
  File scriptFile

  ** The directory containing the this script
  File scriptDir

  ** Home directory of development installation.  By default this
  ** value is initialized by Sys.env["fan.build.devHome"], otherwise
  ** Sys.homeDir is used.
  File devHomeDir

  ** {devHomeDir}/bin/{os}/
  File binDir

  ** {devHomeDir}/lib/
  File libDir

  ** {devHomeDir}/lib/fan
  File libFanDir

  ** {devHomeDir}/lib/java
  File libJavaDir

  ** {devHomeDir}/lib/net
  File libNetDir

  ** This is the global default version to use when building pods.  It
  ** is initialized by Sys.env["fan.build.globalVersion"], otherwise
  ** "0.0.0" is used as a default.
  Version globalVersion

  **
  ** Initialize the environment
  **
  private Void initEnv()
  {
    // init devHomeDir
    devHomeDir = Sys.homeDir
    devHomeProp := Sys.env["fan.build.devHome"]
    if (devHomeProp != null)
    {
      try
      {
        f := File.make(devHomeProp.toUri)
        if (!f.exists || !f.isDir) throw Err.make
        devHomeDir = f
      }
      catch
      {
        log.error("Invalid URI for fan.build.devHome: $devHomeProp")
      }
    }

    // global version
    globalVersion = Version.parse("0.0.0")
    globalVersionProp := Sys.env["fan.build.globalVersion"]
    if (devHomeProp != null)
    {
      try
      {
        globalVersion = Version.parse(globalVersionProp)
      }
      catch
      {
        log.error("Invalid Version for fan.build.globalVersion: $globalVersionProp")
      }
    }

    // directories
    scriptFile = File.make(type.sourceFile.toUri)
    scriptDir  = scriptFile.parent
    binDir     = devHomeDir + `bin/win/`
    libDir     = devHomeDir + `lib/`
    libFanDir  = devHomeDir + `lib/fan/`
    libJavaDir = devHomeDir + `lib/java/`
    libNetDir  = devHomeDir + `lib/net/`

    // debug
    if (log.isVerbose)
    {
      log.printLine("BuildScript Environment:")
      log.printLine("  scriptFile:    $scriptFile")
      log.printLine("  scriptDir:     $scriptDir")
      log.printLine("  devHomeDir:    $devHomeDir")
      log.printLine("  binDir:        $binDir")
      log.printLine("  libDir:        $libDir")
      log.printLine("  libFanDir:     $libFanDir")
      log.printLine("  libJavaDir:    $libJavaDir")
      log.printLine("  libNetDir:     $libNetDir")
      log.printLine("  globalVersion: $globalVersion")
    }
  }

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

 **
 ** Return this script's source file path.
 **
 override Str toStr()
 {
   return type.sourceFile
 }

  **
  ** Get the available targets this script publishes.  Scripts should
  ** override this method to prepend their own targets to what the
  ** super class publishes.  The first target in this list is the default
  ** target of the script.
  **
  virtual Target[] targets()
  {
    return Target[,]
  }

//////////////////////////////////////////////////////////////////////////
// Arguments
//////////////////////////////////////////////////////////////////////////

  **
  ** Parse the arguments passed from the command line.
  ** Return true for success or false to end the script.
  **
  private Bool parseArgs(Str[] args)
  {
    // check for usage
    if (args.contains("-?") || args.contains("-help"))
    {
      usage
      return false
    }

    success := true
    toRun = Target[,]

    // get published targetss
    published := targets
    if (published.isEmpty)
    {
      log.error("No targets available for script")
      return false
    }

    // process each argument
    for (i:=0; i<args.size; ++i)
    {
      arg := args[i]
      if (arg == "-v") log.level = LogLevel.verbose
      else if (arg.startsWith("-")) log.warning("Unknown option $arg")
      else
      {
        // add target to our run list
        target := published.find |Target t->Bool| { return t.name == arg }
        if (target == null)
        {
          log.error("Unknown target '$arg'")
          success = false
        }
        else
        {
          toRun.add(target)
        }
      }
    }

    // if no targets specified, then use the default
    if (toRun.isEmpty) toRun.add(published.first)

    // return success flag
    return success
  }

  **
  ** Dump usage including all this scripts published targets.
  **
  private Void usage()
  {
    log.printLine("usage: ")
    log.printLine("  build [options] <target>*")
    log.printLine("options:")
    log.printLine("  -? -help    print usage summary")
    log.printLine("  -v          verbose logging")
    log.printLine("targets:")
    targets.each |Target t, Int i|
    {
      log.print("  ${t.name.justl(10)}  $t.description")
      if (i == 0) log.print(" (default)")
      log.printLine
    }
  }

//////////////////////////////////////////////////////////////////////////
// Setup
//////////////////////////////////////////////////////////////////////////

  **
  ** The setup callback is invoked before processing of any targets
  ** to ensure that the BuildScript is correctly initialized.  If the
  ** script cannot be setup then report errors via the log and
  ** throw FatalBuildErr to terminate the script.
  **
  virtual Void setup()
  {
  }

  **
  ** Internal callback to validate setup
  **
  internal virtual Void validate()
  {
  }

  **
  ** Check that the specified field is non-null, if not
  ** then log an error and return false.
  **
  internal Bool validateReqField(Str field)
  {
    val := type.field(field).get(this)
    if (val != null) return true
    log.error("Required field not set: '$field'", toStr)
    return false
  }

  **
  ** Convert a list of Uris to directories and verify they all exist.
  **
  internal File[] resolveDirs(Uri[] uris)
  {
    return resolveUris(uris, true)
  }

  **
  ** Convert a list of Uris to files and verify they all exist.
  **
  internal File[] resolveFiles(Uri[] uris)
  {
    return resolveUris(uris, false)
  }

  private File[] resolveUris(Uri[] uris, Bool expectDir)
  {
    files := File[,]
    if (uris == null) return files

    files.capacity = uris.size
    ok := true
    uris.each |Uri uri|
    {
      file := scriptDir + uri
      if (!file.exists || file.isDir != expectDir )
      {
        ok = false
        if (expectDir)
          log.error("Invalid directory", uri.toStr)
        else
          log.error("Invalid file", uri.toStr)
      }
      files.add(file)
    }
    if (!ok) throw FatalBuildErr.make
    return files
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  **
  ** Log an error and return a FatalBuildErr instance
  **
  FatalBuildErr fatal(Str msg, Str aux := null, Err err := null)
  {
    log.error(msg, aux, err)
    return FatalBuildErr.make
  }

//////////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////////

  **
  ** Run the script with the specified arguments.
  ** Return 0 on success or -1 on failure.
  **
  Int main(Str[] args := Sys.args)
  {
    t1 := Duration.now
    success := false
    try
    {
      if (!parseArgs(args)) return -1
      toRun.each |Target t| { t.run }
      success = true
    }
    catch (FatalBuildErr err)
    {
      // error should have alredy been logged
    }
    catch (Err err)
    {
      log.error("Internal build error", toStr)
      err.trace
    }
    t2 := Duration.now

    if (success)
      echo("BUILD SUCCESS [${(t2-t1).toMillis}ms]!")
    else
      echo("BUILD FAILED [${(t2-t1).toMillis}ms]!")
    return success ? 0 : -1
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  ** Log used for error reporting and tracing
  BuildLog log

  ** Targets specified to run by command line
  Target[] toRun

}