//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   5 May 07  Brian Frank  Creation
//

using compiler
using fandoc

**
** ApiToHtmlGenerator generates an HTML file for a Type's API
**
class ApiToHtmlGenerator : HtmlGenerator
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  new make(DocCompiler compiler, Location loc, OutStream out, Type t)
    : super(compiler, loc, out)
  {
    this.t = t
  }

//////////////////////////////////////////////////////////////////////////
// Generator
//////////////////////////////////////////////////////////////////////////

  **
  ** Generate the HTML documentation for this Type.
  **
  override Void generate()
  {
    sorter := |Slot a, Slot b -> Int| { return a.name <=> b.name }
    filter := |Slot s -> Bool|
    {
      v := !s.isSynthetic
      v &= t == Obj.type || s.parent != Obj.type
      v &= t == s.parent
      return v
    }
    this.slots = t.slots.rw.sort(sorter).findAll(filter)

    out.print("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\"\n")
    out.print(" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n")
    out.print("<html>\n")
    out.print("<head>\n")
    out.print("  <title>").print(t.qname).print("</title>\n")
    out.print("  <meta http-equiv='Content-Type' content='text/html; charset=UTF-8'/>\n")
    out.print("  <link rel='stylesheet' type='text/css' href='../fandoc.css'/>\n")
    out.print("</head>\n")
    out.print("<body>\n")

    out.print("<div id='content'>\n")
    generateContent
    out.print("</div>\n")

    out.print("<div id='sidebarBg'></div>\n")
    out.print("<div id='sidebar'>\n")
    generateSidebar
    out.print("</div>\n")

    out.print("</body>\n")
    out.print("</html>\n")
    out.close
  }

//////////////////////////////////////////////////////////////////////////
// Overrides
//////////////////////////////////////////////////////////////////////////

  **
  ** Generate the main content.
  **
  virtual Void generateContent()
  {
    nav
    typeOverview
    typeDetail
    slotsDetail
  }

  **
  ** Generate the sidebar.
  **
  virtual Void generateSidebar()
  {
    actions
    slotsOverview
  }

//////////////////////////////////////////////////////////////////////////
// Methods
//////////////////////////////////////////////////////////////////////////

  **
  ** Generate the page navigation.
  **
  Void nav()
  {
    out.print("<div id='nav'>\n")
    out.print("<ul>\n")
    out.print("  <li><a href='..'>All Pods</a></li>\n")
    out.print("  <li><a href='.'>$t.pod.name Pod</a></li>\n")
    out.print("</ul>\n")
    out.print("</div>\n")
  }

  **
  ** Generate the actions.
  **
  Void actions()
  {
    out.print("<div id='actions'>\n")
    out.print("<h2>Actions</h2>\n")
    out.print("<ul>\n")
    out.print("  <li><a href='${t.name}_src.html'>View Source</a></li>\n")
    out.print("  <li><a href='#'>Show All Slots</a></li>\n")
    out.print("</ul>\n")
    out.print("</div>\n")
  }

  **
  ** Generate the type overview documentation.
  **
  Void typeOverview()
  {
    out.print("<div id='typeOverview'>\n")
    out.print("<h1>$t.qname</h1>\n")
    if (t.base != null) inheritance
    out.print("</div>\n")
  }

  **
  ** Generate the type detail documentation.
  **
  Void typeDetail()
  {
    if (t.doc == null) return
    out.print("<div id='typeDetail'>\n")
    fandoc(t.doc)
    out.print("</div>\n")
  }

  **
  ** Generate the type inheritance.
  **
  Void inheritance()
  {
    chain := Type[t]
    base  := t.base
    while (base != null)
    {
      chain.insert(0, base)
      base = base.base
    }
    out.print("<pre><code>")
    chain.each |Type t, Int i|
    {
      if (i > 0) out.print("\n${Str.spaces(i*2)}")
      if (i == chain.size-1) out.print(t.qname)
      else out.print("<a href='${typeUri(t)}'>$t.qname</a>")
    }
    out.print("</code></pre>")
  }

  **
  ** Generate the slot overview documentation.
  **
  Void slotsOverview()
  {
    out.print("<div id='slotsOverview'>\n")
    out.print("<h2>Slots</h2>\n")
    out.print("<ul>\n")
    slots.each |Slot slot|
    {
      out.print("  <li><a href='#$slot.name'>$slot.name</a></li>\n")
    }
    out.print("</ul>\n")
    out.print("</div>\n")
  }

  **
  ** Generate the slot detail documentation.
  **
  Void slotsDetail()
  {
    out.print("<div id='slotsDetail'>\n")
    out.print("<h2>Slots</h2>\n")
    out.print("<dl>\n")
    slots.each |Slot slot| { slotDetail(slot) }
    out.print("</dl>\n")
    out.print("</div>\n")
  }

  **
  ** Generate the documentation for the given slot.
  **
  Void slotDetail(Slot slot)
  {
    cls := (slot.isField) ? "field" : "method"
    out.print("<dt id='$slot.name' class='$cls'>$slot.name</dt>\n")
    out.print("<dd>\n")

    // Slot spec
    out.print("<p><code>")
    if (slot.isField)
    {
      f := (Field)slot
      slotScope(f)
      typeLink(f.fieldType)
      out.print(" $f.name")
    }
    else
    {
      m := (Method)slot
      if (m.isCtor) out.print("new")
      else
      {
        slotScope(m)
        typeLink(m.returnType)
      }
      out.print(" $m.name")
      out.print("(")
      m.params.each |MethodParam p, Int i|
      {
        if (i > 0) out.print(", ")
        typeLink(p.paramType)
        out.print(" $p.name")
        if (p.hasDefault) out.print(" := def")
      }
      out.print(")")
    }
    out.print("</code></p>\n")

    // Slot comment
    if (slot.doc != null)
      fandoc(slot.doc)

    out.print("</dd>\n")
  }

  **
  ** Write a slot's scope.
  **
  Void slotScope(Slot s)
  {
    if (s.isStatic)    out.print("static ")
    if (s.isProtected) out.print("protected ")
    if (s.isPrivate)   out.print("private ")
    if (s.isInternal)  out.print("internal ")
  }

  **
  ** Return the fandoc Uri for this type.
  **
  Str typeUri(Type t)
  {
    return "../$t.pod.name/${t.name}.html"
  }

  **
  ** Write a type link out in the form <a href='pod/type'>type.name</a>.
  **
  Void typeLink(Type t)
  {
    sig := t.signature

    // Check for method sig
    if (sig.index("|") != null)
    {

      // TODO - need to handle |sys::Int->sys::Void|
      /*
      i := sig.index("::")
      while (i != null)
      {
        j := sig.indexr(">", i)
        if (j == null) j = sig.indexr(",", i)
        if (j == null) j = sig.indexr("|", i)

        sig = sig[0..j] + sig[i+2..-1]
        j = sig.index("::")
      }
      */
    }
    else
    {
      // Simply trim down to just type name
      sig = sig[sig.index("::")+2..-1]

      // Check if need to clean up more for map types
      i := sig.index("::")
      if (i != null)
        sig = sig[0..sig.index(":")] + sig[i+2..-2]
    }
    out.print("<a href='${typeUri(t)}'>$sig</a>")
  }

  **
  ** Write out the fandoc for this text - if an exception
  ** is thrown, write the original text.
  **
  Void fandoc(Str text)
  {
    try
    {
      doc := FandocParser.make.parse(null, InStream.makeForStr(text))
      doc.children.each |DocNode child| { child.write(this) }
    }
    catch { out.print("<p>$text<p>\n") }
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  Type t        // type to documenting
  Slot[] slots  // slots to document

}
