//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   9 May 07  Andy Frank  Creation
//

using compiler
using fandoc

**
** SourceToHtmlGenerator generates an syntax color coded HTML
** file for a Type's source code.
**
class SourceToHtmlGenerator : ApiToHtmlGenerator
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  new make(DocCompiler compiler, Location loc, OutStream out, Type t)
    : super(compiler, loc, out, t)
  {
  }

//////////////////////////////////////////////////////////////////////////
// ApiToHtmlGenerator
//////////////////////////////////////////////////////////////////////////

  **
  ** Generate the main content.
  **
  override Void generateContent()
  {
    nav
    typeOverview
    out.print("<div id='source'>\n")
    file := t.pod.files["/src/$t.sourceFile".toUri]
    convert(file.in.readAllLines, out)
    out.print("</div>\n")
  }

  **
  ** Generate the sidebar.
  **
  override Void generateSidebar()
  {
    out.print("<div id='actions'>\n")
    out.print("<h2>Actions</h2>\n")
    out.print("<ul>\n")
    out.print("  <li><a href='${t.name}.html'>View Fandoc</a></li>\n")
    out.print("</ul>\n")
    out.print("</div>\n")
    slotsOverview
  }

//////////////////////////////////////////////////////////////////////////
// FanToHtml
//////////////////////////////////////////////////////////////////////////

  **
  ** Convert the Fan source code to syntax highlighted HTML markup.
  **
  static Void convert(Str[] lines, OutStream out)
  {
    max := lines.size.toStr.size
    inBlockComment := false

    lines.each |Str line, Int i|
    {
      num := i+1
      pad := max - num.toStr.size + 1

      out.print("<span class='lineNum'>")
      safeStr(out, Str.spaces(pad))
      out.print("${num} </span>")

      out.print("&nbsp;")
      inBlockComment = markup(line, out, inBlockComment)

      out.print("<br/>\n")
    }
  }

//////////////////////////////////////////////////////////////////////////
// Support Methods
//////////////////////////////////////////////////////////////////////////

  **
  ** Markup the given line with syntax highlighting.
  **
  private static Bool markup(Str line, OutStream out, Bool inBlockComment := false)
  {
    inStr := false
    buf   := Buf.make

    if (inBlockComment) out.print("<span class='blockComment'")

    for (i := 0; i<line.size; i++)
    {
      curr := line[i]
      peek := (i < line.size-1) ? line[i+1] : -1
      pre  := (i > 0) ? line[i-1] : -1

      // Currenty in a block comment, keep looping till we find the end
      if (inBlockComment)
      {
        safe(out, curr)
        if (curr == '*' && peek == '/')
        {
          inBlockComment = false
          safe(out, peek)
          out.print("</span>")
          i++
        }
        continue
      }

      // Currently in a Str or Uri, keep looping till we find the end
      if (inStr)
      {
        safe(out, curr)
        if ((curr == '"' && pre != '\\') || curr == '`')
        {
          inStr = false
          out.print("</span>")
        }
        continue
      }

      // We're starting a Str
      if (curr == '"')
      {
        inStr = true
        out.print("<span class='string'>").writeChar(curr)
        continue
      }

      // We're starting a Uri
      if (curr == '`')
      {
        inStr = true
        out.print("<span class='uri'>").writeChar(curr)
        continue
      }

      // We're starting a Char
      if (curr == '\'')
      {
        out.print("<span class='char'>").writeChar(curr)
        safe(out, peek)
        if (peek == '\\')
        {
          safe(out, line[i+2])
          i++
        }
        i += 2
        out.writeChar('\'')
        out.print("</span>")
        continue
      }

      // Block comment
      if (curr == '/' && peek == '*')
      {
        out.print("<span class='blockComment'>")
        inBlockComment = true
        out.writeChar(curr).writeChar(peek)
        i++
        continue
      }

      // Line comment
      if (curr == '/' && peek == '/')
      {
        out.print("<span class='lineComment'>")
        safeStr(out, line[i..-1])
        out.print("</span>")
        return false
      }

      // Fandoc comment
      if (curr == '*' && peek == '*')
      {
        out.print("<span class='fandocComment'>")
        safeStr(out, line[i..-1])
        out.print("</span>")
        return false
      }

      if (isAlpha(curr))
      {
        // Build up buffer to check if this is a keyword
        buf.writeChar(curr)
        continue
      }
      else if (!buf.empty)
      {
        // Word is over, write out checking for keywords
        doBuf(buf, out)

        // Reset buf for next time
        buf.clear
      }

      // Single character
      if (curr == '(' || curr == ')' ||
          curr == '{' || curr == '}' ||
          curr == '[' || curr == ']')
      {
        out.print("<span class='bracket'>").writeChar(curr)
        out.print("</span>")
      }
      else
      {
        safe(out, curr)
      }
    }

    // If buf ended the line, make sure we still check for a keyword
    if (!buf.empty) doBuf(buf, out)

    // If still in block comment, close tag for this line
    if (inBlockComment) out.print("</span>")
    return inBlockComment
  }

  **
  ** Convenience for s.each |Int ch| { safe(out, ch) }
  **
  private static Void safeStr(OutStream out, Str s)
  {
    s.each |Int ch| { safe(out, ch) }
  }

  **
  ** Escape <, &, and > characters.
  **
  private static Void safe(OutStream out, Int ch)
  {
    if (ch == ' ') out.print("&nbsp;")
    else if (ch == '<') out.print("&lt;")
    else if (ch == '>') out.print("&gt;")
    else if (ch == '&') out.print("&amp;")
    else out.writeChar(ch)
  }

  **
  ** Handle writing the buf, checking for keywords. This method
  ** assumes the buf is still in write mode.
  **
  private static Void doBuf(Buf buf, OutStream out)
  {
    buf.flip
    s := buf.readAllStr

    if (isKeyword(s))
    {
      // Keyword
      out.print("<span class='keyword'>$s</span>")
    }
    else
    {
      // Indentifier
      out.print(s)
    }
  }

  **
  ** Return true if the buf contains only alphabetic characters,
  ** numerals or underscores.
  **
  private static Bool isAlpha(Int ch)
  {
    if (ch >= 65 && ch <= 90) return true   // upper alpha
    if (ch >= 97 && ch <= 122) return true  // lower alpha
    if (ch >= 48 && ch <= 57) return true   // numeral
    if (ch == 95) return true               // underscore
    return false
  }

  **
  ** Return true if this Str matches a keyword
  **
  private static Bool isKeyword(Str s)
  {
    switch (s[0])
    {
      case 'a':
        if (s == "abstract") return true
        if (s == "as") return true
        if (s == "assert") return true
      case 'b':
        if (s == "boolean") return true
        if (s == "break") return true
      case 'c':
        if (s == "case") return true
        if (s == "catch") return true
        if (s == "class") return true
        if (s == "const") return true
        if (s == "continue") return true
      case 'd':
        if (s == "default") return true
        if (s == "delegate") return true
        if (s == "do") return true
      case 'e':
        if (s == "else") return true
        if (s == "enum") return true
        if (s == "event") return true
        if (s == "explicit") return true
        if (s == "extends") return true
        if (s == "extern") return true
      case 'f':
        if (s == "false") return true
        if (s == "final") return true
        if (s == "finally") return true
        if (s == "fixed") return true
        if (s == "for") return true
        if (s == "foreach") return true
      case 'g':
        if (s == "goto") return true
      case 'i':
        if (s == "if") return true
        if (s == "implements") return true
        if (s == "implicit") return true
        if (s == "import") return true
        if (s == "instanceof") return true
        if (s == "interface") return true
        if (s == "internal") return true
        if (s == "is") return true
      case 'l':
        if (s == "lock") return true
      case 'm':
        if (s == "mixin") return true
      case 'n':
        if (s == "native") return true
        if (s == "new") return true
        if (s == "null") return true
      case 'o':
        if (s == "object") return true
        if (s == "operator") return true
        if (s == "override") return true
      case 'p':
        if (s == "package") return true
        if (s == "private") return true
        if (s == "protected") return true
        if (s == "public") return true
      case 'r':
        if (s == "readonly") return true
        if (s == "return") return true
      case 's':
        if (s == "sealed") return true
        if (s == "sizeof") return true
        if (s == "static") return true
        if (s == "strictfp") return true
        if (s == "struct") return true
        if (s == "super") return true
        if (s == "switch") return true
        if (s == "synchronized") return true
      case 't':
        if (s == "this") return true
        if (s == "throw") return true
        if (s == "throws") return true
        if (s == "transient") return true
        if (s == "true") return true
        if (s == "try") return true
        if (s == "typeof") return true
      case 'u':
        if (s == "unchecked") return true
        if (s == "unsafe") return true
        if (s == "using") return true
      case 'v':
        if (s == "virtual") return true
        if (s == "void") return true
        if (s == "volatile") return true
        if (s == "while") return true
    }
    return false
  }

}
