//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   9 Feb 07  Brian Frank  Creation
//

class TcpSocketTest : Test
{

//////////////////////////////////////////////////////////////////////////
// Make
//////////////////////////////////////////////////////////////////////////

  Void testMake()
  {
    s := TcpSocket.make
    verifyEq(s.isBound, false)
    verifyEq(s.isConnected, false)
    verifyEq(s.isClosed, false)
    verifyEq(s.localAddress, null)
    verifyEq(s.localPort, null)
    verifyEq(s.remoteAddress, null)
    verifyEq(s.remotePort, null)
    verifyErr(IOErr.type) |,| { s.in }
    verifyErr(IOErr.type) |,| { s.out }
    s.close
  }

//////////////////////////////////////////////////////////////////////////
// Bind
//////////////////////////////////////////////////////////////////////////

  Void testBind()
  {
    verifyBind(null, null)
    verifyBind(IpAddress.local, null)
    verifyBind(null, 1972)
    verifyBind(IpAddress.local, 1973)
  }

  Void verifyBind(IpAddress addr, Int port)
  {
    s := TcpSocket.make
    verifySame(s.bind(addr, port), s)

    // state
    verifyEq(s.isBound, true)
    verifyEq(s.isConnected, false)
    verifyEq(s.isClosed, false)
    verifyErr(IOErr.type) |,| { s.in }
    verifyErr(IOErr.type) |,| { s.out }

    // local address
    if (addr == null)
      verify(s.localAddress != null)
    else
      verifyEq(s.localAddress, addr)

    // local port
    if (port == null)
      verify(s.localPort > 0)
    else
      verifyEq(s.localPort, port)

    // null remote
    verifyEq(s.remoteAddress, null)
    verifyEq(s.remotePort, null)

    // duplicate port
    x := TcpSocket.make
    verifyErr(IOErr.type) |,| { x.bind(null, s.localPort) }

    // cleanup
    s.close
    x.close

    verifyEq(s.isClosed, true)
    verifyErr(IOErr.type) |,| { s.in }
    verifyErr(IOErr.type) |,| { s.out }
  }

//////////////////////////////////////////////////////////////////////////
// Connection Failures
//////////////////////////////////////////////////////////////////////////

  Void testConnectFailures()
  {
    // local, invalid port
    s := TcpSocket.make
    verifyErr(IOErr.type) |,| { s.connect(IpAddress.local, 1969) }
    verifyEq(s.isConnected, false)
    verifyErr(IOErr.type) |,| { s.in }
    verifyErr(IOErr.type) |,| { s.out }
    s.close

    // invalid host
    t1 := Duration.now
    s = TcpSocket.make
    verifyErr(IOErr.type) |,| { s.connect(IpAddress.make("1.1.1.1"), 1969, 100ms) }
    t2 := Duration.now
    verifyEq(s.isConnected, false)
    verify(80ms < t2-t1 && t2-t1 < 120ms)
    s.close

    verifyEq(s.isClosed, true)
    verifyErr(IOErr.type) |,| { s.in }
    verifyErr(IOErr.type) |,| { s.out }
  }

//////////////////////////////////////////////////////////////////////////
// Connect
//////////////////////////////////////////////////////////////////////////

  Void testConnectHttp()
  {
    // connect to www server
    s := TcpSocket.make.connect(IpAddress.make("www.fandev.org"), 80)

    // verify connetion state
    verifyEq(s.isBound, true)
    verifyEq(s.isConnected, true)
    verifyEq(s.isClosed, false)
    verify(s.in != null)
    verify(s.out != null)
    verifyErr(Err.type) |,| { s.options.inBufferSize = 16 }
    verifyErr(Err.type) |,| { s.options.outBufferSize = 16 }

    // send very simple request line
    s.out.print("GET / HTTP/1.0\n\r\n\r").flush

    // read first response line
    res := s.in.readLine
    verify(res.startsWith("HTTP/1.1 200 OK"))

    // cleanup
    s.close
    verifyEq(s.isClosed, true)
    verifyErr(IOErr.type) |,| { s.in }
    verifyErr(IOErr.type) |,| { s.out }
  }

//////////////////////////////////////////////////////////////////////////
// Fork
//////////////////////////////////////////////////////////////////////////

  /* TODO - remove when we finalize that TcpSocket should be const
  Void testFork()
  {
    // verify duplicate name
    verifyErr(ArgErr.type) |,|
    {
       x := TcpSocket.make.connect(IpAddress.make("www.fandev.org"), 80)
       x.fork(Thread.current.name, &runFork(x.localPort, x.remoteAddress.numeric))
    }

    // verify non-const method
    verifyErr(NotImmutableErr.type) |,|
    {
       x := TcpSocket.make.connect(IpAddress.make("www.fandev.org"), 80)
       x.fork(null) |TcpSocket s| { fail }
    }

    // connect to www server
    s := TcpSocket.make.connect(IpAddress.make("www.fandev.org"), 80)
    so := s.options

    // fork
    t := s.fork(null, &runFork(s.localPort, s.remoteAddress.numeric))

    // verify that all methods on s now throw UnsupportedErr
    verifyErr(UnsupportedErr.type) |,| { s.isBound }
    verifyErr(UnsupportedErr.type) |,| { s.isConnected }
    verifyErr(UnsupportedErr.type) |,| { s.isClosed }
    verifyErr(UnsupportedErr.type) |,| { s.localAddress }
    verifyErr(UnsupportedErr.type) |,| { s.localPort }
    verifyErr(UnsupportedErr.type) |,| { s.remoteAddress }
    verifyErr(UnsupportedErr.type) |,| { s.remotePort  }
    verifyErr(UnsupportedErr.type) |,| { s.bind(null, null) }
    verifyErr(UnsupportedErr.type) |,| { s.connect(null, null) }
    verifyErr(UnsupportedErr.type) |,| { s.in }
    verifyErr(UnsupportedErr.type) |,| { s.out }
    verifyErr(UnsupportedErr.type) |,| { s.close }
    verifyErr(UnsupportedErr.type) |,| { s.fork(null, null) }
    verifyErr(UnsupportedErr.type) |,| { s.options }

    // verify that all socket options now throw UnsupportedErr
    verifyErr(UnsupportedErr.type) |,| { echo(so.inBufferSize) }
    verifyErr(UnsupportedErr.type) |,| { so.inBufferSize = 100}
    verifyErr(UnsupportedErr.type) |,| { echo(so.outBufferSize) }
    verifyErr(UnsupportedErr.type) |,| { so.outBufferSize = 100}
    verifyErr(UnsupportedErr.type) |,| { echo(so.keepAlive) }
    verifyErr(UnsupportedErr.type) |,| { so.keepAlive = false }
    verifyErr(UnsupportedErr.type) |,| { echo(so.receiveBufferSize) }
    verifyErr(UnsupportedErr.type) |,| { so.receiveBufferSize = 10}
    verifyErr(UnsupportedErr.type) |,| { echo(so.sendBufferSize) }
    verifyErr(UnsupportedErr.type) |,| { so.sendBufferSize = 10}
    verifyErr(UnsupportedErr.type) |,| { echo(so.linger) }
    verifyErr(UnsupportedErr.type) |,| { so.linger = null }
    verifyErr(UnsupportedErr.type) |,| { echo(so.receiveTimeout) }
    verifyErr(UnsupportedErr.type) |,| { so.receiveTimeout = null }
    verifyErr(UnsupportedErr.type) |,| { echo(so.noDelay) }
    verifyErr(UnsupportedErr.type) |,| { so.noDelay = false }
    verifyErr(UnsupportedErr.type) |,| { echo(so.trafficClass) }
    verifyErr(UnsupportedErr.type) |,| { so.trafficClass = 0 }

    // join and verify response
    verifyEq(t.join, "HTTP/1.1 200 OK")
  }

  static Obj runFork(Int localPort, Str remoteAddr, TcpSocket s)
  {
    // verify state of new detached socket
    if (s.localPort != localPort) return null
    if (s.remoteAddress.numeric != remoteAddr) return null
    if (s.remotePort != 80) return null

    // send very simple request line
    s.out.print("GET / HTTP/1.0\n\r\n\r").flush

    // read first response line
    res := s.in.readLine

    // cleanup and return response
    s.close
    return res
  }
  */

//////////////////////////////////////////////////////////////////////////
// Options
//////////////////////////////////////////////////////////////////////////

  Void testOptions()
  {
    s := TcpSocket.make
    so := s.options

    in := so.inBufferSize
    so.inBufferSize = in*2
    verifyEq(so.inBufferSize, in*2)

    out := so.outBufferSize
    so.outBufferSize = in*2+1
    verifyEq(so.outBufferSize, in*2+1)

    keepAlive := so.keepAlive
    so.keepAlive = !keepAlive
    verifyEq(so.keepAlive, !keepAlive)

    receive := so.receiveBufferSize
    so.receiveBufferSize = receive*2
    verifyEq(so.receiveBufferSize, receive*2)

    send := so.sendBufferSize
    so.sendBufferSize = send*4
    verifyEq(so.sendBufferSize, send*4)

    reuse := so.reuseAddress
    so.reuseAddress = !reuse
    verifyEq(so.reuseAddress, !reuse)

    so.linger = 2sec
    verifyEq(so.linger, 2sec)
    so.linger = null
    verifyEq(so.linger, null)

    so.receiveTimeout = 100ms
    verifyEq(so.receiveTimeout, 100ms)
    so.receiveTimeout = null
    verifyEq(so.receiveTimeout, null)

    noDelay := so.noDelay
    so.noDelay = !noDelay
    verifyEq(so.noDelay, !noDelay)

    tc := so.trafficClass
    so.trafficClass = 0x6
    verifyEq(so.trafficClass, 0x6)

    verifyErr(UnsupportedErr.type) |,| { echo(so.broadcast) }
    verifyErr(UnsupportedErr.type) |,| { so.broadcast = false }

    s.close
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  /*
  Void dump(TcpSocket s)
  {
    echo("---------")
    echo("bound      = $s.isBound")
    echo("connected  = $s.isConnected")
    echo("closed     = $s.isClosed")
    echo("localAddr  = $s.localAddress")
    echo("localPort  = $s.localPort")
    echo("remoteAddr = $s.remoteAddress")
    echo("remotePort = $s.remotePort")
    echo("keepAlive  = $s.options.keepAlive")
    echo("receive    = $s.options.receiveBufferSize")
    echo("send       = $s.options.sendBufferSize")
    echo("reuseAddr  = $s.options.reuseAddress")
    echo("linger     = $s.options.linger")
    echo("timeout    = $s.options.receiveTimeout")
    echo("noDelay    = $s.options.noDelay")
    echo("trafficCls = 0x$s.options.trafficClass.toHex")
  }
  */

}