//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Apr 06  Brian Frank  Creation
//
package fan.sys;

import java.io.*;
import sun.nio.cs.StreamEncoder;

/**
 * Buf
 */
public class Buf
  extends FanObj
  implements OutStream, InStream
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public static Buf make() { return new Buf().init(DefaultCapacity); }
  public static Buf make(int capacity) { return new Buf().init(capacity); }
  public static Buf make(Int capacity) { return new Buf().init((int)capacity.val); }
  public static void make$(Buf self) { self.init(DefaultCapacity ); }
  public static void make$(Buf self, Int capacity) { self.init((int)capacity.val); }

  public static Buf make(byte[] bytes)
  {
    Buf buf  = new Buf();
    buf.buf  = bytes;
    buf.pos  = 0;
    buf.size = bytes.length;
    return buf;
  }

  public Buf init(int capacity)
  {
    this.buf  = new byte[capacity];
    this.pos  = 0;
    this.size = 0;
    return this;
  }

  static final int DefaultCapacity = 1024;

//////////////////////////////////////////////////////////////////////////
// Obj
//////////////////////////////////////////////////////////////////////////

  public Bool equals(Obj that)
  {
    if (that instanceof Buf)
    {
      Buf x = (Buf)that;
      if (size != x.size) return Bool.False;
      for (int i=0; i<size; ++i)
        if (buf[i] != x.buf[i]) return Bool.False;
      return Bool.True;
    }
    return Bool.False;
  }

  public Int hash()
  {
    int hash = 30311;
    for (int i=0; i<size; ++i)
      hash ^= buf[i] << i%32;
    return Int.make(hash);
  }

  public Str toStr()
  {
    StringBuilder s = new StringBuilder();
    s.append("0x[");
    for (int i=0; i<size; ++i)
    {
      int b = buf[i] & 0xFF;
      if (b < 0x10) s.append('0');
      s.append(Integer.toHexString(b));
      if (i > 100) { s.append("..."); break; }
      if ((i+1) % 4 == 0 && i+1<size) s.append(' ');
    }
    s.append(']');
    return Str.make(s.toString());
  }

  public Type type() { return Sys.BufType; }

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  public Bool empty() { return size == 0 ? Bool.True : Bool.False; }

  public final Int capacity()
  {
    return Int.pos(buf.length);
  }

  public final void capacity(Int c)
  {
    int newCapacity = (int)c.val;
    if (newCapacity < size) throw ArgErr.make("capacity < size").val;
    byte[] temp = new byte[newCapacity];
    System.arraycopy(buf, 0, temp, 0, newCapacity);
    buf = temp;
  }

  public Int size()
  {
    return Int.make(size);
  }

  public void size(Int s)
  {
    int newSize = (int)s.val;
    if (newSize > buf.length)
    {
      byte[] temp = new byte[newSize];
      System.arraycopy(buf, 0, temp, 0, buf.length);
      buf  = temp;
    }
    size = newSize;
  }

  public Int pos() { return Int.make(pos); }

  public Int remaining() { return Int.make(size-pos); }

  public Bool more() { return size-pos > 0 ? Bool.True : Bool.False; }

  public Buf seek(Int pos)
  {
    int p = (int)pos.val;
    if (p < 0) p = size + p;
    if (p < 0 || p > size) throw IndexErr.make(pos).val;
    this.pos = p;
    return this;
  }

  public Buf flip()
  {
    this.size = pos;
    this.pos  = 0;
    return this;
  }

  public Int get(Int pos)
  {
    int p = (int)pos.val;
    if (p < 0) p = size + p;
    if (p < 0 || p >= size) throw IndexErr.make(pos).val;
    return Int.make(buf[p] & 0xFF);
  }

  public Buf slice(Range range)
  {
    int s = range.start(size);
    int e = range.end(size);
    int n = e - s + 1;
    if (n < 0) throw IndexErr.make(range).val;

    byte[] slice = new byte[n];
    System.arraycopy(buf, s, slice, 0, n);

    Buf result = new Buf();
    result.buf = slice;
    result.size = n;
    result.charset = charset;
    return result;
  }

//////////////////////////////////////////////////////////////////////////
// Modification
//////////////////////////////////////////////////////////////////////////

  public Buf set(Int pos, Int b)
  {
    int p = (int)pos.val;
    if (p < 0) p = size + p;
    if (p < 0 || p >= size) throw IndexErr.make(pos).val;
    buf[p] = (byte)b.val;
    return this;
  }

  public Buf trim()
  {
    if (size == buf.length) return this;
    byte[] temp = new byte[size];
    System.arraycopy(buf, 0, temp, 0, size);
    this.buf = temp;
    return this;
  }

  public Buf clear()
  {
    this.size = 0;
    this.pos  = 0;
    return this;
  }

//////////////////////////////////////////////////////////////////////////
// OutStream
//////////////////////////////////////////////////////////////////////////

  public OutStream write(Int b) { return w((int)b.val); }
  public final OutStream w(int v)
  {
    if (pos+1 >= buf.length) grow(pos+1);
    buf[pos++] = (byte)v;
    if (pos > size) size = pos;
    return this;
  }

  public OutStream writeBuf(Buf other) { return writeBuf(other, other.remaining()); }
  public OutStream writeBuf(Buf other, Int n)
  {
    int len = (int)n.val;
    if (other.pos + len > other.size) throw IOErr.make("Not enough bytes to write").val;
    grow(pos+len);
    System.arraycopy(other.buf, other.pos, buf, pos, len);
    pos += len;
    if (pos > size) size = pos;
    other.pos += len;
    return this;
  }

  public OutStream writeI2(Int x) { return OutStream$.writeI2(this, x); }

  public OutStream writeI4(Int x) { return OutStream$.writeI4(this, x); }

  public OutStream writeI8(Int x) { return OutStream$.writeI8(this, x); }

  public OutStream writeF4(Float x) { return OutStream$.writeF4(this, x); }

  public OutStream writeF8(Float x) { return OutStream$.writeF8(this, x); }

  public OutStream writeBool(Bool x) { return OutStream$.writeBool(this, x); }

  public OutStream writeUtf(Str x) { return OutStream$.writeUtf(this, x); }

  public OutStream writeChar(Int c)
  {
    if (charsetEncoder == null) charsetEncoder = charset.newEncoder();
    charsetEncoder.encode((char)c.val, (OutStream)this);
    return this;
  }

  public OutStream writeChars(Str s) { return writeChars(s.val, 0, s.val.length()); }
  public OutStream writeChars(Str s, Int off) { return writeChars(s.val, (int)off.val, s.val.length()-(int)off.val); }
  public OutStream writeChars(Str s, Int off, Int len) { return writeChars(s.val, (int)off.val, (int)len.val); }
  public OutStream writeChars(String s, int off, int len)
  {
    if (charsetEncoder == null) charsetEncoder = charset.newEncoder();
    int end = off+len;
    for (int i=off; i<end; ++i)
      charsetEncoder.encode(s.charAt(i), (OutStream)this);
    return this;
  }

  public OutStream print(Obj obj) { return OutStream$.print(this, obj); }

  public OutStream printLine() { return OutStream$.printLine(this, Str.Empty); }
  public OutStream printLine(Obj obj) { return OutStream$.printLine(this, obj); }

  public OutStream writeProps(Map props) { return OutStream$.writeProps(this, props, Bool.True); }
  public OutStream writeProps(Map props, Bool close) { return OutStream$.writeProps(this, props, close); }

//////////////////////////////////////////////////////////////////////////
// InStream
//////////////////////////////////////////////////////////////////////////

  public Int read() {  int n = r(); return n < 0 ? null : Int.pos(n); }
  public final int r()
  {
    if (pos >= size) return -1;
    return buf[pos++] & 0xFF;
  }

  public Int readBuf(Buf other, Int n)
  {
    if (pos >= size) return null;
    int len = Math.min(size-pos, (int)n.val);
    other.grow(other.pos+len);
    System.arraycopy(buf, pos, other.buf, other.pos, len);
    pos += len;
    other.pos += len;
    other.size += len;
    return Int.pos(len);
  }

  public InStream unread(Int n) { return unread((int)n.val); }
  public InStream unread(int n)
  {
    // unreading a buffer is a bit weird - the typical case
    // is that we are pushing back the byte we just read in
    // which case we can just rewind the position; however
    // if we pushing back a different byte then we need
    // to shift the entire buffer and insert the byte
    if (pos > 0 && buf[pos-1] == (byte)n)
    {
      pos--;
    }
    else
    {
      if (size+1 >= buf.length) grow(size+1);
      System.arraycopy(buf, pos, buf, pos+1, size);
      buf[pos] = (byte)n;
      size++;
    }
    return this;
  }

  public Buf readBufFully(Buf buf, Int n) { return InStream$.readBufFully(this, buf, n); }

  public Buf readAllBuf() { return InStream$.readAllBuf(this); }

  public Int readU1() { return InStream$.readU1(this); }

  public Int readS1() { return InStream$.readS1(this); }

  public Int readU2() { return InStream$.readU2(this); }

  public Int readS2() { return InStream$.readS2(this); }

  public Int readU4() { return InStream$.readU4(this); }

  public Int readS4() { return InStream$.readS4(this); }

  public Int readS8() { return InStream$.readS8(this); }

  public Float readF4() { return InStream$.readF4(this); }

  public Float readF8() { return InStream$.readF8(this); }

  public Bool readBool() { return InStream$.readBool(this); }

  public Str readUtf() { return InStream$.readUtf(this); }

  public int rChar()
  {
    if (charsetDecoder == null) charsetDecoder = charset.newDecoder();
    return charsetDecoder.decode(this);
  }

  public Int readChar()
  {
    if (charsetDecoder == null) charsetDecoder = charset.newDecoder();
    int ch = charsetDecoder.decode(this);
    return ch < 0 ? null : Int.pos(ch);
  }

  public InStream unreadChar(Int c)
  {
    if (charsetEncoder == null) charsetEncoder = charset.newEncoder();
    charsetEncoder.encode((char)c.val, (InStream)this);
    return this;
  }

  public Str readLine() { return InStream$.readLine(this, Int.pos[4096]); }
  public Str readLine(Int max) { return InStream$.readLine(this, max); }

  public List readAllLines() { return InStream$.readAllLines(this); }

  public Str readAllStr() { return InStream$.readAllStr(this, Bool.True); }
  public Str readAllStr(Bool normalizeNewlines)  { return InStream$.readAllStr(this, normalizeNewlines); }

  public Map readProps() { return InStream$.readProps(this); }

  public Int pipe(OutStream out) { return InStream$.pipe(this, out, null, Bool.True); }
  public Int pipe(OutStream out, Int n) { return InStream$.pipe(this, out, n, Bool.True); }
  public Int pipe(OutStream out, Int n, Bool close) { return InStream$.pipe(this, out, n, close); }

//////////////////////////////////////////////////////////////////////////
// OutStream and InStream
//////////////////////////////////////////////////////////////////////////

  public void flush() { }

  public Bool close() { return Bool.True; }

  public Charset charset()
  {
    return charset;
  }

  public void charset(Charset charset)
  {
    this.charset = charset;
    this.charsetEncoder = null;
    this.charsetDecoder = null;
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  public byte[] bytes()
  {
    byte[] r = new byte[size];
    System.arraycopy(buf, 0, r, 0, size);
    return r;
  }

  public void grow(int capacity)
  {
    if (buf.length >= capacity) return;
    byte[] temp = new byte[Math.max(capacity, size*2)];
    System.arraycopy(buf, 0, temp, 0, size);
    buf = temp;
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  Charset charset = Charset.utf8();
  Charset.Encoder charsetEncoder;
  Charset.Decoder charsetDecoder;
  public byte[] buf;
  public int pos;
  public int size;

}