//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Mar 06  Brian Frank  Creation
//
package fan.sys;

/**
 * File represents a file or directory in a file system.
 */
public abstract class File
  extends FanObj
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public static File make(Uri uri)
  {
    return new LocalFile(uri);
  }

  public static File os(Str osPath)
  {
    return new LocalFile(new java.io.File(osPath.val));
  }

  public static File createTemp() { return createTemp(null, null, null); }
  public static File createTemp(Str prefix) { return createTemp(prefix, null, null); }
  public static File createTemp(Str prefix, Str suffix) { return createTemp(prefix, suffix, null); }
  public static File createTemp(Str prefix, Str suffix, File dir)
  {
    if (prefix == null || prefix.val.length() == 0) prefix = Str.make("fan");
    if (prefix.val.length() == 1) prefix = Str.make(prefix.val + "xx");
    if (prefix.val.length() == 2) prefix = Str.make(prefix.val + "x");

    if (suffix == null) suffix = Str.make(".tmp");

    java.io.File d = null;
    if (dir != null)
    {
      if (!(dir instanceof LocalFile)) throw IOErr.make("Dir is not on local file system: " + dir).val;
      d = ((LocalFile)dir).file;
    }

    try
    {
      return new LocalFile(java.io.File.createTempFile(prefix.val, suffix.val, d));
    }
    catch (java.io.IOException e)
    {
      throw IOErr.make(e).val;
    }
  }

  protected File(Uri uri)
  {
    this.uri = uri;
  }

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  public final Bool equals(Obj obj)
  {
    if (obj instanceof File)
    {
      return uri.equals(((File)obj).uri);
    }
    return Bool.False;
  }

  public final int hashCode() { return uri.hashCode(); }

  public final Int hash() { return uri.hash(); }

  public final Str toStr() { return uri.toStr(); }

  public Type type() { return Sys.FileType; }

//////////////////////////////////////////////////////////////////////////
// Naming
//////////////////////////////////////////////////////////////////////////

  public final Uri uri()
  {
    return uri;
  }

  public final Bool isDir()
  {
    return uri.isDir();
  }

  public final List path()
  {
    return uri.path();
  }

  public final Str pathStr()
  {
    return uri.pathStr();
  }

  public final Str name()
  {
    return uri.name();
  }

  public final Str basename()
  {
    return uri.basename();
  }

  public final Str ext()
  {
    return uri.ext();
  }

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  public abstract Bool exists();

  public abstract Int size();

  public abstract Time modified();
  public abstract void modified(Time time);

  public abstract Str osPath();

  public abstract File parent();

  public abstract List list();

  public abstract File normalize();

  public abstract File plus(Uri uri);

//////////////////////////////////////////////////////////////////////////
// Management
//////////////////////////////////////////////////////////////////////////

  public abstract File create();

  public abstract void delete();

  public abstract File deleteOnExit();

//////////////////////////////////////////////////////////////////////////
// IO
//////////////////////////////////////////////////////////////////////////

  public final InStream in() { return in(defaultBufSize); }
  public abstract InStream in(Int bufSize);

  public final OutStream out() { return out(Bool.False, defaultBufSize); }
  public final OutStream out(Bool append) { return out(append, defaultBufSize); }
  public abstract OutStream out(Bool append, Int bufSize);

  private static final Int defaultBufSize = Int.make(4096);

  public final Buf readAllBuf()
  {
    return in(defaultBufSize).readAllBuf();
  }

  public final List readAllLines()
  {
    return in(defaultBufSize).readAllLines();
  }

  public final Str readAllStr() { return readAllStr(Bool.True); }
  public final Str readAllStr(Bool normalizeNewlines)
  {
    return in(defaultBufSize).readAllStr(normalizeNewlines);
  }

  public final Map readProps()
  {
    return in(defaultBufSize).readProps();
  }

  public final void writeProps(Map props)
  {
    out(Bool.False, defaultBufSize).writeProps(props, Bool.True);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  final Uri uri;

}