//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Mar 06  Brian Frank  Creation
//
package fan.sys;

import fan.sys.Float;

/**
 * InStream implementation.
 */
public class InStream$
{
  public static InStream makeForStr(Str s)
  {
    return new StrInStream(s);
  }

  public static Buf readAllBuf(InStream self)
  {
    try
    {
      Int size = Int.pos[4096];
      Buf buf = Buf.make(size);
      while (self.readBuf(buf, size) != null);
      buf.flip();
      return buf;
    }
    finally
    {
      try { self.close(); } catch (Exception e) { e.printStackTrace(); }
    }
  }

  public static Buf readBufFully(InStream self, Buf buf, Int n)
  {
    if (buf == null) buf = Buf.make(n);

    long total = n.val;
    long got = 0;
    while (got < total)
    {
      Int r = self.readBuf(buf, Int.make(total-got));
      if (r == null || r.val == 0) throw IOErr.make("Unexpected end of stream").val;
      got += r.val;
    }

    buf.flip();
    return buf;
  }

  public static Int readU1(InStream self)
  {
    int c = self.r();
    if (c < 0) throw IOErr.make("Unexpected end of stream").val;
    return Int.make(c);
  }

  public static Int readS1(InStream self)
  {
    int c = self.r();
    if (c < 0) throw IOErr.make("Unexpected end of stream").val;
    return Int.make((byte)c);
  }

  public static Int readU2(InStream self)
  {
    int c1 = self.r();
    int c2 = self.r();
    if ((c1 | c2) < 0) throw IOErr.make("Unexpected end of stream").val;
    return Int.make(c1 << 8 | c2);
  }

  public static Int readS2(InStream self)
  {
    int c1 = self.r();
    int c2 = self.r();
    if ((c1 | c2) < 0) throw IOErr.make("Unexpected end of stream").val;
    return Int.make((short)(c1 << 8 | c2));
  }

  public static Int readU4(InStream self)
  {
    long c1 = self.r();
    long c2 = self.r();
    long c3 = self.r();
    long c4 = self.r();
    if ((c1 | c2 | c3 | c4) < 0) throw IOErr.make("Unexpected end of stream").val;
    return Int.make((c1 << 24) + (c2 << 16) + (c3 << 8) + c4);
  }

  public static Int readS4(InStream self) { return Int.make(readInt(self)); }
  public static int readInt(InStream self)
  {
    int c1 = self.r();
    int c2 = self.r();
    int c3 = self.r();
    int c4 = self.r();
    if ((c1 | c2 | c3 | c4) < 0) throw IOErr.make("Unexpected end of stream").val;
    return ((c1 << 24) + (c2 << 16) + (c3 << 8) + c4);
  }

  public static Int readS8(InStream self) { return Int.make(readLong(self)); }
  public static long readLong(InStream self)
  {
    long c1 = self.r();
    long c2 = self.r();
    long c3 = self.r();
    long c4 = self.r();
    long c5 = self.r();
    long c6 = self.r();
    long c7 = self.r();
    long c8 = self.r();
    if ((c1 | c2 | c3 | c4 | c5 | c6 | c7 | c8) < 0) throw IOErr.make("Unexpected end of stream").val;
    return ((c1 << 56) + (c2 << 48) + (c3 << 40) + (c4 << 32) +
            (c5 << 24) + (c6 << 16) + (c7 << 8) + c8);
  }

  public static Float readF4(InStream self)
  {
    return Float.make(java.lang.Float.intBitsToFloat(readInt(self)));
  }

  public static Float readF8(InStream self)
  {
    return Float.make(Double.longBitsToDouble(readLong(self)));
  }

  public static Bool readBool(InStream self)
  {
    int n = self.r();
    if (n < 0) throw IOErr.make("Unexpected end of stream").val;
    return n == 0 ? Bool.False : Bool.True;
  }

  public static Str readUtf(InStream self)
  {
    // read two-byte length
    int len1 = self.r();
    int len2 = self.r();
    if ((len1 | len2) < 0) throw IOErr.make("Unexpected end of stream").val;
    int utflen = len1 << 8 | len2;

    char[] buf = new char[utflen]; // char buffer we read into
    int bnum = 0, cnum = 0;        // byte count, char count

    // read the chars
    int c, c2, c3;
    while (bnum < utflen)
    {
      c = self.r(); bnum++;
      if (c < 0) throw IOErr.make("Unexpected end of stream").val;
      switch (c >> 4) {
        case 0: case 1: case 2: case 3: case 4: case 5: case 6: case 7:
          /* 0xxxxxxx*/
          buf[cnum++]=(char)c;
          break;
        case 12: case 13:
          /* 110x xxxx   10xx xxxx*/
          if (bnum >= utflen) throw IOErr.make("UTF encoding error").val;
          c2 = self.r(); bnum++;
          if (c2 < 0) throw IOErr.make("Unexpected end of stream").val;
          if ((c2 & 0xC0) != 0x80) throw IOErr.make("UTF encoding error").val;
          buf[cnum++]=(char)(((c & 0x1F) << 6) | (c2 & 0x3F));
          break;
        case 14:
          /* 1110 xxxx  10xx xxxx  10xx xxxx */
          if (bnum+1 >= utflen) throw IOErr.make("UTF encoding error").val;
          c2 = self.r(); bnum++;
          c3 = self.r(); bnum++;
          if ((c2|c3) < 0) throw IOErr.make("Unexpected end of stream").val;
          if (((c2 & 0xC0) != 0x80) || ((c3 & 0xC0) != 0x80))  throw IOErr.make("UTF encoding error").val;
          buf[cnum++]=(char)(((c & 0x0F) << 12) | ((c2 & 0x3F) << 6) | ((c3 & 0x3F) << 0));
          break;
        default:
          /* 10xx xxxx,  1111 xxxx */
          throw IOErr.make("UTF encoding error").val;
      }
    }

    // allocate as Str
    return Str.make(new String(buf, 0, cnum));
  }

  public static Str readLine(InStream self, Int max)
  {
    // max limit
    int maxChars = (max != null) ? (int)max.val : Integer.MAX_VALUE;

    // read first char, if at end of file bail
    int c = self.rChar();
    if (c < 0) return null;

    // loop reading chars until we hit newline
    // combo or end of stream
    StringBuilder buf = new StringBuilder();
    while (true)
    {
      // check for \n, \r\n, or \r
      if (c == '\n') break;
      if (c == '\r')
      {
        c = self.rChar();
        if (c >= 0 && c != '\n') self.unreadChar(Int.make(c));
        break;
      }

      // append to working buffer
      buf.append((char)c);
      if (buf.length() >= maxChars) break;

      // read next char
      c = self.rChar();
      if (c < 0) break;
    }
    return Str.make(buf.toString());
  }

  public static List readAllLines(InStream self)
  {
    try
    {
      List list = new List(Sys.StrType);
      Str line;
      while ((line = self.readLine()) != null)
        list.add(line);
      return list;
    }
    finally
    {
      try { self.close(); } catch (Exception e) { e.printStackTrace(); }
    }
  }

  public static Str readAllStr(InStream self, Bool normalizeNewlines)
  {
    try
    {
      char[] buf  = new char[4096];
      int n = 0;
      boolean normalize = normalizeNewlines.val;

      // read characters
      int last = -1;
      while (true)
      {
        int c = self.rChar();
        if (c < 0) break;

        // grow buffer if needed
        if (n >= buf.length)
        {
          char[] temp = new char[buf.length*2];
          System.arraycopy(buf, 0, temp, 0, n);
          buf = temp;
        }

        // normalize newlines and add to buffer
        if (normalize)
        {
          if (c == '\r') buf[n++] = '\n';
          else if (last == '\r' && c == '\n') {}
          else buf[n++] = (char)c;
          last = c;
        }
        else
        {
          buf[n++] = (char)c;
        }
      }

      return Str.make(new String(buf, 0, n));
    }
    finally
    {
      try { self.close(); } catch (Exception e) { e.printStackTrace(); }
    }
  }

  public static Map readProps(InStream self)
  {
    Charset origCharset = self.charset();
    self.charset(Charset.utf8());
    try
    {
      Map props = new Map(Sys.StrType, Sys.StrType);

      StringBuilder name = new StringBuilder();
      StringBuilder val = null;
      int inBlockComment = 0;
      boolean inEndOfLineComment = false;
      int c = -1, last = -1;
      int lineNum = 1;

      while (true)
      {
        last = c;
        c = self.rChar();
        if (c < 0) break;

        // end of line
        if (c == '\n' || c == '\r')
        {
          inEndOfLineComment = false;
          if (last == '\r' && c == '\n') continue;
          Str n = Str.makeTrim(name);
          if (val != null)
          {
            props.add(n, Str.makeTrim(val));
            name = new StringBuilder();
            val = null;
          }
          else if (n.val.length() > 0)
            throw IOErr.make("Invalid name/value pair [Line " + lineNum + "]").val;
          lineNum++;
          continue;
        }

        // if in comment
        if (inEndOfLineComment) continue;

        // block comment
        if (inBlockComment > 0)
        {
          if (last == '/' && c == '*') inBlockComment++;
          if (last == '*' && c == '/') inBlockComment--;
          continue;
        }

        // equal
        if (c == '=' && val == null)
        {
          val = new StringBuilder();
          continue;
        }

        // comment
        if (c == '/')
        {
          int peek = self.rChar();
          if (peek < 0) break;
          if (peek == '/') { inEndOfLineComment = true; continue; }
          if (peek == '*') { inBlockComment++; continue; }
          self.unreadChar(Int.make(peek));
        }

        // escape or line continuation
        if (c == '\\')
        {
          int peek = self.rChar();
          if (peek < 0) break;
          else if (peek == 'n')  c = '\n';
          else if (peek == 'r')  c = '\r';
          else if (peek == 't')  c = '\t';
          else if (peek == '\\') c = '\\';
          else if (peek == '\r' || peek == '\n')
          {
            // line continuation
            lineNum++;
            if (peek == '\r')
            {
              peek = self.rChar();
              if (peek != '\n') self.unreadChar(Int.make(peek));
            }
            while (true)
            {
              peek = self.rChar();
              if (peek == ' ' || peek == '\t') continue;
              self.unreadChar(Int.make(peek));
              break;
            }
            continue;
          }
          else if (peek == 'u')
          {
            int n3 = hex(self.rChar());
            int n2 = hex(self.rChar());
            int n1 = hex(self.rChar());
            int n0 = hex(self.rChar());
            if (n3 < 0 || n2 < 0 || n1 < 0 || n0 < 0) throw IOErr.make("Invalid hex value for \\uxxxx [Line " +  lineNum + "]").val;
            c = ((n3 << 12) | (n2 << 8) | (n1 << 4) | n0);
          }
          else throw IOErr.make("Invalid escape sequence [Line " + lineNum + "]").val;
        }

        // normal character
        if (val == null)
          name.append((char)c);
        else
          val.append((char)c);
      }

      Str n = Str.makeTrim(name);
      if (val != null)
        props.add(n, Str.makeTrim(val));
      else if (n.val.length() > 0)
        throw IOErr.make("Invalid name/value pair [Line " + lineNum + "]").val;

      return props;
    }
    finally
    {
      try { self.close(); } catch (Exception e) { e.printStackTrace(); }
      self.charset(origCharset);
    }
  }

  static int hex(int c)
  {
    if ('0' <= c && c <= '9') return c - '0';
    if ('a' <= c && c <= 'f') return c - 'a' + 10;
    if ('A' <= c && c <= 'F') return c - 'A' + 10;
    return -1;
  }

  public static Int pipe(InStream self, OutStream out, Int toPipe, Bool close)
  {
    try
    {
      Int bufSize = Int.pos[4096];
      Buf buf = Buf.make(bufSize);
      long total = 0;
      if (toPipe == null)
      {
        while (true)
        {
          Int n = self.readBuf(buf.clear(), bufSize);
          if (n == null) break;
          out.writeBuf(buf.flip(), buf.remaining());
          total += n.val;
        }
      }
      else
      {
        while (total < toPipe.val )
        {
          if (toPipe.val - total < bufSize.val) bufSize = Int.make(toPipe.val - total);
          Int n = self.readBuf(buf, bufSize);
          if (n == null) throw IOErr.make("Unexpected end of stream").val;
          out.writeBuf(buf.flip(), buf.remaining());
          total += n.val;
        }
      }
      return Int.make(total);
    }
    finally
    {
      if (close.val) self.close();
    }
  }

  public static Bool close(InStream self)
  {
    return Bool.True;
  }

}