//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Mar 06  Brian Frank  Creation
//
package fan.sys;

import fan.sys.Float;

/**
 * OutStream implementation.
 */
public class OutStream$
{
  public static OutStream makeForStrBuf(StrBuf buf)
  {
    return new StrBufOutStream(buf);
  }

  public static OutStream writeI2(OutStream self, Int x)
  {
    int v = (int)x.val;
    return self.w((v >>> 8) & 0xFF)
               .w((v >>> 0) & 0xFF);
  }

  public static OutStream writeI4(OutStream self, Int x) { return writeI4(self, (int)x.val); }
  public static OutStream writeI4(OutStream self, int v)
  {
    return self.w((v >>> 24) & 0xFF)
               .w((v >>> 16) & 0xFF)
               .w((v >>> 8)  & 0xFF)
               .w((v >>> 0)  & 0xFF);
  }

  public static OutStream writeI8(OutStream self, Int x) { return writeI8(self, x.val); }
  public static OutStream writeI8(OutStream self, long v)
  {
    return self.w((int)(v >>> 56) & 0xFF)
               .w((int)(v >>> 48) & 0xFF)
               .w((int)(v >>> 40) & 0xFF)
               .w((int)(v >>> 32) & 0xFF)
               .w((int)(v >>> 24) & 0xFF)
               .w((int)(v >>> 16) & 0xFF)
               .w((int)(v >>> 8)  & 0xFF)
               .w((int)(v >>> 0)  & 0xFF);
  }

  public static OutStream writeF4(OutStream self, Float x)
  {
    return writeI4(self, java.lang.Float.floatToIntBits((float)x.val));
  }

  public static OutStream writeF8(OutStream self, Float x)
  {
    return writeI8(self, Double.doubleToLongBits(x.val));
  }

  public static OutStream writeBool(OutStream self, Bool x)
  {
    return self.w(x.val ? 1 : 0);
  }

  public static OutStream writeUtf(OutStream self, Str x)
  {
    String s = x.val;
    int slen = s.length();
    int utflen = 0;

    // first we have to figure out the utf length
    for (int i=0; i<slen; ++i)
    {
      int c = s.charAt(i);
      if (c <= 0x007F)
        utflen +=1;
      else if (c > 0x07FF)
        utflen += 3;
      else
        utflen += 2;
    }

    // sanity check
    if (utflen > 65536) throw IOErr.make("String too big").val;

    // write length as 2 byte value
    self.w((utflen >>> 8) & 0xFF);
    self.w((utflen >>> 0) & 0xFF);

    // write characters
    for (int i=0; i<slen; ++i)
    {
      int c = s.charAt(i);
      if (c <= 0x007F)
      {
        self.w(c);
      }
      else if (c > 0x07FF)
      {
        self.w(0xE0 | ((c >> 12) & 0x0F));
        self.w(0x80 | ((c >>  6) & 0x3F));
        self.w(0x80 | ((c >>  0) & 0x3F));
      }
      else
      {
        self.w(0xC0 | ((c >>  6) & 0x1F));
        self.w(0x80 | ((c >>  0) & 0x3F));
      }
    }
    return self;
  }

  public static OutStream print(OutStream self, Obj obj)
  {
    Str s = obj == null ? Str.Null : obj.toStr();
    self.writeChars(s, Int.Zero, s.size());
    return self;
  }

  public static OutStream printLine(OutStream self, Obj obj)
  {
    Str s = obj == null ? Str.Null : obj.toStr();
    self.writeChars(s, Int.Zero, s.size());
    self.writeChar(Int.pos['\n']);
    return self;
  }

  public static OutStream writeProps(OutStream self, Map props, Bool close)
  {
    Charset origCharset = self.charset();
    self.charset(Charset.utf8());
    try
    {
      List keys = props.keys().sort();
      int size = keys.sz();
      Int eq = Int.pos('=');
      Int nl = Int.pos('\n');
      for (int i=0; i<size; ++i)
      {
        Str key = (Str)keys.get(i);
        Str val = (Str)props.get(key);
        writePropStr(self, key.val);
        self.writeChar(eq);
        writePropStr(self, val.val);
        self.writeChar(nl);
      }
      return self;
    }
    finally
    {
      try { if (close.val) self.close(); } catch (Exception e) { e.printStackTrace(); }
      self.charset(origCharset);
    }
  }

  static void writePropStr(OutStream self, String s)
  {
    int len = s.length();
    for (int i=0; i<len; ++i)
    {
      int ch = s.charAt(i);
      int peek = i+1<len ? s.charAt(i+1) : -1;

      // escape special chars
      switch (ch)
      {
        case '\n': self.writeChar(Int.pos['\\']).writeChar(Int.pos['n']); continue;
        case '\r': self.writeChar(Int.pos['\\']).writeChar(Int.pos['r']); continue;
        case '\t': self.writeChar(Int.pos['\\']).writeChar(Int.pos['t']); continue;
        case '\\': self.writeChar(Int.pos['\\']).writeChar(Int.pos['\\']); continue;
      }

      // escape control chars, comments, and =
      if ((ch < ' ') || (ch == '/' && (peek == '/' || peek == '*')) || (ch == '='))
      {
        Int nib1 = Int.make((ch>>4)&0xf).toDigit(Int.pos[16]);
        Int nib2 = Int.make((ch>>0)&0xf).toDigit(Int.pos[16]);

        self.writeChar(Int.pos['\\']).writeChar(Int.pos['u'])
            .writeChar(Int.pos['0']).writeChar(Int.pos['0'])
            .writeChar(nib1).writeChar(nib2);
        continue;
      }

      // normal character
      self.writeChar(Int.pos(ch));
    }
  }

  public static void flush(OutStream self)
  {
  }

  public static Bool close(OutStream self)
  {
    return Bool.True;
  }

}