//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.*;
import fanx.compiler.*;
import fanx.compiler.parser.*;

/**
 * BinaryExpr is used for binary expressions with a left hand side
 * and a right hand side including ||, &&, ===, is, and as.  Note that
 * many common binary operations are actually modeled as ShortcutExpr
 * to enable method based operator overloading.
 */
public class BinaryExpr
  extends Expr
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public BinaryExpr(ILocation location, Token token, Expr lhs, Expr rhs)
  {
    super(location);
    this.opToken = token.type;
    switch (token.type)
    {
      case Token.ASSIGN:      id = ASSIGN;    break;  // =
      case Token.DOUBLE_AMP:  id = BOOL_AND;  break;  // &&
      case Token.DOUBLE_PIPE: id = BOOL_OR;   break;  // ||
      case Token.SAME:        id = SAME;      break;  // ===
      case Token.NOT_SAME:    id = NOT_SAME;  break;  // !==
      case Token.AS:          id = AS;        break;  // as
      case Token.IS:          id = IS;        break;  // is
      default: throw new IllegalStateException(token.toString());
    }
    this.lhs = lhs;
    this.rhs = rhs;
  }

  public BinaryExpr(ILocation location, Token token)
  {
    this(location, token, null, null);
  }

//////////////////////////////////////////////////////////////////////////
// Expr
//////////////////////////////////////////////////////////////////////////

  public int id()
  {
    return id;
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  protected void walkChildren(Visitor v)
  {
    lhs = walk(v, lhs);
    rhs = walk(v, rhs);
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public String toString()
  {
    return "(" + lhs + ") " + Token.toString(opToken) + " (" + rhs + ")";
  }

  public void tree(AstWriter out)
  {
    out.w("BinaryExpr " ).w(Token.toString(opToken)).w(" " + leaveToString()).nl();
    out.indent();
    out.w("lhs: "); lhs.tree(out);
    out.w("rhs: "); rhs.tree(out);
    out.unindent();
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public int id;       // expression id: OR, AND, SAME, ASSIGN, IS, AS
  public int opToken;  // operator token type (Token.AND, etc)
  public Expr lhs;     // left hand side
  public Expr rhs;     // right hand side

}