//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.compiler.ast;

import java.util.ArrayList;
import fan.sys.*;
import fanx.util.*;
import fanx.compiler.*;

/**
 * Expr is the base class for expressions
 */
public abstract class Expr
  extends Node
{

//////////////////////////////////////////////////////////////////////////
// Construction
//////////////////////////////////////////////////////////////////////////

  public Expr(ILocation location)
  {
    super(location);
  }

//////////////////////////////////////////////////////////////////////////
// Overrides
//////////////////////////////////////////////////////////////////////////

  /**
   * Get expression type id constant (defined below).
   */
  public abstract int id();

  /**
   * If this expression is a single variable name, then return
   * the string name, otherwise return null.
   */
  public String toSimpleName()
  {
    return null;
  }

  /**
   * Attempt to convert this expression into a TypeRef if this
   * expression could be a type declaration, otherwise return null.
   */
  public TypeRef toTypeRef()
  {
    return null;
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  /**
   * Return this expression as an Int literal usable in a tableswitch,
   * or null if this Expr doesn't represent a constant Int.  Expressions
   * which work as table switch cases:
   *   - IntLiteral
   *   - Enum
   */
  public Int asTableSwitchCase()
  {
    return null;
  }

  /**
   * Given a list of Expr instances, find the common base type
   * they all share.  This method does not take into account
   * the null literal.  It is used for type inference for lists
   * and maps.
   */
  public static Type commonType(ArrayList exprs)
  {
    ArrayList acc = new ArrayList(exprs.size());
    for (int i=0; i<exprs.size(); ++i)
    {
      Expr expr = (Expr)exprs.get(i);
      if (expr.type == null) throw new IllegalStateException();
      if (expr.id() != Expr.NULL_LITERAL) acc.add(expr.type);
    }
    return Type.common((Type[])acc.toArray(new Type[acc.size()]));
  }

//////////////////////////////////////////////////////////////////////////
// Tree
//////////////////////////////////////////////////////////////////////////

  public final Node walk(Visitor v)
  {
    v.enterExpr(this);
    walkChildren(v);
    v.exitExpr(this);
    return v.visit(this);
  }

  protected void walkChildren(Visitor v)
  {
  }

//////////////////////////////////////////////////////////////////////////
// Debug
//////////////////////////////////////////////////////////////////////////

  public abstract String toString();

  public String leaveToString() { return leave ? "leave" : "no-leave"; }

  public void tree(AstWriter out)
  {
    out.w(StrUtil.getName(getClass()) + " " + toString());
    if (type != null) out.w(" (" + type + ") " + leaveToString());
    out.nl();
  }

  public void pretty(AstWriter out)
  {
    out.w(toString());
  }

//////////////////////////////////////////////////////////////////////////
// Id Constants
//////////////////////////////////////////////////////////////////////////

  public static final int NULL_LITERAL     = 0;    // LiteralExpr
  public static final int BOOL_LITERAL     = 1;
  public static final int INT_LITERAL      = 2;
  public static final int FLOAT_LITERAL    = 3;
  public static final int STR_LITERAL      = 4;
  public static final int DURATION_LITERAL = 5;
  public static final int URI_LITERAL      = 6;
  public static final int TYPE_LITERAL     = 7;
  public static final int RANGE_LITERAL    = 8;    // RangeLiteralExpr
  public static final int LIST_LITERAL     = 9;    // ListLiteralExpr
  public static final int MAP_LITERAL      = 10;    // MapLiteralExpr
  public static final int BOOL_NOT         = 11;   // UnaryExpr
  public static final int NEGATE           = 12;
  public static final int POSITIVE         = 13;
  public static final int PRE_INCREMENT    = 14;
  public static final int PRE_DECREMENT    = 15;
  public static final int POST_INCREMENT   = 16;
  public static final int POST_DECREMENT   = 17;
  public static final int ASSIGN           = 20;   // BinaryExpr
  public static final int SAME             = 21;
  public static final int NOT_SAME         = 22;
  public static final int BOOL_OR          = 23;
  public static final int BOOL_AND         = 24;
  public static final int IS               = 25;
  public static final int AS               = 26;
  public static final int CALL             = 27;   // CallExpr
  public static final int NEW              = 28;   // NewExpr
  public static final int SHORTCUT         = 29;   // ShortcutExpr
  public static final int FIELD            = 30;   // FieldExpr
  public static final int LOCAL_VAR        = 31;   // LocalVarExpr
  public static final int THIS             = 32;   // ThisExpr
  public static final int SUPER            = 33;   // SuperExpr
  public static final int TYPE_VAR         = 34;   // TypeVarExpr
  public static final int UNKNOWN_VAR      = 35;   // UnknownVarExpr
  public static final int CAST             = 36;   // CastExpr
  public static final int CLOSURE          = 37;   // Closure
  public static final int TERNARY          = 38;   // TernaryExpr

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  public Type type;            // type the expression resolves to (and leaves on stack)
  public boolean leave = true; // should this expression leave a value on the stack
  public boolean err;          // is this expression in error

}