//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   5 Mar 06  Brian Frank  Creation
//
package fanx.compiler.steps;

import java.util.*;
import fan.sys.*;
import fanx.fcode.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.compiler.parser.*;

/**
 * ClosureVars (cvars) is used to pull local variables used by closures
 * into an auto-generated anonymous class so that they can be
 * used outside and inside the closure as well as have a lifetime
 * once the method returns:
 *   1) scan for methods with closures
 *   2) look for method vars flagged as usedInClosure
 *   3) define the cvars class
 *   4) reorder the method vars
 *   5) walk the method body
 *      a) remap local var access to cvars field access
 *      b) accumulate all the ClosureExprs
 *   6) walk accumlated ClosureExprs in method body
 *      a) add $cvars field to closure implementation class
 *      b) add $cvars parameter to implementation class constructor
 *      c) pass $cvars arg from method body to implementation constructor
 *      c) walk implementation class code and remap local var access
 */
public class ClosureVars
  extends CompilerStep
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public ClosureVars(Compiler compiler)
  {
    super(compiler);
  }

//////////////////////////////////////////////////////////////////////////
// Run
//////////////////////////////////////////////////////////////////////////

  public void run()
  {
    scan();
    addClassDefs(cvarsToAdd);
  }

  void scan()
  {
    for (int i=0; i<compiler.classDefs.length; ++i)
    {
      ClassDef cls = compiler.classDefs[i];
      MethodDef[] methods = cls.methods();
      for (int j=0; j<methods.length; ++j)
      {
        if (methods[j].hasLocalVarUsedInClosure())
          process(cls, methods[j]);
      }
    }
  }

//////////////////////////////////////////////////////////////////////////
// Process
//////////////////////////////////////////////////////////////////////////

  void process(ClassDef cls, MethodDef method)
  {
    this.loc      = cls.location;
    this.cls      = cls;
    this.method   = method;

    defineCvarsClass();
    reorderVars();
    insertCvarsInit();
    remapVarsInMethod();
    remapVarsInClosures();
  }

//////////////////////////////////////////////////////////////////////////
// Define Cvars Class
//////////////////////////////////////////////////////////////////////////

  void defineCvarsClass()
  {
    // define class
    String name = cls.name + "$" + method.name + "$Cvars";
    cvars = new ClassDef(loc, cls.unit, name, FConst.Internal | FConst.Synthetic);
    cvars.synthetic = true;
    cvars.type = new ShimType(compiler.pod, cvars);
    cvars.base = new TypeRef(loc, Sys.ObjType);

    // generate the fields
    MethodVar[] vars = (MethodVar[])method.vars.toArray(new MethodVar[method.vars.size()]);
    for (int i=0; i<vars.length; ++i)
    {
      MethodVar var = vars[i];
      if (var.usedInClosure)
      {
        FieldDef f = mapToField(var);
        cvars.slots.add(f);
        var.cvarsField = mapField(cvars.type, f);
      }
    }

    // generate the constructor
    MethodDef ctor = DefaultCtor.makeDefaultCtor(loc, FConst.Internal | FConst.Synthetic);
    cvars.slots.add(ctor);
    cvarsCtor = mapMethod(cvars.type, ctor);

    // restub the slot defs to shim slots
    cvars.type.restubSlots();

    // put into accumlator to add to compiler pipeline at end of step
    cvarsToAdd.add(cvars);
  }

  FieldDef mapToField(MethodVar var)
  {
    FieldDef f = new FieldDef(loc);
    f.name  = var.name;
    f.type  = new TypeRef(loc, var.type);
    f.flags = FConst.Internal | FConst.Synthetic;
    return f;
  }

//////////////////////////////////////////////////////////////////////////
// Reorder Vars
//////////////////////////////////////////////////////////////////////////

  void reorderVars()
  {
    // remove any non-parameter, locally defined variables
    // from the list which are to moved into the cvars class
    ArrayList vars = method.vars;
    MethodVar[] origVars = (MethodVar[])vars.toArray(new MethodVar[vars.size()]);
    for (int i=origVars.length-1; i>=0; --i)
      if (!origVars[i].isParam() && origVars[i].usedInClosure)
        vars.remove(i);

    // now insert the cvars (right after params so that we can
    // use optimized register access such as ILOAD_2 for Java)
    cvarsLocal = new MethodVar(-1, cvars.type, "$cvars", 0, null);
    vars.add(method.params.size(), cvarsLocal);

    // re-index the registers
    int reg = method.isStatic() ? 0 : 1;
    for (int i=0; i<vars.size(); ++i)
      ((MethodVar)vars.get(i)).register = reg++;
  }

//////////////////////////////////////////////////////////////////////////
// Insert Cvars Initialization
//////////////////////////////////////////////////////////////////////////

  void insertCvarsInit()
  {
    //  method(Foo x)
    //  {
    //    $cvars := $Cvars.make()
    //    $cvars.x = x // for all params
    //    ...
    //  }

    LocalVarExpr local = new LocalVarExpr(loc, cvarsLocal);
    local.type = cvars.type;

    CallExpr ctor = new CallExpr(loc, new TypeVarExpr(loc, cvars.type), cvarsCtor);
    ctor.type = cvars.type;

    BinaryExpr assign = new BinaryExpr(loc, new Token(Token.ASSIGN), local, ctor);
    assign.type = cvars.type;
    assign.leave = false;

    method.code.add(0, new ExprStmt(loc, assign));

    // init any params we are going to remap to cvars
    for (int i=0; i<method.vars.size(); ++i)
    {
      MethodVar var = (MethodVar)method.vars.get(i);
      if (!var.isParam()) break;
      if (var.cvarsField == null) continue;

      local = new LocalVarExpr(loc, var);
      local.noRemapToCvars = true; // don't want to replace this access with cvars field

      assign = new BinaryExpr(loc, new Token(Token.ASSIGN),
        new FieldExpr(loc, var.cvarsField, new LocalVarExpr(loc, cvarsLocal)),
        local);
      assign.leave = false;
      method.code.add(1, new ExprStmt(loc, assign));
    }
  }

//////////////////////////////////////////////////////////////////////////
// Remap Vars in Method
//////////////////////////////////////////////////////////////////////////

  void remapVarsInMethod()
  {
    closures.clear();
    method.code.walk(new Visitor()
    {
      public Node visit(Node node)
      {
        if (node instanceof LocalVarDefStmt) return remapInMethod((LocalVarDefStmt)node);
        if (node instanceof LocalVarExpr) return remapInMethod((LocalVarExpr)node);
        if (node instanceof ClosureExpr) closures.add(node);
        return node;
      }
    });
  }

  Node remapInMethod(LocalVarDefStmt local)
  {
    // Foo x;       ->  nop
    // Foo x := y;  ->  $cvars.x = y

    if (local.var.cvarsField == null) return local;
    if (local.init == null) return new NopStmt(local.location);

    ILocation loc = local.location;
    FieldExpr fieldExpr = new FieldExpr(loc, local.var.cvarsField, new LocalVarExpr(loc, cvarsLocal));
    BinaryExpr assign = new BinaryExpr(loc, new Token(Token.ASSIGN), fieldExpr, local.init);
    assign.leave = false;
    return new ExprStmt(loc, assign);
  }

  Node remapInMethod(LocalVarExpr local)
  {
    // x -> $cvars.x
    if (local.id() == Expr.THIS) return local;
    if (local.id() == Expr.SUPER) return local;
    if (local.var.cvarsField == null) return local;
    if (local.noRemapToCvars) return local;
    ILocation loc = local.location;
    return new FieldExpr(loc, local.var.cvarsField, new LocalVarExpr(loc, cvarsLocal));
  }

//////////////////////////////////////////////////////////////////////////
// Remap Vars in Closures
//////////////////////////////////////////////////////////////////////////

  void remapVarsInClosures()
  {
    // closures now contains all the ClosureExpr we found inside
    // the current method body, now we need to walk them
    for (int i=0; i<closures.size(); ++i)
      remapVarInClosure((ClosureExpr)closures.get(i));
  }

  void remapVarInClosure(ClosureExpr closure)
  {
    // add cvars field to closure implementation class
    ILocation loc = closure.location;
    FieldDef field = new FieldDef(loc);
    field.name  = "$cvars";
    field.type  = new TypeRef(loc, cvars.type);
    field.flags = FConst.Internal | FConst.Synthetic;

    // add to class and shim type
    closure.cls.slots.add(field);
    closureCvarsField = mapField(closure.cls.type, field);

    // add parameter to closure implementation constructor
    MethodDef ctor = closure.cls.method("$make");
    ctor.params.add(new MethodParamDef(loc, cvars.type, "$cvars"));
    MethodVar paramVar = new MethodVar(ctor.params.size(), cvars.type, "$cvars", FConst.Param, null);
    ctor.vars.add(paramVar);
    Method m = mapMethod(closure.cls.type, ctor);

    // restub the shim type
    closure.cls.type.restubSlots();

    // set field in constructor
    BinaryExpr assign = new BinaryExpr(loc, new Token(Token.ASSIGN),
      new FieldExpr(loc, closureCvarsField, new ThisExpr(loc)),
      new LocalVarExpr(loc, paramVar));
    assign.leave = false;
    ctor.code.add(0, new ExprStmt(loc, assign));

    // pass cvars instance to closure class constructor
    CallExpr ctorCall = closure.substitute;
    ctorCall.method = ctor.method;
    ctorCall.args.add(new LocalVarExpr(loc, cvarsLocal));

    // walk closure implementation
    MethodDef doCall = closure.cls.method("doCall");
    doCall.code.walk(new Visitor()
    {
      public Node visit(Node node)
      {
        if (node instanceof LocalVarExpr) return remapInClosure((LocalVarExpr)node);
        return node;
      }
    });
  }

  Node remapInClosure(LocalVarExpr local)
  {
    if (local.var == null || local.var.cvarsField == null) return local;
    ILocation loc = local.location;
    return new FieldExpr(loc, local.var.cvarsField, new FieldExpr(loc, closureCvarsField, new ThisExpr(loc)));
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  ArrayList cvarsToAdd = new ArrayList(64);
  ILocation loc;
  ClassDef cls;                // class containing closure
  MethodDef method;            // method containing 1 or more closures
  MethodVar cvarsLocal;        // local var referencing cvars in method body
  ClassDef cvars;              // cvars class implementation
  Method cvarsCtor;            // constructor for cvars class
  ArrayList closures = new ArrayList();  // acc for closures found in method body
  Field closureCvarsField;     // cvars field for current closure class implementation

}