//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   17 Apr 06  Brian Frank  Creation
//
package fanx.compiler.steps;

import java.util.*;
import fan.sys.*;
import fanx.fcode.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.compiler.parser.*;

/**
 * FieldAccessors walks through all the field definitions and
 * auto-generates MethodDefs for the getters and setters.
 */
public class FieldAccessors
  extends CompilerStep
{

  public FieldAccessors(Compiler compiler)
  {
    super(compiler);
  }

  public void run()
  {
    for (int i=0; i<compiler.classDefs.length; ++i)
      processClass(compiler.classDefs[i]);
  }

  void processClass(ClassDef cls)
  {
    if (cls.isEnum()) return; // TODO - enums should be able to have field accessors

    FieldDef[] fields = cls.fields();
    for (int i=0; i<fields.length; ++i)
      processField(cls, fields[i]);
  }

  void processField(ClassDef cls, FieldDef f)
  {
    // getter
    ILocation loc = f.get != null ? f.get.location : f.location;
    MethodDef get = new MethodDef(loc);
    get.accessorFor = f;
    get.flags = f.getFlags | FConst.Getter | FConst.Synthetic;
    get.name  = f.name;
    get.ret   = f.type;
    if (f.get != null) { get.code = f.get; f.get = null; }
    else
    {
      ReturnStmt ret = new ReturnStmt(loc);
      if (f.isStatic())
        ret.expr = new UnknownVarExpr(loc, new TypeVarExpr(loc, cls.type), f.name);
      else
        ret.expr = new UnknownVarExpr(loc, new ThisExpr(loc), f.name);
      get.code = new Block(loc);
      get.code.add(ret);
    }
    f.getMethod = get;
    cls.slots.add(get);

    // setter
    loc = f.set != null ? f.set.location : f.location;
    MethodDef set = new MethodDef(loc);
    set.accessorFor = f;
    set.flags = f.setFlags | FConst.Setter| FConst.Synthetic;
    set.name  = f.name;
    set.ret   = new TypeRef(loc, Sys.VoidType);
    set.params.add(new MethodParamDef(f.type, "val"));
    if (f.set != null) { set.code = f.set; f.set = null; }
    else
    {
      BinaryExpr assign = new BinaryExpr(loc, new Token(Token.ASSIGN));
      if (f.isStatic())
        assign.lhs = new UnknownVarExpr(loc, new TypeVarExpr(loc, cls.type), f.name);
      else
        assign.lhs = new UnknownVarExpr(loc, new ThisExpr(loc), f.name);
      assign.rhs = new UnknownVarExpr(loc, "val");
      set.code = new Block(loc);
      set.code.add(new ExprStmt(loc, assign));
    }
    f.setMethod = get;
    cls.slots.add(set);
  }

}