//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Dec 05  Brian Frank  Creation
//
package fanx.compiler.steps;

import java.io.*;
import java.util.*;
import fan.sys.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.compiler.parser.*;
import fanx.compiler.resolver.*;
import fanx.fcode.*;
import fanx.util.Stack;

/**
 * Walk the AST to resolve:
 *   - Resolve all the Expr nodes
 *   - Resolve all Stmt nodes
 *   - Resolve all MethodAlias
 */
public class ResolveExpr
  extends VisitorStep
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public ResolveExpr(Compiler compiler)
  {
    super(compiler);
  }

//////////////////////////////////////////////////////////////////////////
// Visitor Callbacks
//////////////////////////////////////////////////////////////////////////

  public void run()
  {
    super.run();
    bombIfErr();
  }

  public Node visit(Node node)
  {
    if (node instanceof Expr) return resolveExpr((Expr)node);
    if (node instanceof Stmt) return resolveStmt((Stmt)node);
    return node;
  }

  public void enterClassDef(ClassDef d)
  {
    super.enterClassDef(d);
  }

  public void exitClassDef(ClassDef d)
  {
    super.exitClassDef(d);
  }

  public void enterMethodDef(MethodDef m)
  {
    super.enterMethodDef(m);
    initMethodVars();
  }

  public void exitMethodDef(MethodDef m)
  {
    super.exitMethodDef(m);
  }

//////////////////////////////////////////////////////////////////////////
// Statements
//////////////////////////////////////////////////////////////////////////

  /**
   * Resolve Stmt.
   */
  Node resolveStmt(Stmt stmt)
  {
    switch (stmt.id())
    {
      case Stmt.EXPR:          resolveExpr((ExprStmt)stmt); break;
      case Stmt.FOR:           resolveFor((ForStmt)stmt); break;
      case Stmt.BREAK:         resolveBreak((BreakStmt)stmt); break;
      case Stmt.CONTINUE:      resolveContinue((ContinueStmt)stmt); break;
      case Stmt.LOCAL_VAR_DEF: resolveLocalVarDef((LocalVarDefStmt)stmt); break;
    }
    return stmt;
  }

  void resolveExpr(ExprStmt stmt)
  {
    // stand alone expr statements, shouldn't be left on the stack
    stmt.expr.leave = false;
  }

  void resolveLocalVarDef(LocalVarDefStmt def)
  {
    // check for type inference
    if (def.type == null)
      def.type = new TypeRef(def.location, def.init.type);

    // bind to scope as a method variable
    def.var = bindToMethodVar(def);
  }

  void resolveFor(ForStmt stmt)
  {
    // don't leave update expression on the stack
    if (stmt.update != null) stmt.update.leave = false;
  }

  void resolveBreak(BreakStmt stmt)
  {
    // find which loop we're inside of (checked in CheckErrors)
    stmt.loop = stmtStack.findLoop();
  }

  void resolveContinue(ContinueStmt stmt)
  {
    // find which loop we're inside of (checked in CheckErrors)
    stmt.loop = stmtStack.findLoop();
  }

//////////////////////////////////////////////////////////////////////////
// Expressions
//////////////////////////////////////////////////////////////////////////

  /**
   * Ensure a Expr's type field maps to a valid Type.
   */
  Expr resolveExpr(Expr expr)
  {
    // if an expression underneath this guy has already
    // been found in error, then just skip it
    if (expr.err) return expr;

    try
    {
      // do the actual resolve
      expr = doResolveExpr(expr);
    }
    catch (CompileException e)
    {
      // error out this guy and everybody above him in the tree
      exprStack.err();
    }

    return expr;
  }

  /**
   * Ensure a Expr's type field maps to a valid Type.
   */
  private Expr doResolveExpr(Expr expr)
  {
    switch (expr.id())
    {
      case Expr.NULL_LITERAL:     expr.type = Sys.ObjType;       break;
      case Expr.BOOL_LITERAL:     expr.type = Sys.BoolType;      break;
      case Expr.INT_LITERAL:      expr.type = Sys.IntType;       break;
      case Expr.FLOAT_LITERAL:    expr.type = Sys.FloatType;      break;
      case Expr.STR_LITERAL:      expr.type = Sys.StrType;       break;
      case Expr.DURATION_LITERAL: expr.type = Sys.DurationType;  break;
      case Expr.URI_LITERAL:      expr.type = Sys.UriType;       break;
      case Expr.TYPE_LITERAL:     expr.type = Sys.TypeType;      break;
      case Expr.RANGE_LITERAL:    expr.type = Sys.RangeType;     break;
      case Expr.LIST_LITERAL:     expr = resolveList((ListLiteralExpr)expr);  break;
      case Expr.MAP_LITERAL:      expr = resolveMap((MapLiteralExpr)expr);  break;
      case Expr.BOOL_NOT:         expr.type = Sys.BoolType; break;
      case Expr.NEGATE:
      case Expr.POSITIVE:
      case Expr.PRE_INCREMENT:
      case Expr.PRE_DECREMENT:
      case Expr.POST_INCREMENT:
      case Expr.POST_DECREMENT:   expr.type = ((UnaryExpr)expr).operand.type; break;
      case Expr.ASSIGN:           expr = resolveAssign((BinaryExpr)expr); break;
      case Expr.SAME:
      case Expr.NOT_SAME:         expr.type = Sys.BoolType; break;
      case Expr.BOOL_OR:
      case Expr.BOOL_AND:
      case Expr.IS:               expr.type = Sys.BoolType; break;
      case Expr.AS:               expr.type = ((BinaryExpr)expr).rhs.type; break;
      case Expr.CALL:
      case Expr.NEW:              expr = resolveCall((CallExpr)expr); break;
      case Expr.SHORTCUT:         expr = resolveShortcut((ShortcutExpr)expr); break;
      case Expr.THIS:             expr = resolveThis((ThisExpr)expr); break;
      case Expr.SUPER:            expr = resolveSuper((SuperExpr)expr); break;
      case Expr.TYPE_VAR:         resolveTypeVar((TypeVarExpr)expr); break;
      case Expr.UNKNOWN_VAR:      expr = resolveVar((UnknownVarExpr)expr); break;
      case Expr.CAST:             expr.type = ((CastExpr)expr).toType.type; break;
      case Expr.CLOSURE:          resolveClosure((ClosureExpr)expr); break;
      case Expr.TERNARY:          resolveTernary((TernaryExpr)expr); break;
    }

    // sanity check
    if (expr.type == null)
      throw new IllegalStateException(expr.getClass().getName() + ": " + expr);

    return expr;
  }

  /**
   * Resolve list literal
   */
  Expr resolveList(ListLiteralExpr expr)
  {
    Type of;
    if (expr.explicitType != null)
    {
      of = expr.explicitType.type;
    }
    else
    {
      // infer from list item expressions
      of = Expr.commonType(expr.items);
    }
    expr.type = of.toListOf();
    return expr;
  }

  /**
   * Resolve map literal
   */
  Expr resolveMap(MapLiteralExpr expr)
  {
    if (expr.mapType != null)
    {
      expr.type = expr.mapType.type;
    }
    else
    {
      // interface from key/val expressions
      Type k = Expr.commonType(expr.keys);
      Type v = Expr.commonType(expr.vals);
      expr.type = new MapType(k, v);
    }
    return expr;
  }

  /**
   * Resolve an assignment operation and insert casts if necessary (=, *=, /=, %=, +=, -=, |=, &=, ^=).
   */
  Expr resolveAssign(BinaryExpr expr)
  {
    // check auto-cast (TODO - this shouldn't be needed anymore with TypeChecker????)
//    if (expr.lhs.type != expr.rhs.type)
//      autoCast(expr, true);

    // check for left hand side the [] shortcut, because []= is set
    if (expr.lhs instanceof ShortcutExpr)
    {
      ShortcutExpr shortcut = (ShortcutExpr)expr.lhs;
      if (shortcut.op == ShortcutExpr.GET)
      {
        shortcut.op = ShortcutExpr.SET;
        shortcut.name = "set";
        shortcut.args.add(expr.rhs);
        shortcut.method = null;
        resolveCall(shortcut);
        return shortcut;
      }
    }

    // check for left hand side the -> shortcut, because a->x=b is trap.a("x", [b])
    if (expr.lhs instanceof CallExpr && ((CallExpr)expr.lhs).isDynamic)
    {
      CallExpr call = (CallExpr)expr.lhs;
      call.args.add(expr.rhs);
      return resolveCall(call);
    }

    // type check
    Type lhs = expr.lhs.type;
    Type rhs = expr.rhs.type;
    /* TODO done in CheckTypes
    if (expr.id() == Expr.ASSIGN)
    {
      if (!rhs.is(lhs))
        throw err("'" + rhs + "' is not assignable to '" + lhs + "'", expr);
    }
    else
    {
      if (!lhs.isNumeric() || lhs != rhs)
        throw err("Cannot apply '" + Token.toString(expr.opToken) + "' operator to '" + lhs + "' and '" + lhs + "'", expr);
    }
    */
    expr.type = rhs;

    return expr;
  }

  /**
   * Resolve the this pointer.
   */
  Expr resolveThis(ThisExpr expr)
  {
    if (curClassDef.isClosure() && curMethod.name.equals("doCall"))
    {
      ClosureExpr closure = curClassDef.closure;
      return new FieldExpr(expr, closure.outerThisField(), new ThisExpr(expr));
    }

    expr.type = curType;
    return expr;
  }

  /**
   * Resolve the super pointer.
   */
  Expr resolveSuper(SuperExpr expr)
  {
    if (curClassDef.isClosure() && curMethod.name.equals("doCall"))
    {
      ClosureExpr closure = curClassDef.closure;
      FieldExpr f = new FieldExpr(expr, closure.outerThisField(), new ThisExpr(expr));
      f.type = f.type.base();
      return f;
    }

    expr.type = curType.base();
    return expr;
  }

  /**
   * Resolve a call to it's Method and return type.
   */
  Expr resolveCall(CallExpr call)
  {
    if (call.isDynamic)
    {
      call.method = Sys.ObjType.method("trap", true);
      call.type = Sys.ObjType;
    }
    else
    {
      // use CallResolver helper class
      new CallResolver(this, curClassDef, curMethod, call).resolve();
    }

    // if non-static/ctor call on this, insert implicit "this" node
    insertImplicitThis(call);

    return call;
  }

  /**
   * If the call has not explicit target, and is a normal
   * instance method, then we need to insert an implicit this.
   *
   * OLD:  We also use this opportunity to check if we are calling
   * and Obj method on a mixin - although we fake this out
   * from the Fan side of things, mixin interfaces do not
   * actually extend Obj in Java/C# - therefore we need to
   * insert an explicit cast to make the VM happy (this is the
   * "this" side of what we do for the arguments in CallResolver)
   * NEW: the above has been replaced by just using java.lang.Object
   * where ever we wouldn't used Obj
   */
  private void insertImplicitThis(CallExpr call)
  {
    if (!call.method.isStatic().val && !call.method.isCtor().val)
    {
      if (!inStatic && call.target == null)
      {
        call.target = new ThisExpr(call.location);
        call.target.type = curType;
      }

      /* don't do this anymore, we use java.lang.Object for Obj now
      if (call.target != null && call.target.type.isMixin().val && call.method.parent() == Sys.ObjType)
        call.target = new CastExpr(Sys.ObjType, call.target);
      */
    }
  }

  /**
   * Resolve an UnknownVar to it's replacement node.
   */
  Expr resolveVar(UnknownVarExpr var)
  {
    // check for type literal: Foo.type
    if (var.target instanceof TypeVarExpr && var.name.equals("type"))
    {
      LiteralExpr literal = new LiteralExpr(var.location, Expr.TYPE_LITERAL, var.target.type);
      literal.type = Sys.TypeType;
      return literal;
    }

    // if there is no target
    Slot slot = null;
    if (var.target == null)
    {
      // attempt to a name in the current scope
      MethodVar binding = resolveLocal(var.name);
      if (binding != null)
        return new LocalVarExpr(var.location, binding);

      // if a closure, check parent scope; closure auto-generated classes
      // must always be processed after their original declaration classes,
      // so we can guarantee that the enclosingLocals have been resolved
      if (curClassDef.isClosure())
      {
        // TODO: this code doesn't handle nested closures
        ClosureExpr closure = curClassDef.closure;
        binding = closure.resolveLocal(var.name);
        if (binding != null)
        {
          // for now just mark the local var as being used in a closure,
          // and return a LocalVarExpr; we will pull this out into a cvars
          // class in the ClosureVars step
          binding.usedInClosure = true;
          return new LocalVarExpr(var.location, binding);
        }

        // check if we are in a closure - they get implicit access
        slot = closure.enclosingClass.type.slot(var.name, false);
        if (slot != null && !slot.isStatic().val)
          var.target = new FieldExpr(var, closure.outerThisField(), new ThisExpr(var));
      }
    }

    // at this point it can't be a local variable, so it must be
    // a slot on either myself or the variable's target
    Type onType;
    if (var.target == null)
      onType = curType;
    else
      onType = var.target.type;
    if (onType == null) throw new IllegalStateException();

    // try to find the slot on the target type
    if (slot == null)
      slot = onType.slot(var.name, false);

    if (slot == null)
    {
      // if there was an explicit target, then give up
      if (var.target == null)
        throw err("Unknown variable '" + var.name + "'", var);
      else
        throw err("Unknown slot '" + onType + "." + var.name + "'", var);
    }

    // no arg method
    if (slot instanceof Method)
    {
      CallExpr call = new CallExpr(var, var.target, (Method)slot);
      new CallResolver(this, curClassDef, curMethod, call).resolve();
      insertImplicitThis(call);
      return call;
    }

    // field
    else
    {
      if (slot.isStatic().val)
      {
        return new FieldExpr(var, (Field)slot, null);
      }
      else
      {
        if (var.target == null && !inStatic) var.target = new ThisExpr(var.location);
        return new FieldExpr(var, (Field)slot, var.target);
      }
    }
  }

  /**
   * Resolve TypeVarExpr.
   */
  void resolveTypeVar(TypeVarExpr var)
  {
    if (var.type == null)
      var.type = resolveType(var, var.typeRef);
    if (var.type == null)
      throw err("Unresolved type '" + var + "'", var);
  }

  /**
   * Resolve a ?: ternary expression
   */
  Expr resolveTernary(TernaryExpr expr)
  {
    if (expr.trueExpr.id() == Expr.NULL_LITERAL)
      expr.type = expr.falseExpr.type;
    else if (expr.falseExpr.id() == Expr.NULL_LITERAL)
      expr.type = expr.trueExpr.type;
    else
      expr.type = Type.common(new Type[] {expr.trueExpr.type, expr.falseExpr.type});
    return expr;
  }

//////////////////////////////////////////////////////////////////////////
// Shortcuts
//////////////////////////////////////////////////////////////////////////

  /**
   * Resolve ShortcutExpr.
   */
  Expr resolveShortcut(ShortcutExpr expr)
  {
    // if the target of a "slice" is a TypeVarExpr, then this isn't
    // a slice, but a List literal with one item: Str["a"]
    if (expr.op == ShortcutExpr.GET && expr.target instanceof TypeVarExpr)
    {
      TypeVarExpr type = (TypeVarExpr)expr.target;
      ListLiteralExpr list = new ListLiteralExpr(expr, type.toTypeRef());
      list.items.add(expr.args.get(0));
      return resolveList(list);
    }

    // if a binary operation
    if (expr.args.size() == 1)
    {
      // extract lhs and rhs
      Expr lhs = expr.target;
      Expr rhs = (Expr)expr.args.get(0);

      // if arg is Range, then slice(), not get()
      if (expr.op == ShortcutExpr.GET && rhs.type == Sys.RangeType)
      {
        expr.op = ShortcutExpr.SLICE;
        expr.name = "slice";
      }

      // string concat is always optimized, and performs a bit
      // difference since a non-string can be used as the lhs
      if (expr.opToken == Token.PLUS && (lhs.type.isStr() || rhs.type.isStr()))
      {
        expr.type = Sys.StrType;
        return expr;
      }

      // insert auto casts if necessary
      //autoCast(expr, false);
    }

    // resolve the call (we optimize this in the assembler)
    resolveCall(expr);

    // the comparision operations are special case that call a method
    // that return an int, but leave a boolean on the stack
    switch (expr.opToken)
    {
      case Token.LT:  case Token.LTEQ:
      case Token.GT:  case Token.GTEQ:
        expr.type = Sys.BoolType;
    }

    return expr;
  }

  /**
   * If the left and right hand sides are both numeric, but not at the same precision
   * level, then insert a cast operation to the most precise numeric type.  If rightOnly
   * is true, then don't insert any casts on the left hand side.
   */
  void autoCast(ShortcutExpr expr, boolean rightOnly)
  {
    // only process binary operations
    if (expr.args.size() != 1) return;

    // extract lhs and rhs
/* TODO
    Expr lhsExpr = expr.target;
    Expr rhsExpr = (Expr)expr.args.get(0);
    int lhs = lhsExpr.type.builtin;
    int rhs = rhsExpr.type.builtin;

    switch (lhs)
    {
      case Type.INT:
        if (!rightOnly)
        {
          if (rhs == Type.FLOAT) { expr.target = new CastExpr(lib.Float, lhsExpr); return; }
        }
        break;
      case Type.FLOAT:
        if (rhs == Type.INT)    { expr.args.set(0, new CastExpr(lib.Float, rhsExpr)); return; }
        break;
    }
*/
  }

//////////////////////////////////////////////////////////////////////////
// Closure
//////////////////////////////////////////////////////////////////////////

  /**
   * ClosureExpr will just output it's substitute expression.  But we take
   * this opportunity to capture the local variables in the closure's scope
   * and cache them on the ClosureExpr.
   */
  private void resolveClosure(ClosureExpr expr)
  {
    expr.type = expr.substitute.type;
    expr.enclosingLocals = getLocalsInScope();
  }

//////////////////////////////////////////////////////////////////////////
// Scope
//////////////////////////////////////////////////////////////////////////

  /**
   * Setup the MethodVars for the parameters.
   */
  private void initMethodVars()
  {
    MethodDef m = curMethod;
    int reg = m.isStatic() ?  0 : 1;

    // params (use null for block scope to indicate scoped to entire method)
    for (int i=0; i<m.params.size(); ++i)
    {
      MethodParamDef p = (MethodParamDef)m.params.get(i);
      MethodVar var = new MethodVar(reg++, p.type.type, p.name, FConst.Param, null);
      var.paramDef = p;
      m.vars.add(var);
    }
  }

  /**
   * Bind the specified local variable definition to a
   * MethodVar (and register number).
   */
  private MethodVar bindToMethodVar(LocalVarDefStmt def)
  {
    // make sure it doesn't exist in the current scope
    if (resolveLocal(def.name) != null)
      err("Variable '" + def.name + "' is already defined in current block", def);

    // alloc register and add MethodDef (maybe this + number of parameters)
    int reg = curMethod.vars.size();
    if (!inStatic) reg++;
    MethodVar var = new MethodVar(reg, def.type.type, def.name, 0, currentBlock());
    curMethod.vars.add(var);
    return var;
  }

  /**
   * Resolve a local variable for using current block stack.
   */
  private MethodVar resolveLocal(String name)
  {
    if (curMethod == null) return null;
    ArrayList vars = curMethod.vars;
    for (int i=0; i<vars.size(); ++i)
    {
      MethodVar var = (MethodVar)vars.get(i);
      if (var.name.equals(name) && isBlockInScope(var.scope))
        return var;
    }
    return null;
  }

  /**
   * Get a list of all the local method variables that
   * are currently in scope.
   */
  private ArrayList getLocalsInScope()
  {
    ArrayList vars = curMethod.vars;
    ArrayList acc = new ArrayList(vars.size());
    for (int i=0; i<vars.size(); ++i)
    {
      MethodVar var = (MethodVar)vars.get(i);
      if (isBlockInScope(var.scope)) acc.add(var);
    }
    return acc;
  }

  /**
   * Get the current block which defines our scope.  We make
   * a special case for "for" loops which can declare variables.
   */
  private Block currentBlock()
  {
    if (stmtStack.peek() instanceof ForStmt)
      return ((ForStmt)stmtStack.peek()).block;
    else
      return (Block)blockStack.peek();
  }

  /**
   * Check if the specified block is currently in scope.  We make
   * a specialcase for "for" loops which can declare variables.
   */
  private boolean isBlockInScope(Block block)
  {
    // the null block within the whole method (ctorChains or defaultParams)
    if (block == null) return true;

    // special case for "for" loops
    if (stmtStack.peek() instanceof ForStmt)
    {
      if (((ForStmt)stmtStack.peek()).block == block)
        return true;
    }

    for (int i=0; i<blockStack.n; ++i)
      if (block == blockStack.stack[i]) return true;

    return false;
  }

//////////////////////////////////////////////////////////////////////////
// StmtStack
//////////////////////////////////////////////////////////////////////////

  public void enterStmt(Stmt stmt) { stmtStack.push(stmt); }
  public void exitStmt(Stmt stmt)  { stmtStack.pop(); }

  static class StmtStack extends Stack
  {
    Stmt findLoop()
    {
      for (int i=n-1; i>=0; --i)
      {
        Stmt stmt = (Stmt)stack[i];
        if (stmt.id() == Stmt.WHILE) return stmt;
        if (stmt.id() == Stmt.FOR)   return stmt;
      }
      return null;
    }
  }

//////////////////////////////////////////////////////////////////////////
// BlockStack
//////////////////////////////////////////////////////////////////////////

  public void enterBlock(Block block) { blockStack.push(block); }
  public void exitBlock(Block block)  { blockStack.pop(); }

  static class BlockStack extends Stack
  {
  }

//////////////////////////////////////////////////////////////////////////
// ExprStack
//////////////////////////////////////////////////////////////////////////

  public void enterExpr(Expr expr) { exprStack.push(expr); }
  public void exitExpr(Expr expr)  { exprStack.pop(); }

  static class ExprStack extends Stack
  {
    void err()
    {
      for (int i=0; i<n; ++i)
        ((Expr)stack[i]).err = true;
    }
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  StmtStack  stmtStack  = new StmtStack();
  BlockStack blockStack = new BlockStack ();
  ExprStack  exprStack  = new ExprStack();

}