//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Dec 05  Brian Frank  Creation
//
package fanx.compiler.steps;

import java.io.*;
import java.util.*;
import fan.sys.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;

/**
 * Walk the AST to resolve TypeRefs into a Type:
 *  - Resolve all the TypeRef nodes
 *  - Resolve all UnknownVarExpr nodes which map to TypeVarExprs
 */
public class ResolveTypeRefs
  extends VisitorStep
{

  /**
   * Constructor
   */
  public ResolveTypeRefs(Compiler compiler)
  {
    super(compiler);
  }

  /**
   * Run
   */
  public void run()
  {
    super.run();
    resolveInheritance();
  }

  /**
   * Visit nodes which are (or might be) a TypeRef.
   */
  public Node visit(Node node)
  {
    if (node instanceof TypeRef) resolveType((TypeRef)node);
    else if (node instanceof UnknownVarExpr) return resolveTypeVar((UnknownVarExpr)node);
    return node;
  }

  /**
   * Ensure a TypeRef's type field maps to a valid Type.
   */
  void resolveType(TypeRef ref)
  {
    // if already resolved, then we are done
    if (ref.type != null) return;

    // handle parameterized List
    if (ref instanceof ListTypeRef)
    {
      ref.type = ((ListTypeRef)ref).of.type.toListOf();
      return;
    }

    // handle parameterized Map
    if (ref instanceof MapTypeRef)
    {
      ref.type = ((MapTypeRef)ref).toMapType();
      return;
    }

    // handle parameterized Method
    if (ref instanceof MethodTypeRef)
    {
      ref.type = ((MethodTypeRef)ref).toMethodType();
      return;
    }

    // resolve based on name
    ref.type = resolveType(ref, ref);
    if (ref.type != null) return;

    // if compiling sys, V and K are special parameterized types
    if (compiler.isSys)
    {
      ref.type = Sys.genericParameterType(ref.id);
      if (ref.type != null) return;
    }

    // no luck
    throw err("Unresolved type '" + ref.id + "'", ref);
  }

  /**
   * If an UnknownVarExpr maps to a Type, replace it with a TypeVarExpr
   */
  Node resolveTypeVar(UnknownVarExpr var)
  {
    // switch to pod::type makes this obsolete
    /*
    // collapse dotted names in case this is fully qualified
    // type name like "java.lang.Integer" (first we will try
    // "java", then "java.lang", then "java.lang.Integer" until
    // we get a hit)
    String name = var.name;
    UnknownVarExpr base = var;
    while (base.target instanceof UnknownVarExpr)
    {
      base = (UnknownVarExpr)base.target;
      name = base.name + "." + name;
    }
    */

    String name = var.name;
    UnknownVarExpr base = var;

    // if the var target is not null, then this variable is inside
    // a chain of expressions and it can't ever be a TypeVar
    if (base.target != null) return var;

    // see if name maps to a type name (to scope a static member), if
    // so then replace the entire variable chain with a TypeVarExpr
    Type type = resolveType(base, new TypeRef(var.location, name));
    if (type != null)
    {
      return new TypeVarExpr(base, type);
    }

    // just leave it and we will try again in a later pass
    return var;
  }

  /**
   * Resolve inheritance
   */
  private void resolveInheritance()
  {
    for (int i=0; i<compiler.classDefs.length; ++i)
      resovleInheritance(compiler.classDefs[i]);
  }

  private void resovleInheritance(ClassDef cls)
  {
     // if base already specified, this was a hand coded cls
     if (cls.base != null) return;

     // if no base types defined
     if (cls.bases.size() == 0)
     {
       if (!cls.qname.equals("sys::Obj"))
         cls.base = new TypeRef(cls.location, Sys.ObjType);
       return;
     }

     // check if base class specified
     TypeRef first = (TypeRef)cls.bases.get(0);
     if (first.type.isMixin().val)
     {
       cls.base = new TypeRef(cls.location, Sys.ObjType);
     }
     else
     {
       cls.base = first;
       cls.bases.remove(0);
     }

     // everything else is mixin
     cls.mixins = cls.bases;
     for (int i=0; i<cls.mixins.size(); ++i)
     {
       TypeRef t = (TypeRef)cls.mixins.get(i);
       if (!t.type.isMixin().val)
         err("Invalid extends", t.location);
     }
  }

    /*
    if (curt == Token.EXTENDS)
    {
      if (mixin)
      {
        cls.base = new TypeRef(loc, Sys.ObjType);
        consume();
        cls.mixins.add( type() );
        while (curt == Token.COMMA)
        {
          consume();
          cls.mixins.add( type() );
        }
      }
      else
      {
        if (isEnum) throw err("An enum can only implement mixins (no extends)", cur);
        consume();
        cls.base = type();
      }
    }
    else
    {
      if (isEnum)
        cls.base = new TypeRef(loc, Sys.EnumType);
      else if (!cls.qname.equals("sys::Obj"))
        cls.base = new TypeRef(loc, Sys.ObjType);
    }

    // mixins
    if (curt == Token.MIXIN)
    {
      if (mixin) throw err("A mixin can only extend other mixins", cur);
      consume();
      cls.mixins.add( type() );
      while (curt == Token.COMMA)
      {
        consume();
        cls.mixins.add( type() );
      }
    }
    */


}