//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Dec 05  Brian Frank  Creation
//
package fanx.compiler.steps;

import java.io.*;
import java.util.*;
import fan.sys.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.compiler.resolver.*;

/**
 * VisitorStep is an abstract base class for CompilerSteps which
 * walk the abstract syntax tree using the Visitor API.
 */
public abstract class VisitorStep
  extends CompilerStep
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public VisitorStep(Compiler compiler)
  {
    super(compiler);
  }

//////////////////////////////////////////////////////////////////////////
// Run
//////////////////////////////////////////////////////////////////////////

  /**
   * Walk all the ClassDefs with the a Visitor that maps to our callbacks.
   */
  public void run()
  {
    // init convenience fields
    this.classDefs = compiler.classDefs;

    // initialize a type checker, since most walkers need these utils
    // TODO, refactor into NameSpace?
    this.typeChecker = new TypeChecker(compiler);

    // default behavior is to just walk and let subclasses override visit()
    walk();
  }

  /**
   * Walk the all class defs.
   */
  public void walk()
  {
    for (int i=0; i<classDefs.length; ++i)
    {
      curClassDef = classDefs[i];
      curUnit     = curClassDef.unit;
      curType     = curClassDef.type;
      walk(classDefs[i]);
    }
  }

  /**
   * Walk the specified node (note this doesn't setup all the fields)
   */
  public void walk(Node node)
  {
    node.walk(new Visitor()
    {
      public Node visit(Node n) { return VisitorStep.this.visit(n); }
      public void enterClassDef(ClassDef d)   { VisitorStep.this.enterClassDef(d); }
      public void exitClassDef(ClassDef d)    { VisitorStep.this.exitClassDef(d); }
      public void enterFieldDef(FieldDef d) { VisitorStep.this.enterFieldDef(d); }
      public void exitFieldDef(FieldDef d)  { VisitorStep.this.exitFieldDef(d); }
      public void enterMethodDef(MethodDef d) { VisitorStep.this.enterMethodDef(d); }
      public void exitMethodDef(MethodDef d)  { VisitorStep.this.exitMethodDef(d); }
      public void enterStmt(Stmt s) { VisitorStep.this.enterStmt(s); }
      public void exitStmt(Stmt s)  { VisitorStep.this.exitStmt(s); }
      public void enterBlock(Block b) { VisitorStep.this.enterBlock(b); }
      public void exitBlock(Block b)  { VisitorStep.this.exitBlock(b); }
      public void enterExpr(Expr e) { VisitorStep.this.enterExpr(e); }
      public void exitExpr(Expr e)  { VisitorStep.this.exitExpr(e); }
    });
  }

  /**
   * Used to initialize an embedded VistorStep
   */
  public void init(ClassDef cls)
  {
    this.classDefs        = compiler.classDefs;
    this.typeChecker      = new TypeChecker(compiler);
    this.curUnit          = cls.unit;
    this.curClassDef      = cls;
    this.curType          = cls.type;
    enterClassDef(cls);
  }

//////////////////////////////////////////////////////////////////////////
// Callbacks
//////////////////////////////////////////////////////////////////////////

  public abstract Node visit(Node node);

  public void enterClassDef(ClassDef def)
  {
  }

  public void exitClassDef(ClassDef m)
  {
  }

  public void enterFieldDef(FieldDef f)
  {
    curField = f;
    inStatic = f.isStatic();
  }

  public void exitFieldDef(FieldDef f)
  {
    curField = null;
  }

  public void enterMethodDef(MethodDef m)
  {
    curMethod = m;
    inStatic  = m.isStatic();
  }

  public void exitMethodDef(MethodDef m)
  {
    curMethod = null;
  }

  public void enterStmt(Stmt stmt) {}
  public void exitStmt(Stmt stmt)  {}

  public void enterBlock(Block block) {}
  public void exitBlock(Block block)  {}

  public void enterExpr(Expr expr) {}
  public void exitExpr(Expr expr)  {}

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  /**
   * Convenience for resolver.resolveType(loc, this.curUnit, ref).
   */
  public Type resolveType(ILocation loc, TypeRef ref)
  {
    return compiler.ns.resolveType(this.curUnit, loc, ref);
  }

  /**
   * Convenience for typeChecker.is(expected, expr).
   */
  public boolean is(Expr expr, Type expected)
  {
    return typeChecker.is(expr, expected);
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  //
  // NOTE: new fields must be mapped in embedded constructor
  //

  ClassDef[] classDefs;               // convenience: compiler.classDefs
  TypeChecker typeChecker;            // provides type checking utilities
  CompilationUnit curUnit;            // current for run()
  ClassDef curClassDef;               // current for run()
  ShimType curType;                   // current for run()
  FieldDef curField;                  // enterFielddDef to exitFieldDef
  MethodDef curMethod;                // enterMethodDef to exitMethodDef
  boolean inStatic;                   // curMethod.flags.isStatic()

}