//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   3 Sep 05  Brian Frank  Creation
//
package fanx.compiler.steps;

import java.io.*;
import java.util.*;
import java.util.zip.*;
import fanx.fcode.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;
import fanx.compiler.ast.*;
import fanx.util.*;

/**
 * WritePod writes the FPod to a zip file.
 */
public class WritePod
  extends CompilerStep
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  public WritePod(Compiler compiler)
  {
    super(compiler);
  }

//////////////////////////////////////////////////////////////////////////
// Run
//////////////////////////////////////////////////////////////////////////

  public void run()
  {
    File dir = compiler.outDir;
    mkdir(dir);

    FPod pod = compiler.fpod;
    this.podFile = new File(compiler.outDir, pod.podName+".pod");

    log("Zip", podFile);

    FStore store = null;
    try
    {
      // open zip store
      store = FStore.openToWrite(new BufferedOutputStream(new FileOutputStream(podFile)));

      // write fpod data structures into zip file
      pod.store = store;
      pod.write();

      // write resource files
      writeRes(store);

      // if including fandoc write it out too
      if (compiler.includeDoc) writeDoc(store);

      // if including source write it out too
      if (compiler.includeSrc) writeSrc(store);
    }
    catch (CompileException e)
    {
      throw e;
    }
    catch (Exception e)
    {
      throw err("Cannot write", podFile, e);
    }
    finally
    {
      if (store != null) { try { store.close(); } catch (Exception e) {} }
    }
  }

//////////////////////////////////////////////////////////////////////////
// Resource
//////////////////////////////////////////////////////////////////////////

  private void writeRes(FStore store)
    throws IOException
  {
    for (int i=0; i<compiler.resFiles.length; ++i)
      writeRes(store, compiler.resFiles[i]);
  }

  private void writeRes(FStore store, File file)
    throws IOException
  {
    String path = file.getPath().substring(compiler.srcDir.getPath().length()+1);
    path = path.replace('\\', '/');
    writeRes(store, file, path);
  }

  private void writeRes(FStore store, File file, String path)
    throws IOException
  {
    try
    {
      OutputStream out = store.write(path);
      InputStream in = new BufferedInputStream(new FileInputStream(file));
      try
      {
        FileUtil.pipe(in, out);
      }
      finally
      {
        out.close();
        in.close();
      }
    }
    catch (IOException e)
    {
      throw err("Cannot write resource file '" + file + "' to pod", podFile, e);
    }
  }

//////////////////////////////////////////////////////////////////////////
// Doc
//////////////////////////////////////////////////////////////////////////

  private void writeDoc(FStore store)
    throws IOException
  {
    ClassDef[] classes = compiler.classDefs;
    for (int i=0; i<classes.length; ++i)
      if (!classes[i].isSynthetic())
        writeDoc(store, classes[i]);
  }

  private void writeDoc(FStore store, ClassDef cls)
    throws IOException
  {
    OutputStream fout = store.write("doc/" + cls.name + ".apidoc");
    PrintWriter out = new PrintWriter(new OutputStreamWriter(fout, "UTF-8"));
    try
    {
      FDoc.write(out, cls.qname, cls.doc);
      for (int i=0; i<cls.slots.size(); ++i)
      {
        SlotDef slot = (SlotDef)cls.slots.get(i);
        if (slot.doc == null) continue;
        FDoc.write(out, slot.slot().qname().val, slot.doc);
      }
    }
    finally
    {
      out.close();
    }
  }

//////////////////////////////////////////////////////////////////////////
// Src
//////////////////////////////////////////////////////////////////////////

  private void writeSrc(FStore store)
    throws IOException
  {
    for (int i=0; i<compiler.srcFiles.length; ++i)
      writeRes(store, compiler.srcFiles[i], "src/" + compiler.srcFiles[i].getName());
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  File podFile;

}