//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Dec 05  Brian Frank  Creation
//
package fanx.fcode;

import java.io.*;
import java.util.*;
import fanx.util.*;
import fan.sys.*;
import fan.sys.Float;

/**
 * FTable is a 16-bit indexed lookup table for pod constants.
 */
public abstract class FTable
{

//////////////////////////////////////////////////////////////////////////
// Constructor
//////////////////////////////////////////////////////////////////////////

  protected FTable(FPod pod)
  {
    this.pod = pod;
  }

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  /**
   * Get the size of the table.
   */
  public final int size() { return size; }

  /**
   * Get the object identified by the specified 16-bit index.
   */
  public final Object get(int index)
  {
    return table[index];
  }

  /**
   * Perform a reverse lookup to map a value to it's index (only
   * available at compile time).  If the value isn't in the table
   * yet, then add it.
   */
  public final int put(Object val)
  {
    Integer index = (Integer)reverse.get(val);
    if (index == null)
    {
      int n = size++;
      if (n == table.length)
      {
        Object[] temp = new Object[n*2];
        System.arraycopy(table, 0, temp, 0, n);
        table = temp;
      }
      table[n] = val;
      reverse.put(val, index = new Integer(n));
    }
    return index.intValue();
  }

  /**
   * Initialize for compilation (prime the tables to being adding)
   */
  public FTable initForCompile()
  {
    table = new Object[256];
    reverse = new HashMap(256*3);
    return this;
  }

  /**
   * Dump to the specified print writer.
   */
  public void dump(FPod pod, PrintWriter out)
  {
    for (int i=0; i<size; ++i)
    {
      out.print(StrUtil.padr("  [" + i + "] ", 8));
      out.println(toString(i));
    }
    out.flush();
  }

  /**
   * Get the value at specified index formated as a String.
   */
  public String toString(int index)
  {
    return table[index].toString();
  }

  /**
   * Serialize.
   */
  public abstract FTable read(FStore.Input in)
    throws IOException;

  /**
   * Deserialize.
   */
  public abstract void write(FStore.Output out)
    throws IOException;

//////////////////////////////////////////////////////////////////////////
// Names
//////////////////////////////////////////////////////////////////////////

  static class Names extends FTable
  {
    Names(FPod pod) { super(pod); }

    public FTable read(FStore.Input in) throws IOException
    {
      if (in == null) { size = 0; return this; }
      size = in.u2();
      table = new Object[size];
      for (int i=0; i<size; ++i)
        table[i] = in.utf().intern();
      return this;
    }

    public void write(FStore.Output out) throws IOException
    {
      out.u2(size);
      for (int i=0; i<size; ++i) out.utf((String)table[i]);
      out.close();
    }
  }

//////////////////////////////////////////////////////////////////////////
// TypeRefs
//////////////////////////////////////////////////////////////////////////

  static class TypeRefs extends FTable
  {
    TypeRefs(FPod pod) { super(pod); }

    public String toString(int index)
    {
      if (index == -1) return "null";
      return ((FTypeRef)table[index]).sig(pod);
    }

    public FTable read(FStore.Input in) throws IOException
    {
      if (in == null) { size = 0; return this; }
      size = in.u2();
      table = new Object[size];
      for (int i=0; i<size; ++i)
        table[i] = FTypeRef.read(in);
      return this;
    }

    public void write(FStore.Output out) throws IOException
    {
      out.u2(size);
      for (int i=0; i<size; ++i)
        ((FTypeRef)table[i]).write(out);
      out.close();
    }
  }

//////////////////////////////////////////////////////////////////////////
// FieldRefs
//////////////////////////////////////////////////////////////////////////

  static class FieldRefs extends FTable
  {
    FieldRefs(FPod pod) { super(pod); }

    public String toString(int index)
    {
      int[] val = ((FTuple)table[index]).val;
      return pod.typeRefs.toString(val[0]) + "." + pod.name(val[1]) +
        " -> " + pod.typeRefs.toString(val[2]);
    }

    public FTable read(FStore.Input in) throws IOException
    {
      if (in == null) { size = 0; return this; }
      size = in.u2();
      table = new Object[size];
      for (int i=0; i<size; ++i)
      {
        int[] x = { in.u2(), in.u2(), in.u2() };
        table[i] = new FTuple(x);
      }
      return this;
    }

    public void write(FStore.Output out) throws IOException
    {
      out.u2(size);
      for (int i=0; i<size; ++i)
      {
        int[] x = ((FTuple)table[i]).val;
        out.u2(x[0]);  // parent
        out.u2(x[1]);  // name
        out.u2(x[2]);  // type
      }
      out.close();
    }
  }

//////////////////////////////////////////////////////////////////////////
// MethodRefs
//////////////////////////////////////////////////////////////////////////

  static class MethodRefs extends FTable
  {
    MethodRefs(FPod pod) { super(pod); }

    public String toString(int index)
    {
      int[] val = ((FTuple)table[index]).val;
      StringBuilder s = new StringBuilder();
      s.append(pod.typeRefs.toString(val[0])).append(".").append(pod.name(val[1]));
      s.append('(');
      for (int i=3; i<val.length; ++i)
      {
        if (i > 3) s.append(", ");
        s.append(pod.typeRefs.toString(val[i]));
      }
      s.append(") -> ").append(pod.typeRefs.toString(val[2]));
      return s.toString();
    }

    public FTable read(FStore.Input in) throws IOException
    {
      if (in == null) { size = 0; return this; }
      size = in.u2();
      table = new Object[size];
      for (int i=0; i<size; ++i)
      {
        int parent = in.u2();
        int name   = in.u2();
        int ret    = in.u2();
        int paramn = in.u1();
        int[] x = new int[3+paramn];
        x[0] = parent;
        x[1] = name;
        x[2] = ret;
        for (int j=0; j<paramn; ++j)
          x[j+3] = in.u2();
        table[i] = new FTuple(x);
      }
      return this;
    }

    public void write(FStore.Output out) throws IOException
    {
      out.u2(size);
      for (int i=0; i<size; ++i)
      {
        int[] x = ((FTuple)table[i]).val;
        out.u2(x[0]);  // parent
        out.u2(x[1]);  // name
        out.u2(x[2]);  // ret
        out.u1(x.length-3); // param count
        for (int j=3; j<x.length; ++j)
          out.u2(x[j]); // param[x]
      }
      out.close();
    }
  }

//////////////////////////////////////////////////////////////////////////
// Ints
//////////////////////////////////////////////////////////////////////////

  static class Ints extends FTable
  {
    Ints(FPod pod) { super(pod); }

    public FTable read(FStore.Input in) throws IOException
    {
       if (in == null) { size = 0; return this; }
       size = in.u2();
       table = new Object[size];
       for (int i=0; i<size; ++i)
         table[i] = Int.make( in.u8() );
       return this;
    }

    public void write(FStore.Output out) throws IOException
    {
      out.u2(size);
      for (int i=0; i<size; ++i) out.u8(((Int)table[i]).val);
      out.close();
    }
  }

//////////////////////////////////////////////////////////////////////////
// Floats
//////////////////////////////////////////////////////////////////////////

  static class Floats extends FTable
  {
    Floats(FPod pod) { super(pod); }

    public FTable read(FStore.Input in) throws IOException
    {
       if (in == null) { size = 0; return this; }
       size = in.u2();
       table = new Object[size];
       for (int i=0; i<size; ++i)
         table[i] = Float.make( in.f8() );
       return this;
    }

    public void write(FStore.Output out) throws IOException
    {
      out.u2(size);
      for (int i=0; i<size; ++i) out.f8(((Float)table[i]).val);
      out.close();
    }
  }

//////////////////////////////////////////////////////////////////////////
// Strs
//////////////////////////////////////////////////////////////////////////

  static class Strs extends FTable
  {
    Strs(FPod pod) { super(pod); }

    public FTable read(FStore.Input in) throws IOException
    {
       if (in == null) { size = 0; return this; }
       size = in.u2();
       table = new Object[size];
       for (int i=0; i<size; ++i)
         table[i] = Str.make(in.utf()).intern();
       return this;
    }

    public void write(FStore.Output out) throws IOException
    {
      out.u2(size);
      for (int i=0; i<size; ++i) out.utf(((Str)table[i]).val);
      out.close();
    }
  }

//////////////////////////////////////////////////////////////////////////
// Durations
//////////////////////////////////////////////////////////////////////////

  static class Durations extends FTable
  {
    Durations(FPod pod) { super(pod); }

    public FTable read(FStore.Input in) throws IOException
    {
       if (in == null) { size = 0; return this; }
       size = in.u2();
       table = new Object[size];
       for (int i=0; i<size; ++i)
         table[i] = Duration.make( in.u8() );
       return this;
    }

    public void write(FStore.Output out) throws IOException
    {
      out.u2(size);
      for (int i=0; i<size; ++i) out.u8(((Duration)table[i]).ticks);
      out.close();
    }
  }

//////////////////////////////////////////////////////////////////////////
// Uris
//////////////////////////////////////////////////////////////////////////

  static class Uris extends FTable
  {
    Uris(FPod pod) { super(pod); }

    public FTable read(FStore.Input in) throws IOException
    {
       if (in == null) { size = 0; return this; }
       size = in.u2();
       table = new Object[size];
       for (int i=0; i<size; ++i)
         table[i] = Uri.parse( Str.make(in.utf()) );
       return this;
    }

    public void write(FStore.Output out) throws IOException
    {
      out.u2(size);
      for (int i=0; i<size; ++i) out.utf(table[i].toString());
      out.close();
    }
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  FPod pod;
  int size;
  Object[] table;
  HashMap reverse;
  int tupleFormat;

}