//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   20 Jan 06  Brian Frank  Creation
//
package fanx.test;

import fan.sys.*;

/**
 * MethodTest
 */
public class MethodTest
  extends CompileTest
{

//////////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////////

  public void doRun()
    throws Exception
  {
    verifyTyping();
    verifyCall();
  }

//////////////////////////////////////////////////////////////////////////
// Typing
//////////////////////////////////////////////////////////////////////////

  public void verifyTyping()
    throws Exception
  {
    verifySig("| |", new Type[] { }, Sys.VoidType);

    verifySig("|->Str|", new Type[] { }, Sys.StrType);

    verifySig("|->Void|", new Type[] { }, Sys.VoidType);

    verifySig("|Bool x-> Str|", new Type[] { Sys.BoolType }, Sys.StrType);

    verifySig("|Str a, Duration b -> Int|",
              new Type[] { Sys.StrType, Sys.DurationType },
              Sys.IntType);

    verifySig("|Bool a, Int b, Int c, Str d, Type e, Obj f, Str g, Bool h -> Int|",
             new Type[] { Sys.BoolType, Sys.IntType, Sys.IntType, Sys.StrType,
               Sys.TypeType, Sys.ObjType, Sys.StrType, Sys.BoolType  },
             Sys.IntType);

    verifySig("|Bool a, Int b, Int c, Str d, Type e, Obj f, Str g, Bool h, Duration i -> Int|",
             new Type[] { Sys.BoolType, Sys.IntType, Sys.IntType, Sys.StrType,
               Sys.TypeType, Sys.ObjType, Sys.StrType, Sys.BoolType, Sys.DurationType  },
             Sys.IntType);

    verifySigErr("|Bool a-> Int,Bool|", "Expected '|', not ','");
  }

  void verifySigErr(String sig, String msg)
    throws Exception
  {
    verifyErr("static Obj f(" + sig + " m) { return null }", msg);
  }

  void verifySig(String sig, Type[] p, Type r)
    throws Exception
  {
    // test as method param signature
    Type t = compileToType("class Foo { static Obj f(" + sig + " m) { return null } }");
    Method m = t.method("f", true);
    MethodParam params = (MethodParam)m.params().get(0);
    verifySig(params.paramType(), p, r);

    // test as type literal
    t = compileToType("class Foo { static Type f() { return " + sig + ".type } }");
    m = t.method("f", true);
    verifySig((Type)m.call0(), p, r);

    // test as closure signature
    Class cls;
    if (r.isVoid())
      cls = compileToClass("class Foo { static Method f() { return " + sig + " { return; } }}");
    else
      cls = compileToClass("class Foo { static Method f() { return " + sig + " { return null; } }}");
    m = (Method)invoke(cls, "f");
    verifySig(m.type(), p, r);
  }

  void verifySig(Type t, Type[] p, Type r)
  {
    for (int i=0; i<=Method.MaxIndirectParams; ++i)
    {
      Method c = t.method("call"+i, true);

      // verify return
      verify(c.returnType() == r);

      // verify p0..pn params
      for (int j=0; j<p.length && j<i; ++j)
      {
        verify(((MethodParam)c.params().get(j)).paramType() == p[j]);
      }

      // verify rest left at Obj
      for (int j=p.length; j<Method.MaxIndirectParams && j<i; ++j)
      {
        verify(((MethodParam)c.params().get(j)).paramType() == Sys.ObjType);
      }
    }
  }

//////////////////////////////////////////////////////////////////////////
// Call
//////////////////////////////////////////////////////////////////////////

  public void verifyCall()
    throws Exception
  {
    // static
    verify("Obj f(|Str s, Bool c -> Type| m) { return m.call([\"sys::Str\", true]) }",
      new Obj[] { Slot.findMethod("sys::Type.find", true) }, Sys.StrType);
    verify("Obj f(|Str s, Bool c -> Type| m) { return m.call([\"sys::Str\", true, 55]) }",
      new Obj[] { Slot.findMethod("sys::Type.find", true) }, Sys.StrType);

    // instance
    verify("Obj f(|Int a, Int b -> Int| m) { return m.call([4, 2])}",
      new Obj[] { Slot.findMethod("sys::Int.plus", true) }, Int.make(6));
    verify("Obj f(|Int a, Int b -> Int| m) { return m.call([4, 2])}",
      new Obj[] { Slot.findMethod("sys::Int.star", true) }, Int.make(8));
    verify("Obj f(|Int a, Int b -> Int| m) { return m.call([4, 2, 3])}",
      new Obj[] { Slot.findMethod("sys::Int.star", true) }, Int.make(8));
    verify("Obj f(|Int a, Int b -> Int| m) { return m.call([-3])}",
      new Obj[] { Slot.findMethod("sys::Int.negate", true) }, Int.make(3));
    verify("Obj f(|Int a, Int b -> Int| m) { return m.call([4, 2])}",
      new Obj[] { Slot.findMethod("sys::Int.negate", true) }, Int.make(-4));
  }

}