//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   15 Sep 05  Brian Frank  Creation
//
package fanx.tools;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.zip.*;
import fanx.util.*;
import fanx.compiler.*;
import fanx.compiler.Compiler;

/**
 * Fanc is the main entry point for the Fan compiler.  It handles
 * all the argument processing and misc commands like help, version,
 * It launches a Compiler for each source directory based on the target
 * platform.
 */
public class Fanc
{

//////////////////////////////////////////////////////////////////////////
// Run
//////////////////////////////////////////////////////////////////////////

  /**
   * Main entry point for compiler.
   */
  public int run(String[] args)
  {
    long t1 = System.currentTimeMillis();
    boolean success = true;

    // process args
    if (!processArgs(args)) return 0;

    // process each directory specified
    for (int i=0; i<srcDirs.size(); ++i)
    {
      File srcDir = (File)srcDirs.get(i);
      try
      {
        compile(srcDir);
      }
      catch (CompileException e)
      {
        // all errors should already be logged by Compiler
        success = false;
        break;
      }
      catch (Throwable e)
      {
        log.error("Internal compiler error", srcDir, e);
        success = false;
        break;
      }
    }

    long t2 = System.currentTimeMillis();
    if (success)
    {
      log.println("SUCCESS (" + (t2-t1) + "ms)");
      log.flush();
      return 0;
    }
    else
    {
      log.println("FAILED (" + (t2-t1) + "ms)");
      log.flush();
      return -1;
    }
  }

  /**
   * Process command line args and return false if we should exit.
   */
  boolean processArgs(String[] args)
  {
    // process args
    for (int i=0; i<args.length; ++i)
    {
      String a = args[i].intern();
      if (a.length() == 0) continue;
      if (a == "-help" || a == "-h" || a == "-?")
      {
        help();
        return false;
      }
      else if (a == "-version")
      {
        Fan.version();
        return false;
      }
      else if (a == "-props")
      {
        props();
        return false;
      }
      else if (a == "-d ")
      {
        if (i+1 >= args.length)
        {
          println("ERROR: must specified dir with -d option");
          return false;
        }
        outDir = new File(args[++i]);
      }
      else if (a == "-doc")
      {
        doc = true;
      }
      else if (a == "-src")
      {
        src = true;
      }
      else if (a == "-silent")
      {
        log.silent = true;
      }
      else if (a == "-v")
      {
        log.verbose = true;
      }
      else if (a.charAt(0) == '-')
      {
        println("WARNING: Unknown option " + a);
      }
      else
      {
        srcDirs.add(new File(a));
      }
    }

    // if no dirs were specified, assume current dir
    if (srcDirs.size() == 0) srcDirs.add(new File("."));

    return true;
  }

  /**
   * Dump help usage.
   */
  void help()
  {
    println("Fan Compiler");
    println("Usage:");
    println("  fanc [options] <srcDirs>");
    println("Options:");
    println("  -help, -h, -?  print usage help");
    println("  -version       print version information");
    println("  -d <dir>       output directory for pod file");
    println("  -doc           include fandoc in pod");
    println("  -src           include source code in pod");
    println("  -v             verbose mode (more logging)");
    println("  -silent        silent mode (no logging)");
  }

  /**
   * Dump system properties
   */
  void props()
  {
    ArrayList keys = new ArrayList();
    Enumeration e = System.getProperties().propertyNames();
    while (e.hasMoreElements())
      keys.add(e.nextElement());

    Collections.sort(keys);

    for (int i=0; i<keys.size(); ++i)
    {
      String key = keys.get(i).toString();
      String val = System.getProperty(key);
      System.out.println(key + "=" + StrUtil.asCode(val));
    }
  }

  /**
   * Compile the specified source directories
   */
  void compile(File srcDir)
    throws Exception
  {
    srcDir = new File(srcDir.getCanonicalPath());
    Compiler compiler = new Compiler();
    compiler.log = log;
    compiler.includeDoc = doc;
    compiler.includeSrc = src;
    if (outDir != null) compiler.outDir = outDir;
    compiler.compile(srcDir);
  }

//////////////////////////////////////////////////////////////////////////
// Utils
//////////////////////////////////////////////////////////////////////////

  public File[] fileList(String s)
  {
    ArrayList acc = new ArrayList();
    StringTokenizer st = new StringTokenizer(s, ";");
    while (st.hasMoreTokens())
      acc.add(new File(st.nextToken()));
    return (File[])acc.toArray(new File[acc.size()]);
  }

  public String fileList(File[] list)
  {
    StringBuilder s = new StringBuilder();
    for (int i=0; i<list.length; ++i)
    {
      if (i > 0) s.append(';');
      s.append(list[i]);
    }
    return s.toString();
  }

  public static void println(String s)
  {
    System.out.println(s);
  }

//////////////////////////////////////////////////////////////////////////
// Main
//////////////////////////////////////////////////////////////////////////

  public static void main(String[] args)
    throws Exception
  {
    System.exit(new Fanc().run(args));
  }

//////////////////////////////////////////////////////////////////////////
// Fields
//////////////////////////////////////////////////////////////////////////

  File outDir = null;           // -d output directory
  boolean doc = false;          // include fandoc
  boolean src = false;          // include source
  Log log = new Log();          // logging, -v verbose output
  ArrayList srcDirs = new ArrayList();   // directories to build


}