//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   26 Feb 07  Andy Frank  Creation
//

using System;
using System.IO;
using System.Text;

namespace Fan.Sys
{
  /// <summary>
  /// Buf
  /// </summary>
  public class Buf : FanObj, OutStream, InStream
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public static Buf make() { return new Buf().init(DefaultCapacity); }
    public static Buf make(int capacity) { return new Buf().init(capacity); }
    public static Buf make(Int capacity) { return new Buf().init((int)capacity.val); }
    public static void make_(Buf self) { self.init(DefaultCapacity); }
    public static void make_(Buf self, Int capacity) { self.init((int)capacity.val); }

    public static Buf make(byte[] bytes)
    {
      Buf buf  = new Buf();
      buf.m_buf  = bytes;
      buf.m_pos  = 0;
      buf.m_size = bytes.Length;
      return buf;
    }

    public Buf init(int capacity)
    {
      this.m_buf  = new byte[capacity];
      this.m_pos  = 0;
      this.m_size = 0;
      return this;
    }

    internal static readonly int DefaultCapacity = 1024;

  //////////////////////////////////////////////////////////////////////////
  // Obj
  //////////////////////////////////////////////////////////////////////////

    public override Bool equals(Obj that)
    {
      if (that is Buf)
      {
        Buf x = (Buf)that;
        if (m_size != x.m_size) return Bool.False;
        for (int i=0; i<m_size; i++)
          if (m_buf[i] != x.m_buf[i]) return Bool.False;
        return Bool.True;
      }
      return Bool.False;
    }

    public override Int hash()
    {
      int hash = 30311;
      for (int i=0; i<m_size; i++)
        hash ^= m_buf[i] << i%32;
      return Int.make(hash);
    }

    public override Str toStr()
    {
      StringBuilder s = new StringBuilder();
      s.Append("0x[");
      for (int i=0; i<m_size; i++)
      {
        int b = m_buf[i] & 0xFF;
        if (b < 0x10) s.Append('0');
        s.Append(b.ToString("X").ToLower());
        if (i > 100) { s.Append("..."); break; }
        if ((i+1) % 4 == 0 && i+1<m_size) s.Append(' ');
      }
      s.Append(']');
      return Str.make(s.ToString());
    }

    public override Type type() { return Sys.BufType; }

  //////////////////////////////////////////////////////////////////////////
  // Access
  //////////////////////////////////////////////////////////////////////////

    public Bool empty() { return m_size == 0 ? Bool.True : Bool.False; }

    public Int capacity()
    {
      return Int.pos(m_buf.Length);
    }

    public void capacity(Int c)
    {
      int newCapacity = (int)c.val;
      if (newCapacity < m_size) throw ArgErr.make("capacity < size").val;
      byte[] temp = new byte[newCapacity];
      Array.Copy(m_buf, 0, temp, 0, newCapacity);
      m_buf = temp;
    }

    public Int size()
    {
      return Int.make(m_size);
    }

    public void size(Int s)
    {
      int newSize = (int)s.val;
      if (newSize > m_buf.Length)
      {
        byte[] temp = new byte[newSize];
        Array.Copy(m_buf, 0, temp, 0, m_buf.Length);
        m_buf  = temp;
      }
      m_size = newSize;
    }

    public Int pos() { return Int.make(m_pos); }

    public Int remaining() { return Int.make(m_size-m_pos); }

    public Bool more() { return m_size-m_pos > 0 ? Bool.True : Bool.False; }

    public Buf seek(Int pos)
    {
      int p = (int)pos.val;
      if (p < 0) p = m_size + p;
      if (p < 0 || p > m_size) throw IndexErr.make(pos).val;
      this.m_pos = p;
      return this;
    }

    public Buf flip()
    {
      this.m_size = m_pos;
      this.m_pos  = 0;
      return this;
    }

    public Int get(Int pos)
    {
      int p = (int)pos.val;
      if (p < 0) p = m_size + p;
      if (p < 0 || p >= m_size) throw IndexErr.make(pos).val;
      return Int.make(m_buf[p] & 0xFF);
    }

    public Buf slice(Range range)
    {
      int s = range.start(m_size);
      int e = range.end(m_size);
      int n = e - s + 1;
      if (n < 0) throw IndexErr.make(range).val;

      byte[] slice = new byte[n];
      Array.Copy(m_buf, s, slice, 0, n);

      Buf result = new Buf();
      result.m_buf = slice;
      result.m_size = n;
      result.m_charset = m_charset;
      return result;
    }

  //////////////////////////////////////////////////////////////////////////
  // Modification
  //////////////////////////////////////////////////////////////////////////

    public Buf set(Int pos, Int b)
    {
      int p = (int)pos.val;
      if (p < 0) p = m_size + p;
      if (p < 0 || p >= m_size) throw IndexErr.make(pos).val;
      m_buf[p] = (byte)b.val;
      return this;
    }

    public Buf trim()
    {
      if (m_size == m_buf.Length) return this;
      byte[] temp = new byte[m_size];
      Array.Copy(m_buf, 0, temp, 0, m_size);
      this.m_buf = temp;
      return this;
    }

    public Buf clear()
    {
      this.m_size = 0;
      this.m_pos  = 0;
      return this;
    }

  //////////////////////////////////////////////////////////////////////////
  // OutStream
  //////////////////////////////////////////////////////////////////////////

    public OutStream write(Int b) { return w((int)b.val); }
    public OutStream w(int v)
    {
      if (m_pos+1 >= m_buf.Length) grow(m_pos+1);
      m_buf[m_pos++] = (byte)v;
      if (m_pos > m_size) m_size = m_pos;
      return this;
    }

    public OutStream writeBuf(Buf other) { return writeBuf(other, other.remaining()); }
    public OutStream writeBuf(Buf other, Int n)
    {
      int len = (int)n.val;
      if (other.m_pos + len > other.m_size) throw IOErr.make("Not enough bytes to write").val;
      grow(m_pos+len);
      Array.Copy(other.m_buf, other.m_pos, m_buf, m_pos, len);
      m_pos += len;
      if (m_pos > m_size) m_size = m_pos;
      other.m_pos += len;
      return this;
    }

    public OutStream writeI2(Int x) { return OutStream_.writeI2(this, x); }

    public OutStream writeI4(Int x) { return OutStream_.writeI4(this, x); }

    public OutStream writeI8(Int x) { return OutStream_.writeI8(this, x); }

    public OutStream writeF4(Float x) { return OutStream_.writeF4(this, x); }

    public OutStream writeF8(Float x) { return OutStream_.writeF8(this, x); }

    public OutStream writeBool(Bool x) { return OutStream_.writeBool(this, x); }

    public OutStream writeUtf(Str x) { return OutStream_.writeUtf(this, x); }

    public OutStream writeChar(Int c)
    {
      if (m_charsetEncoder == null) m_charsetEncoder = m_charset.newEncoder();
      m_charsetEncoder.encode((char)c.val, (OutStream)this);
      return this;
    }

    public OutStream writeChars(Str s) { return writeChars(s.val, 0, s.val.Length); }
    public OutStream writeChars(Str s, Int off) { return writeChars(s.val, (int)off.val, s.val.Length-(int)off.val); }
    public OutStream writeChars(Str s, Int off, Int len) { return writeChars(s.val, (int)off.val, (int)len.val); }
    public OutStream writeChars(String s, int off, int len)
    {
      if (m_charsetEncoder == null) m_charsetEncoder = m_charset.newEncoder();
      int end = off+len;
      for (int i=off; i<end; i++)
        m_charsetEncoder.encode(s[i], (OutStream)this);
      return this;
    }

    public OutStream print(Obj obj) { return OutStream_.print(this, obj); }

    public OutStream printLine() { return OutStream_.printLine(this, Str.Empty); }
    public OutStream printLine(Obj obj) { return OutStream_.printLine(this, obj); }

    public OutStream writeProps(Map props) { return OutStream_.writeProps(this, props, Bool.True); }
    public OutStream writeProps(Map props, Bool close) { return OutStream_.writeProps(this, props, close); }

  //////////////////////////////////////////////////////////////////////////
  // InStream
  //////////////////////////////////////////////////////////////////////////

    public Int read() {  int n = r(); return n < 0 ? null : Int.pos(n); }
    public int r()
    {
      if (m_pos >= m_size) return -1;
      return m_buf[m_pos++] & 0xFF;
    }

    public Int readBuf(Buf other, Int n)
    {
      if (m_pos >= m_size) return null;
      int len = Math.Min(m_size-m_pos, (int)n.val);
      other.grow(other.m_pos+len);
      Array.Copy(m_buf, m_pos, other.m_buf, other.m_pos, len);
      m_pos += len;
      other.m_pos += len;
      other.m_size += len;
      return Int.pos(len);
    }

    public InStream unread(Int n) { return unread((int)n.val); }
    public InStream unread(int n)
    {
      // unreading a buffer is a bit weird - the typical case
      // is that we are pushing back the byte we just read in
      // which case we can just rewind the position; however
      // if we pushing back a different byte then we need
      // to shift the entire buffer and insert the byte
      if (m_pos > 0 && m_buf[m_pos-1] == (byte)n)
      {
        m_pos--;
      }
      else
      {
        if (m_size+1 >= m_buf.Length) grow(m_size+1);
        Array.Copy(m_buf, m_pos, m_buf, m_pos+1, m_size);
        m_buf[m_pos] = (byte)n;
        m_size++;
      }
      return this;
    }

    public Buf readBufFully(Buf buf, Int n) { return InStream_.readBufFully(this, buf, n); }

    public Buf readAllBuf() { return InStream_.readAllBuf(this); }

    public Int readU1() { return InStream_.readU1(this); }

    public Int readS1() { return InStream_.readS1(this); }

    public Int readU2() { return InStream_.readU2(this); }

    public Int readS2() { return InStream_.readS2(this); }

    public Int readU4() { return InStream_.readU4(this); }

    public Int readS4() { return InStream_.readS4(this); }

    public Int readS8() { return InStream_.readS8(this); }

    public Float readF4() { return InStream_.readF4(this); }

    public Float readF8() { return InStream_.readF8(this); }

    public Bool readBool() { return InStream_.readBool(this); }

    public Str readUtf() { return InStream_.readUtf(this); }

    public int rChar()
    {
      if (m_charsetDecoder == null) m_charsetDecoder = m_charset.newDecoder();
      return m_charsetDecoder.decode(this);
    }

    public Int readChar()
    {
      if (m_charsetDecoder == null) m_charsetDecoder = m_charset.newDecoder();
      int ch = m_charsetDecoder.decode(this);
      return ch < 0 ? null : Int.pos(ch);
    }

    public InStream unreadChar(Int c)
    {
      if (m_charsetEncoder == null) m_charsetEncoder = m_charset.newEncoder();
      m_charsetEncoder.encode((char)c.val, (InStream)this);
      return this;
    }

    public Str readLine() { return InStream_.readLine(this, Int.m_pos[4096]); }
    public Str readLine(Int max) { return InStream_.readLine(this, max); }

    public List readAllLines() { return InStream_.readAllLines(this); }

    public Str readAllStr() { return InStream_.readAllStr(this, Bool.True); }
    public Str readAllStr(Bool normalizeNewlines)  { return InStream_.readAllStr(this, normalizeNewlines); }

    public Map readProps() { return InStream_.readProps(this); }

    public Int pipe(OutStream o) { return InStream_.pipe(this, o, null, Bool.True); }
    public Int pipe(OutStream o, Int n) { return InStream_.pipe(this, o, n, Bool.True); }
    public Int pipe(OutStream o, Int n, Bool close) { return InStream_.pipe(this, o, n, close); }

  //////////////////////////////////////////////////////////////////////////
  // OutStream and InStream
  //////////////////////////////////////////////////////////////////////////

    public void flush() { }

    public Bool close() { return Bool.True; }

    public Charset charset()
    {
      return m_charset;
    }

    public void charset(Charset charset)
    {
      m_charset = charset;
      m_charsetEncoder = null;
      m_charsetDecoder = null;
    }

  //////////////////////////////////////////////////////////////////////////
  // Utils
  //////////////////////////////////////////////////////////////////////////

    public byte[] bytes()
    {
      byte[] r = new byte[m_size];
      Array.Copy(m_buf, 0, r, 0, m_size);
      return r;
    }

    public void grow(int capacity)
    {
      if (m_buf.Length >= capacity) return;
      byte[] temp = new byte[Math.Max(capacity, m_size*2)];
      Array.Copy(m_buf, 0, temp, 0, m_size);
      m_buf = temp;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    Charset m_charset = Charset.utf8();
    Charset.Encoder m_charsetEncoder;
    Charset.Decoder m_charsetDecoder;

    public byte[] m_buf;
    public int m_pos;
    public int m_size;

  }
}