//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   18 Oct 06  Andy Frank  Creation
//

using System;

namespace Fan.Sys
{
  /// <summary>
  /// Duration
  /// </summary>
  public sealed class Duration : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Duration make(Int ticks) { return make(ticks.val); }
    public static Duration make(long ticks)
    {
      if (ticks == 0) return Zero;
      return new Duration(ticks);
    }

    public static Duration makeMillis(long ms)
    {
      return make(ms*1000000L);
    }

    public static Duration makeSec(long secs)
    {
      return make(secs*1000000000L);
    }

    public static Duration now()
    {
      return new Duration(DateTime.Now.Ticks * 100);
    }

    private Duration(long ticks)
    {
      this.m_ticks = ticks;
    }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override bool Equals(Object obj)
    {
      return m_ticks == (obj as Duration).m_ticks;
    }

    public override Bool equals(Obj obj)
    {
      if (obj is Duration)
        return m_ticks == ((Duration)obj).m_ticks ? Bool.True : Bool.False;
      else
        return Bool.False;
    }

    public override Int compare(Obj obj)
    {
      long that = ((Duration)obj).m_ticks;
      if (m_ticks < that) return Int.LT; return m_ticks  == that ? Int.EQ : Int.GT;
    }

    public override int GetHashCode()
    {
      //return (int)(ticks ^ (ticks >>> 32));
      return (int)(m_ticks ^ (m_ticks >> 32));
    }

    public override Int hash()
    {
      return Int.make(m_ticks);
    }

    public Int ticks()
    {
      return Int.make(m_ticks);
    }

    public override Type type()
    {
      return Sys.DurationType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Operators
  //////////////////////////////////////////////////////////////////////////

    public Duration negate()
    {
      return make(-m_ticks);
    }

    public Duration plus(Duration x)
    {
      return make(m_ticks + x.m_ticks);
    }

    public Duration minus(Duration x)
    {
      return make(m_ticks - x.m_ticks);
    }

    public Duration star(Float x)
    {
      return make((long)(m_ticks * x.val));
    }

    public Duration slash(Float x)
    {
      return make((long)(m_ticks / x.val));
    }

    public Duration percent(Float x)
    {
      return make((long)(m_ticks % x.val));
    }

  //////////////////////////////////////////////////////////////////////////
  // Conversion
  //////////////////////////////////////////////////////////////////////////

    public override Str toStr()
    {
      if (m_ticks == 0) return ZeroStr;

      // if clean millisecond boundary
      long ns = m_ticks;
      if (ns % 1000000L == 0)
      {
        if (ns % 86400000000000L == 0)
        {
          return Str.make(ns/86400000000000L + "days");
        }
        if (ns % 3600000000000L == 0)
        {
          return Str.make(ns/3600000000000L + "hr");
        }
        if (ns % 60000000000L == 0)
        {
          return Str.make(ns/60000000000L + "min");
        }
        if (ns % 1000000000L == 0)
        {
          return Str.make(ns/1000000000L + "sec");
        }
        return Str.make(ns/1000000L + "ms");
      }

      // return in nanoseconds
      return Str.make(ns + "ns");
    }

    public Int toMillis()
    {
      return Int.make(m_ticks/1000000L);
    }

  //////////////////////////////////////////////////////////////////////////
  // C#
  //////////////////////////////////////////////////////////////////////////

    public long sec()
    {
      return m_ticks/1000000000L;
    }

    public long millis()
    {
      return m_ticks/1000000L;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public static readonly Duration Zero = new Duration(0);
    public static readonly Str ZeroStr = Str.make("0ns");

    public readonly long m_ticks;

  }
}