//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   24 Sep 06  Andy Frank  Creation
//

using System;
using System.IO;

namespace Fan.Sys
{
  /// <summary>
  /// Err.
  /// </summary>
  public class Err : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Java to Fan Mapping
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Map a java exception to it's Fan Err counter part.  Common runtime
    /// exceptions are mapped into explicit Fan types.  Otherwise we just
    /// wrap the exception with a generic Err.
    /// </summary>
    public static Err make(Exception ex)
    {
      // NOTE: everything in this list must be synchronized
      // with the mapping used below for FCodeEmit.errTable()
      // and tested in TryTest
      if (ex is Val)                         return ((Val)ex).m_err;
      if (ex is NullReferenceException)      return new NullErr(ex);
      if (ex is InvalidCastException)        return new CastErr(ex);
      if (ex is IndexOutOfRangeException)    return new IndexErr(ex);
      if (ex is ArgumentOutOfRangeException) return new IndexErr(ex);
      if (ex is ArgumentException)           return new ArgErr(ex);
      //if (ex is UnsupportedOperationException)  return new UnsupportedErr(ex);
      if (ex is IOException)                 return new IOErr(ex);
      return new Err(new Err.Val(), ex);
    }

    /*/
     * This method is used by FCodeEmit to generate extra entries in the
     * exception table - for example if fcode says to trap NullErr, then
     * we also need to trap java.lang.NullPointerException.  Basically this
     * is the inverse of the mapping done in make(Throwable).
     */
    /*
    public static String fanToJava(String jtype)
    {
      if (jtype.equals("fan/sys/NullErr"))  return "java/lang/NullPointerException";
      if (jtype.equals("fan/sys/CastErr"))  return "java/lang/ClassCastException";
      if (jtype.equals("fan/sys/IndexErr")) return "java/lang/IndexOutOfBoundsException";
      if (jtype.equals("fan/sys/ArgErr"))   return "java/lang/IllegalArgumentException";
      if (jtype.equals("fan/sys/UnsupportedErr")) return "java/lang/UnsupportedOperationException";
      if (jtype.equals("fan/sys/IOErr"))    return "java/io/IOException";
      return null;
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // C# Convenience
  //////////////////////////////////////////////////////////////////////////

    public static Err make(string msg) { return make(Str.make(msg)); }
    public static Err make(string msg, Exception e) { return make(Str.make(msg)); }

  //////////////////////////////////////////////////////////////////////////
  // Fan Constructors
  //////////////////////////////////////////////////////////////////////////

    public static Err make() { return make((Str)null, (Err)null); }
    public static Err make(Str msg) { return make(msg, null); }
    public static Err make(Str msg, Err cause)
    {
      Err err = new Err(new Err.Val());
      make_(err, msg, cause);
      return err;
    }

    public static void make_(Err self) { make_(self, null);  }
    public static void make_(Err self, Str msg) { make_(self, msg, null); }
    public static void make_(Err self, Str msg, Err cause)
    {
      self.m_message = msg;
      self.m_cause   = cause;
    }

  //////////////////////////////////////////////////////////////////////////
  // C# Constructors
  //////////////////////////////////////////////////////////////////////////

    public Err() {}

    /// <summary>
    /// All subclasses must call this constructor with their
    /// typed Val exception which the real .NET exception we use.
    /// </summary>
    public Err(Val val)
    {
      this.val = val;
      val.m_err = this;
    }

    /// <summary>
    /// This constructor is used by special subclasses which provide
    /// a transparent mapping between .NET and Fan exception types.
    /// </summary>
    public Err(Val val, Exception actual)
    {
      this.val = val;
      val.m_err = this;
      this.m_actual = actual;
      this.m_message = Str.make(actual.ToString());
    }

  //////////////////////////////////////////////////////////////////////////
  // Methods
  //////////////////////////////////////////////////////////////////////////

    public Str message()
    {
      return m_message;
    }

    public Err cause()
    {
      return m_cause;
    }

    public void trace()
    {
      if (m_actual != null)
        Console.WriteLine(m_actual.ToString());
      else
        Console.WriteLine(val.ToString());
      if (m_cause != null) m_cause.trace();   // TODO - how to handle?
    }

    public void trace(OutStream @out)
    {
      StringWriter writer = new StringWriter();
      if (m_actual != null)
        writer.WriteLine(m_actual.ToString());
      else
        writer.WriteLine(val.ToString());
      writer.Flush();
      @out.writeChars(Str.make(writer.ToString())).flush();
      if (m_cause != null) m_cause.trace();  // TODO - how to handle?
    }

    public override Type type()
    {
      return Sys.ErrType;
    }

    public override Str toStr()
    {
      if (m_message == null)
        return type().qname().toStr();
      else
        return Str.make(type().qname() + ": " + m_message);
    }

  //////////////////////////////////////////////////////////////////////////
  // Val
  //////////////////////////////////////////////////////////////////////////

    public class Val : Exception
    {
      public override string Message { get { return m_err.ToString(); }}
      public Err err() { return m_err; }
      internal Err m_err;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public readonly Val val;
    internal Str m_message;
    internal Err m_cause = null;
    internal Exception m_actual;

  }
}