//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Sep 06  Andy Frank  Creation
//

using System;
using Fanx.Fcode;

namespace Fan.Sys
{
  /// <summary>
  /// Facets encapsulate meta-data as a map of Str name/value pairs.
  /// </summary>
  public sealed class Facets : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // FCode
  //////////////////////////////////////////////////////////////////////////

    public static Facets make(FPod fpod, object spec1, object spec2, FFacet[] facets)
    {
      Facets r = make();
      for (int i=0; i<facets.Length; i++)
      {
        FFacet f = facets[i];
        if ((f.m_flags & FConst.Reflective) != 0)
        {
          string name = fpod.name(f.m_name);
          try
          {
            string val  = f.m_data.utf();
            r.set(Str.make(name), Str.make(val));
          }
          catch (Exception)
          {
            string spec = spec1.ToString();
            if (spec2 != null) spec += "." + spec2;
            Console.WriteLine("ERROR: Invalid facet value: " + spec + "/" + name);
          }
        }
      }
      return r.ro();
    }

  //////////////////////////////////////////////////////////////////////////
  // Constructors
  //////////////////////////////////////////////////////////////////////////

    public static Facets make()
    {
      return new Facets(new Map(Sys.StrType, Sys.StrType));
    }

    public Facets(Map map)
    {
      this.m_map = map;
    }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override Bool equals(Obj that)
    {
      if (that is Facets)
      {
        return m_map.equals(((Facets)that).m_map);
      }
      return Bool.False;
    }

    public override Int hash() { return m_map.hash(); }

    public override Str toStr() { return m_map.toStr(); }

    public override Type type() { return Sys.FacetsType; }

  //////////////////////////////////////////////////////////////////////////
  // Map
  //////////////////////////////////////////////////////////////////////////

    public Bool isEmpty() { return m_map.isEmpty(); }

    public Int size() { return m_map.size(); }

    public List keys() { return m_map.keys(); }

  //////////////////////////////////////////////////////////////////////////
  // Getters
  //////////////////////////////////////////////////////////////////////////

    public Str get(Str key) { return (Str)m_map.get(key); }
    public Str get(Str key, Str def)
    {
      Str val = (Str)m_map.get(key);
      if (val == null) return def;
      return val;
    }

    public Bool getBool(Str key) { return getBool(key, null); }
    public Bool getBool(Str key, Bool def)
    {
      Str val = (Str)m_map.get(key);
      if (val == null) return def;
      return val.toBool();
    }

    public Int getInt(Str key) { return getInt(key, null); }
    public Int getInt(Str key, Int def)
    {
      Str val = (Str)m_map.get(key);
      if (val == null) return def;
      return val.toInt();
    }

    public Float getFloat(Str key) { return getFloat(key, null); }
    public Float getFloat(Str key, Float def)
    {
      Str val = (Str)m_map.get(key);
      if (val == null) return def;
      return val.toFloat();
    }

  //////////////////////////////////////////////////////////////////////////
  // Setters
  //////////////////////////////////////////////////////////////////////////

    public Facets set(Str key, Str val)
    {
      m_map.set(key, val);
      return this;
    }

    public Facets setBool(Str key, Bool val)
    {
      return set(key, val.toStr());
    }

    public Facets setInt(Str key, Int val)
    {
      return set(key, val.toStr());
    }

    public Facets setFloat(Str key, Float val)
    {
      return set(key, val.toStr());
    }

    public Str remove(Str key)
    {
      return (Str)m_map.remove(key);
    }

    public void clear()
    {
      m_map.clear();
    }

  //////////////////////////////////////////////////////////////////////////
  // Readonly
  //////////////////////////////////////////////////////////////////////////

    public Bool isRW()
    {
      return m_map.isRW();
    }

    public Bool isRO()
    {
      return m_map.isRO();
    }

    public Facets rw()
    {
      if (isRW().val) return this;
      return new Facets(m_map.rw());
    }

    public Facets ro()
    {
      if (isRO().val) return this;
      return new Facets(m_map.ro());
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    private Map m_map;

  }
}