//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Dec 06  Andy Frank  Creation
//

using System.Collections;
using Fanx.Fcode;

namespace Fan.Sys
{
  /// <summary>
  /// GenericType is the base class for ListType, MapType, and MethodType
  /// which all support parameterization of the generic parameter types (such
  /// as A-H, V, K).  Instances of GenericType are used to represent generic
  /// instances (for example an instance of ListType is used to represent Str[]).
  /// </summary>
  public abstract class GenericType : Type
  {

  //////////////////////////////////////////////////////////////////////////
  // Variegate
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// This method is done during fcode reflect to variegate a variant type (List,
    /// Map, Method).  At this point we parameterize the variance types from the
    /// specified tuple binary fcode representation.
    /// </summary>
    static Type variegate(Pod pod, Type type, FTuple ftuple)
    {
      Variances var = new Variances().load(pod, ftuple.val);
      if (type == Sys.ListType) return var.v.toListOf();
      throw Err.make("Not variant type: " + type).val;
    }

    class Variances
    {
     public Variances load(Pod pod, int[] val)
     {
       int varCount = (val.Length-2)/2;
       for (int i=0; i<varCount; i++)
         load(pod, val[i*2+2], val[i*2+3]);
       return this;
     }

     public void load(Pod pod, int id, int typeRef)
     {
       Type type = pod.findType(typeRef);
       switch (id)
       {
         case 'A': a = type; break;
         case 'B': b = type; break;
         case 'C': c = type; break;
         case 'D': d = type; break;
         case 'E': e = type; break;
         case 'F': f = type; break;
         case 'G': g = type; break;
         case 'H': h = type; break;
         case 'K': k = type; break;
         case 'R': r = type; break;
         case 'V': v = type; break;
         default: throw Err.make("Unknown variance " + (char)id + " [" + id + "]").val;
       }
     }

     internal Type a, b, c, d, e, f, g, h, k, r, v;
    }

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    internal GenericType(Type baseType) :
      base(baseType.m_pod, baseType.m_name.val, baseType.m_flags, baseType.m_facets)
    {
    }

  //////////////////////////////////////////////////////////////////////////
  // Type
  //////////////////////////////////////////////////////////////////////////

    public override abstract Type @base();

    public override List mixins()
    {
      return @base().mixins();
    }

    public override abstract Str signature();

    public override bool isGenericInstance() { return true; }

    public override bool @is(Type type)
    {
      if (type == this || type == @base()) return true;
      return @base().@is(type);
    }

    public override Map @params()
    {
      if (m_pars == null) m_pars = makeParams();
      return m_pars;
    }

    internal abstract Map makeParams();

  //////////////////////////////////////////////////////////////////////////
  // Reflect
  //////////////////////////////////////////////////////////////////////////

    /**
     * On reflection, we parameterize the master's methods.
     */
    protected override void doReflect()
    {
      // ensure master type is reflected
      Type master = @base();
      master.reflect();
      List masterSlots = master.m_slots;

      // allocate slot data structures
      m_fields = new List(Sys.FieldType, master.m_fields.sz());
      m_methods = new List(Sys.MethodType, master.m_methods.sz());
      m_slots = new List(Sys.SlotType, masterSlots.sz());
      m_slotsByName = new Hashtable(masterSlots.sz()*3);

      // parameterize master's slots
      for (int i=0; i<masterSlots.sz(); i++)
      {
        Slot slot = (Slot)masterSlots.get(i);
        if (slot is Method)
        {
          slot = parameterize((Method)slot);
          m_methods.add(slot);
        }
        else
        {
          m_fields.add(slot);
        }
        m_slots.add(slot);
        m_slotsByName[slot.m_name.val] = slot;
      }
    }

    /**
     * Parameterize the specified method (if reuse if generic
     * parameterization isn't necessary).
     */
    internal Method parameterize(Method m)
    {
      // if not generic, short circuit and reuse original
      if (!m.isGenericMethod()) return m;

      // new signature
      Type ret;
      List pars = new List(Sys.MethodParamType, m.m_pars.sz());

      // parameterize return type
      if (m.m_returnType.isGenericParameter())
        ret = parameterize(m.m_returnType);
      else
        ret = m.m_returnType;

      // narrow pars (or just reuse if not parameterized)
      for (int i=0; i<m.m_pars.sz(); i++)
      {
        MethodParam p = (MethodParam)m.m_pars.get(i);
        if (p.m_paramType.isGenericParameter())
        {
          pars.add(new MethodParam(p.m_name, parameterize(p.m_paramType), p.m_mask));
        }
        else
        {
          pars.add(p);
        }
      }

      return new Method(this, m.m_name, m.m_flags, m.m_facets, ret, m.m_inheritedReturnType, pars, m);
    }

    /**
     * Parameterize t, where t is a generic parameter type such as V.
     */
    internal Type parameterize(Type t)
    {
      if (t is ListType)
        return parameterizeListType((ListType)t);
      else if (t is MethodType)
        return parameterizeMethodType((MethodType)t);
      else
        return doParameterize(t);
    }

    /**
     * Recursively parameterize the a generic list type.
     */
    internal Type parameterizeListType(ListType t)
    {
      return doParameterize(t.m_v).toListOf();
    }

    /**
     * Recursively parameterize the pars of a method type.
     */
    internal MethodType parameterizeMethodType(MethodType t)
    {
      Type[] pars = new Type[t.m_pars.Length];
      for (int i=0; i<pars.Length; i++)
      {
        Type param = t.m_pars[i];
        if (param.isGenericParameter()) param = doParameterize(param);
        pars[i] = param;
      }

      Type ret = t.m_ret;
      if (ret.isGenericParameter()) ret = doParameterize(ret);

      return new MethodType(pars, ret);
    }

    /**
     * Parameterize t, where t is a generic parameter type such as V.
     */
    protected abstract Type doParameterize(Type t);

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    private Map m_pars;
  }

}