//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   28 Jan 07  Andy Frank  Creation
//

using FileInfo = System.IO.FileInfo;
using FileStream = System.IO.FileStream;
using FileSystemInfo = System.IO.FileSystemInfo;
using DirectoryInfo  = System.IO.DirectoryInfo;

namespace Fan.Sys
{
  /// <summary>
  /// LocalFile represents a file or directory in the local file system.
  /// </summary>
  public class LocalFile : File
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Uri toUri(FileSystemInfo file)
    {
      string path = file.FullName;
      path = path.Replace('\\', '/');
      if (file is DirectoryInfo)
      {
        // add trailing slash if not present
        if (path.Length == 0 || path[path.Length-1] != '/')
          path = path + "/";
      }

      // deal with Windoze drive name
      if (path.Length > 2 && path[1] == ':' && path[0] != '/')
        path = "/" + path;

      Uri uri = Uri.parse(path);
      return uri;
    }

    public static FileSystemInfo toFile(Uri uri)
    {
      // Convert to Windows path - TODO - what about Mono?
      string path = uri.pathStr().val;

      // remove starting slash
      if (path.Length > 3 && path[2] == ':' && path[0] == '/')
        path = path.Substring(1);

      // remove trailing slash
      if (path[path.Length-1] == '/')
        path = path.Substring(0, path.Length-1);

      // use backslashes
      path = path.Replace('/', '\\');

      if (System.IO.File.Exists(path)) return new FileInfo(path);
      if (System.IO.Directory.Exists(path)) return new DirectoryInfo(path);

      if (uri.isDir().val) return new DirectoryInfo(path);
      else return new FileInfo(path);
    }

    public LocalFile(Uri uri) : this(uri, toFile(uri)) {}

    public LocalFile(FileSystemInfo file): this(toUri(file), file) {}

    public LocalFile(Uri uri, FileSystemInfo file) : base(uri)
    {
      this.m_file = file;
      if (m_file.Exists)
      {
        if (m_file is DirectoryInfo)
        {
          if (!uri.isDir().val)
            throw IOErr.make("Must use trailing slash for dir: " + uri).val;
        }
        else
        {
          if (uri.isDir().val)
            throw IOErr.make("Cannot use trailing slash for file: " + uri).val;
        }
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Obj
  //////////////////////////////////////////////////////////////////////////

    public override Type type() { return Sys.LocalFileType; }

  //////////////////////////////////////////////////////////////////////////
  // File
  //////////////////////////////////////////////////////////////////////////

    public override Bool exists()
    {
      return m_file.Exists ? Bool.True : Bool.False;
    }

    public override Int size()
    {
      if (m_file is DirectoryInfo) return null;
      m_file.Refresh();
      return Int.make((m_file as FileInfo).Length);
    }

    public override Time modified()
    {
      return Time.net(m_file.LastAccessTime);
    }

    public override void modified(Time time)
    {
      m_file.LastAccessTime = time.net();
    }

    public override Str osPath()
    {
      return Str.make(m_file.FullName);
    }

    public override File parent()
    {
      Uri parent = m_uri.parent();
      if (parent == null) return null;
      return new LocalFile(parent);
    }

    public override List list()
    {
      int len = 0;
      FileSystemInfo[] list = null;
      if (m_file is DirectoryInfo)
      {
        list = (m_file as DirectoryInfo).GetFileSystemInfos();
        len = list.Length;
      }
      List acc = new List(Sys.FileType, len);
      for (int i=0; i<len; i++)
        acc.add(new LocalFile(list[i]));
      return acc;
    }

    public override File normalize()
    {
      //try
      //{
        return this; //new LocalFile(file.getCanonicalFile());
      //}
      //catch (System.IO.IOException e)
      //{
      //  throw IOErr.make(e).val;
      //}
    }

    public override File plus(Uri uri)
    {
      return make(m_uri.plus(uri));
    }

  //////////////////////////////////////////////////////////////////////////
  // File Management
  //////////////////////////////////////////////////////////////////////////

    public override File create()
    {
      if (isDir().val)
        createDir();
      else
        createFile();
      return this;
    }

    private void createFile()
    {
      if (m_file.Exists)
      {
        if (m_file is DirectoryInfo)
          throw IOErr.make("Already exists as dir: " + m_file).val;
      }
      else
      {
        try
        {
          // TODO - not sure how this should work yet
          FileStream fs = (m_file as FileInfo).Create();
          fs.Close();
          m_file.Refresh();
        }
        catch (System.IO.IOException e)
        {
          throw IOErr.make(e).val;
        }
      }
    }

    private void createDir()
    {
      if (m_file.Exists)
      {
        if (!(m_file is DirectoryInfo))
          throw IOErr.make("Already exists as file: " + m_file).val;
      }
      else
      {
        try
        {
          System.IO.Directory.CreateDirectory(m_file.FullName);
          m_file.Refresh();
        }
        catch (System.Exception e)
        {
          throw IOErr.make(e).val;
        }
      }
    }

    public override void delete()
    {
      if (!exists().val) return;

      if (m_file is DirectoryInfo)
      {
        List kids = list();
        for (int i=0; i<kids.sz(); i++)
          (kids.get(i) as File).delete();
      }

      try
      {
        m_file.Delete();
        m_file.Refresh();
      }
      catch (System.Exception)
      {
        throw IOErr.make("Cannot delete: " + m_file).val;
      }
    }

    public override File deleteOnExit()
    {
/*
      file.deleteOnExit();
*/
      return this;
    }

  //////////////////////////////////////////////////////////////////////////
  // IO
  //////////////////////////////////////////////////////////////////////////

    public override InStream @in(Int bufSize)
    {
      try
      {
        System.IO.Stream stream = (m_file as FileInfo).OpenRead();
        return SysInStream.make(stream, bufSize);
      }
      catch (System.IO.IOException e)
      {
        throw IOErr.make(e).val;
      }
    }

    public override OutStream @out(Bool append, Int bufSize)
    {
      try
      {
        System.IO.Stream stream = (m_file as FileInfo).Open(
          append.val ? System.IO.FileMode.Append : System.IO.FileMode.Create,
          System.IO.FileAccess.Write);
        m_file.Refresh();
        return SysOutStream.make(stream, bufSize);
      }
      catch (System.IO.IOException e)
      {
        throw IOErr.make(e).val;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    internal FileSystemInfo m_file = null;

  }
}