//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Sep 06  Andy Frank  Creation
//

using System;
using System.Reflection;
using System.Text;
using Fanx.Fcode;
using Fanx.Util;

namespace Fan.Sys
{
  /// <summary>
  /// Method is an invocable operation on a Type.
  /// </summary>
  public class Method : Slot
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Constructor used by Type.reflect.
    /// </summary>
    public Method(Type parent, Str name, int flags, Facets facets, Type returnType, Type inheritedReturnType, List pars)
      : base(parent, name, flags, facets)
    {
      this.m_returnType = returnType;
      this.m_inheritedReturnType = inheritedReturnType;
      this.m_pars       = pars;
      this.m_mask       = toMask(parent, returnType, pars);
      this.m_generic    = null;
    }

    /// <summary>
    /// Constructor used by GenericType and we are given the generic
    /// method that is being parameterized.
    /// </summary>
    public Method(Type parent, Str name, int flags, Facets facets, Type returnType, Type inheritedReturnType, List pars, Method generic)
      : base(parent, name, flags, facets)
    {
      this.m_returnType = returnType;
      this.m_inheritedReturnType = inheritedReturnType;
      this.m_pars       = pars;
      this.m_mask       = 0;
      this.m_generic    = generic;
    }

    /// <summary>
    /// Compute if the method signature contains generic parameter types.
    /// </summary>
    private static int toMask(Type parent, Type returnType, List pars)
    {
      // we only use generics in Sys
      if (parent.pod() != Sys.SysPod) return 0;

      int p = returnType.isGenericParameter() ? 1 : 0;
      for (int i=0; i<pars.sz(); i++)
        p |= ((MethodParam)pars.get(i)).m_paramType.isGenericParameter() ? 1 : 0;

      int mask = 0;
      if (p != 0) mask |= GENERIC;
      return mask;
    }

  //////////////////////////////////////////////////////////////////////////
  // Methods
  //////////////////////////////////////////////////////////////////////////

    public override Type type() { return Sys.MethodType; }

    public Type returnType() { return m_returnType; }

    public List @params() { return m_pars.ro(); }

    public override Str signature()
    {
      StringBuilder s = new StringBuilder();
      s.Append(returnType()).Append(' ').Append(m_name).Append('(');
      for (int i=0; i<m_pars.sz(); i++)
      {
        if (i > 0) s.Append(", ");
        MethodParam p = (MethodParam)m_pars.get(i);
        s.Append(p.paramType()).Append(' ').Append(p.name());
      }
      s.Append(')');
      return Str.make(s.ToString());
    }

  //////////////////////////////////////////////////////////////////////////
  // Generics
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Return if this method contains generic parameters in it's signature.
    /// </summary>
    public bool isGenericMethod()
    {
      return (m_mask & GENERIC) != 0;
    }

    /// <summary>
    /// Return if this method is the parameterization of a generic method,
    /// with all the generic parameters filled in with real types.
    /// </summary>
    public bool isGenericInstance()
    {
      return m_generic != null;
    }

    /// <summary>
    /// If isGenericInstance is true, then return the generic method which
    /// this method instantiates.  The generic method may be used to access
    /// the actual signatures used in the Java code (via getRawType).  If
    /// this method is not a generic instance, return null.
    /// </summary>
    public Method getGenericMethod()
    {
      return m_generic;
    }

  //////////////////////////////////////////////////////////////////////////
  // Invoke
  //////////////////////////////////////////////////////////////////////////

    // TODO - support for calling ctor via reflection
    public virtual Obj call(List args)
    {
      int argsSize = args == null ? 0 : args.sz();

      bool isStatic = (m_flags & (FConst.Static|FConst.Ctor)) != 0;
      int p = checkArgs(argsSize, isStatic);
      Obj[] a = new Obj[p];

      if (isStatic)
      {
        if (args != null && a.Length > 0) args.toArray(a, 0, a.Length);
        return invoke(null, a);
      }
      else
      {
        Obj i = args.get(0);
        if (a.Length > 0) args.toArray(a, 1, a.Length);
        return invoke(i, a);
      }

    }

    public Obj callOn(Obj target, List args)
    {
      int argsSize = args == null ? 0 : args.sz();

      // from checkArgs perspective we pass true for isStatic
      // because we don't include target as part of arguments
      int p = checkArgs(argsSize, true);

      Obj[] a = new Obj[p];
      if (args != null && a.Length > 0) args.toArray(a, 0, a.Length);
      return invoke(target, a);
    }

    public virtual Obj call0()
    {
      bool isStatic = (m_flags & (FConst.Static|FConst.Ctor)) != 0;
      checkArgs(0, isStatic);
      return invoke(null, noArgs);
    }

    public virtual Obj call1(Obj a)
    {
      bool isStatic = (m_flags & (FConst.Static|FConst.Ctor)) != 0;
      int p = checkArgs(1, isStatic);
      Obj[] args;
      if (isStatic)
      {
        args = new Obj[p];
        if (p > 0) args[0] = a;
        return invoke(null, args);
      }
      else
      {
        args = new Obj[p];
        return invoke(a, args);
      }
    }

    public virtual Obj call2(Obj a, Obj b)
    {
      bool isStatic = (m_flags & (FConst.Static|FConst.Ctor)) != 0;
      int p = checkArgs(2, isStatic);
      Obj[] args;
      if (isStatic)
      {
        args = new Obj[p];
        if (p > 0) args[0] = a;
        if (p > 1) args[1] = b;
        return invoke(null, args);
      }
      else
      {
        args = new Obj[p];
        if (p > 0) args[0] = b;
        return invoke(a, args);
      }
    }

    public virtual Obj call3(Obj a, Obj b, Obj c)
    {
      bool isStatic = (m_flags & (FConst.Static|FConst.Ctor)) != 0;
      int p = checkArgs(3, isStatic);
      Obj[] args;
      if (isStatic)
      {
        args = new Obj[p];
        if (p > 0) args[0] = a;
        if (p > 1) args[1] = b;
        if (p > 2) args[2] = c;
        return invoke(null, args);
      }
      else
      {
        args = new Obj[p];
        if (p > 0) args[0] = b;
        if (p > 1) args[1] = c;
        return invoke(a, args);
      }
    }

    public virtual Obj call4(Obj a, Obj b, Obj c, Obj d)
    {
      bool isStatic = (m_flags & (FConst.Static|FConst.Ctor)) != 0;
      int p = checkArgs(4, isStatic);
      Obj[] args;
      if (isStatic)
      {
        args = new Obj[p];
        if (p > 0) args[0] = a;
        if (p > 1) args[1] = b;
        if (p > 2) args[2] = c;
        if (p > 3) args[3] = d;
        return invoke(null, args);
      }
      else
      {
        args = new Obj[p];
        if (p > 0) args[0] = b;
        if (p > 1) args[1] = c;
        if (p > 2) args[2] = d;
        return invoke(a, args);
      }
    }

    public virtual Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e)
    {
      bool isStatic = (m_flags & (FConst.Static|FConst.Ctor)) != 0;
      int p = checkArgs(5, isStatic);
      Obj[] args;
      if (isStatic)
      {
        args = new Obj[p];
        if (p > 0) args[0] = a;
        if (p > 1) args[1] = b;
        if (p > 2) args[2] = c;
        if (p > 3) args[3] = d;
        if (p > 4) args[4] = e;
        return invoke(null, args);
      }
      else
      {
        args = new Obj[p];
        if (p > 0) args[0] = b;
        if (p > 1) args[1] = c;
        if (p > 2) args[2] = d;
        if (p > 3) args[3] = e;
        return invoke(a, args);
      }
    }

    public virtual Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f)
    {
      bool isStatic = (m_flags & (FConst.Static|FConst.Ctor)) != 0;
      int p = checkArgs(6, isStatic);
      Obj[] args;
      if (isStatic)
      {
        args = new Obj[p];
        if (p > 0) args[0] = a;
        if (p > 1) args[1] = b;
        if (p > 2) args[2] = c;
        if (p > 3) args[3] = d;
        if (p > 4) args[4] = e;
        if (p > 5) args[5] = f;
        return invoke(null, args);
      }
      else
      {
        args = new Obj[p];
        if (p > 0) args[0] = b;
        if (p > 1) args[1] = c;
        if (p > 2) args[2] = d;
        if (p > 3) args[3] = e;
        if (p > 4) args[4] = f;
        return invoke(a, args);
      }
    }

    public virtual Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g)
    {
      bool isStatic = (m_flags & (FConst.Static|FConst.Ctor)) != 0;
      int p = checkArgs(7, isStatic);
      Obj[] args;
      if (isStatic)
      {
        args = new Obj[p];
        if (p > 0) args[0] = a;
        if (p > 1) args[1] = b;
        if (p > 2) args[2] = c;
        if (p > 3) args[3] = d;
        if (p > 4) args[4] = e;
        if (p > 5) args[5] = f;
        if (p > 6) args[6] = g;
        return invoke(null, args);
      }
      else
      {
        args = new Obj[p];
        if (p > 0) args[0] = b;
        if (p > 1) args[1] = c;
        if (p > 2) args[2] = d;
        if (p > 3) args[3] = e;
        if (p > 4) args[4] = f;
        if (p > 5) args[5] = g;
        return invoke(a, args);
      }
    }

    public virtual Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h)
    {
      bool isStatic = (m_flags & (FConst.Static|FConst.Ctor)) != 0;
      int p = checkArgs(8, isStatic);
      Obj[] args;
      if (isStatic)
      {
        args = new Obj[p];
        if (p > 0) args[0] = a;
        if (p > 1) args[1] = b;
        if (p > 2) args[2] = c;
        if (p > 3) args[3] = d;
        if (p > 4) args[4] = e;
        if (p > 5) args[5] = f;
        if (p > 6) args[6] = g;
        if (p > 7) args[7] = h;
        return invoke(null, args);
      }
      else
      {
        args = new Obj[p];
        if (p > 0) args[0] = b;
        if (p > 1) args[1] = c;
        if (p > 2) args[2] = d;
        if (p > 3) args[3] = e;
        if (p > 4) args[4] = f;
        if (p > 5) args[5] = g;
        if (p > 6) args[6] = h;
        return invoke(a, args);
      }
    }

    private int checkArgs(int args, bool isStatic)
    {
      // ensure parent has finished emitting so that reflect is populated
      m_parent.finish();

      // compuate min/max parameters - reflect contains all the method versions
      // with full params at index zero, and full defaults at reflect.length-1
      int max = m_pars.sz();
      int min = max-m_reflect.Length+1;

      // do checking
      if (isStatic)
      {
        if (args < min) throw ArgErr.make("Too few arguments: " + args + " < " + min+".."+max).val;
      }
      else
      {
        if (args < min+1) throw ArgErr.make("Too few arguments: " + args + " < instance+" + min+".."+max).val;
      }

      // return size of arguments to pass to java method
      if (!isStatic) args--;
      if (args <= max) return args;
      return max;
    }

  //////////////////////////////////////////////////////////////////////////
  // Reflection
  //////////////////////////////////////////////////////////////////////////

    internal Obj invoke(object instance, object[] args)
    {
      try
      {
        BindingFlags flags = BindingFlags.Public | BindingFlags.InvokeMethod;
        if ((this.m_flags & (FConst.Static|FConst.Ctor)) != 0)
          flags |= BindingFlags.Static;
        else
          flags |= BindingFlags.Instance;

        System.Type t = (instance == null) ? m_parent.emit() : instance.GetType();
        return (Obj)t.InvokeMember(m_name.val, flags, null, instance, args);
      }
      catch (TargetInvocationException e)
      {
        if (e.GetBaseException() is Err.Val)
          throw (Err.Val)e.GetBaseException();
        else
          throw Err.make(e.GetBaseException()).val;
      }
      catch (Exception e)
      {
        if (m_reflect == null)
          throw Err.make("Method not mapped to System.Reflection correctly").val;

        //System.Console.WriteLine(e.StackTrace);
        throw Err.make("Cannot call '" + this + "': " + e).val;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Indirect
  //////////////////////////////////////////////////////////////////////////

    public static readonly int MaxIndirectParams = 8;  // max callX()

    /// <summary>
    /// Constructor used by Indirect.
    /// </summary>
    protected Method(MethodType methodType)
      : base(null, null, FConst.Internal, null)
    {
      this.m_returnType = methodType.m_ret;
      this.m_pars       = methodType.toMethodParams();
      this.m_mask       = 0;
      this.m_generic    = null;
    }

    /// <summary>
    /// Indirect is the base class for the IndirectX classes, which are
    /// used as the common base classes for closures and curried methods.
    /// An Indirect method takes a MethodType for it's type, and
    /// also extends Method with the call() implementation.
    /// </summary>
    public abstract class Indirect : Method
    {
      protected Indirect(MethodType type) : base(type) { this.m_type = type; }

      public override Str  name()  { return Str.make(GetType().Name); }
      public override Type type()  { return m_type; }
      public override Str  toStr() { return m_type.signature(); }
      public Err tooFewArgs(int given) { throw Err.make("Too few arguments: " + given + " < " + m_type.m_pars.Length).val; }

      MethodType m_type;
    }

    public abstract class Indirect0 : Indirect
    {
      protected Indirect0(MethodType type) : base(type) {}
      public override Obj call(List args) { return call0(); }
      public override abstract Obj call0();
      public override Obj call1(Obj a) { return call0(); }
      public override Obj call2(Obj a, Obj b) { return call0(); }
      public override Obj call3(Obj a, Obj b, Obj c) { return call0(); }
      public override Obj call4(Obj a, Obj b, Obj c, Obj d) { return call0(); }
      public override Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return call0(); }
      public override Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return call0(); }
      public override Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call0(); }
      public override Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call0(); }
    }

    public abstract class Indirect1 : Indirect
    {
      protected Indirect1(MethodType type) : base(type) {}
      public override Obj call(List args) { return call1(args.get(0)); }
      public override Obj call0() { throw Err.make("Too few arguments: 0 < 1").val; }
      public override abstract Obj call1(Obj a);
      public override Obj call2(Obj a, Obj b) { return call1(a); }
      public override Obj call3(Obj a, Obj b, Obj c) { return call1(a); }
      public override Obj call4(Obj a, Obj b, Obj c, Obj d) { return call1(a); }
      public override Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return call1(a); }
      public override Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return call1(a); }
      public override Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call1(a); }
      public override Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call1(a); }
    }

    public abstract class Indirect2 : Indirect
    {
      protected Indirect2(MethodType type) : base(type) {}
      public override Obj call(List args) { return call2(args.get(0), args.get(1)); }
      public override Obj call0() { throw Err.make("Too few arguments: 0 < 2").val; }
      public override Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 2").val; }
      public override abstract Obj call2(Obj a, Obj b);
      public override Obj call3(Obj a, Obj b, Obj c) { return call2(a, b); }
      public override Obj call4(Obj a, Obj b, Obj c, Obj d) { return call2(a, b); }
      public override Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return call2(a, b); }
      public override Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return call2(a, b); }
      public override Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call2(a, b); }
      public override Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call2(a, b); }
    }

    public abstract class Indirect3 : Indirect
    {
      protected Indirect3(MethodType type) : base(type) {}
      public override Obj call(List args) { return call3(args.get(0), args.get(1), args.get(2)); }
      public override Obj call0() { throw Err.make("Too few arguments: 0 < 3").val; }
      public override Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 3").val; }
      public override Obj call2(Obj a, Obj b)  { throw Err.make("Too few arguments: 2 < 3").val; }
      public override abstract Obj call3(Obj a, Obj b, Obj c);
      public override Obj call4(Obj a, Obj b, Obj c, Obj d) { return call3(a, b, c); }
      public override Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return call3(a, b, c); }
      public override Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return call3(a, b, c); }
      public override Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call3(a, b, c); }
      public override Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call3(a, b, c); }
    }

    public abstract class Indirect4 : Indirect
    {
      protected Indirect4(MethodType type) : base(type) {}
      public override Obj call(List args) { return call4(args.get(0), args.get(1), args.get(2), args.get(3)); }
      public override Obj call0() { throw Err.make("Too few arguments: 0 < 4").val; }
      public override Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 4").val; }
      public override Obj call2(Obj a, Obj b)  { throw Err.make("Too few arguments: 2 < 4").val; }
      public override Obj call3(Obj a, Obj b, Obj c) { throw Err.make("Too few arguments: 3 < 4").val; }
      public override abstract Obj call4(Obj a, Obj b, Obj c, Obj d);
      public override Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { return call4(a, b, c, d); }
      public override Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return call4(a, b, c, d); }
      public override Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call4(a, b, c, d); }
      public override Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call4(a, b, c, d); }
    }

    public abstract class Indirect5 : Indirect
    {
      protected Indirect5(MethodType type) : base(type) {}
      public override Obj call(List args) { return call5(args.get(0), args.get(1), args.get(2), args.get(3), args.get(4)); }
      public override Obj call0() { throw Err.make("Too few arguments: 0 < 5").val; }
      public override Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 5").val; }
      public override Obj call2(Obj a, Obj b)  { throw Err.make("Too few arguments: 2 < 5").val; }
      public override Obj call3(Obj a, Obj b, Obj c) { throw Err.make("Too few arguments: 3 < 5").val; }
      public override Obj call4(Obj a, Obj b, Obj c, Obj d) { throw Err.make("Too few arguments: 4 < 5").val; }
      public override abstract Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e);
      public override Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { return call5(a, b, c, d, e); }
      public override Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call5(a, b, c, d, e); }
      public override Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call5(a, b, c, d, e); }
    }

    public abstract class Indirect6 : Indirect
    {
      protected Indirect6(MethodType type) : base(type) {}
      public override Obj call(List args) { return call6(args.get(0), args.get(1), args.get(2), args.get(3), args.get(4), args.get(5)); }
      public override Obj call0() { throw Err.make("Too few arguments: 0 < 6").val; }
      public override Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 6").val; }
      public override Obj call2(Obj a, Obj b)  { throw Err.make("Too few arguments: 2 < 6").val; }
      public override Obj call3(Obj a, Obj b, Obj c) { throw Err.make("Too few arguments: 3 < 6").val; }
      public override Obj call4(Obj a, Obj b, Obj c, Obj d) { throw Err.make("Too few arguments: 4 < 6").val; }
      public override Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { throw Err.make("Too few arguments: 5 < 6").val; }
      public override abstract Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f);
      public override Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { return call6(a, b, c, d, e, f); }
      public override Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call6(a, b, c, d, e, f); }
    }

    public abstract class Indirect7 : Indirect
    {
      protected Indirect7(MethodType type) : base(type) {}
      public override Obj call(List args) { return call7(args.get(0), args.get(1), args.get(2), args.get(3), args.get(4), args.get(5), args.get(6)); }
      public override Obj call0() { throw Err.make("Too few arguments: 0 < 7").val; }
      public override Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 7").val; }
      public override Obj call2(Obj a, Obj b)  { throw Err.make("Too few arguments: 2 < 7").val; }
      public override Obj call3(Obj a, Obj b, Obj c) { throw Err.make("Too few arguments: 3 < 7").val; }
      public override Obj call4(Obj a, Obj b, Obj c, Obj d) { throw Err.make("Too few arguments: 4 < 7").val; }
      public override Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { throw Err.make("Too few arguments: 5 < 7").val; }
      public override Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { throw Err.make("Too few arguments: 6 < 7").val; }
      public override abstract Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g);
      public override Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { return call7(a, b, c, d, e, f, g); }
    }

    public abstract class Indirect8 : Indirect
    {
      protected Indirect8(MethodType type) : base(type) {}
      public override Obj call(List args) { return call8(args.get(0), args.get(1), args.get(2), args.get(3), args.get(4), args.get(5), args.get(6), args.get(7)); }
      public override Obj call0() { throw Err.make("Too few arguments: 0 < 8").val; }
      public override Obj call1(Obj a) { throw Err.make("Too few arguments: 1 < 8").val; }
      public override Obj call2(Obj a, Obj b)  { throw Err.make("Too few arguments: 2 < 8").val; }
      public override Obj call3(Obj a, Obj b, Obj c) { throw Err.make("Too few arguments: 3 < 8").val; }
      public override Obj call4(Obj a, Obj b, Obj c, Obj d) { throw Err.make("Too few arguments: 4 < 8").val; }
      public override Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { throw Err.make("Too few arguments: 5 < 8").val; }
      public override Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { throw Err.make("Too few arguments: 6 < 8").val; }
      public override Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { throw Err.make("Too few arguments: 7 < 8").val; }
      public override abstract Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h);
    }

    public abstract class IndirectX : Indirect
    {
      protected IndirectX(MethodType type) : base(type) {}
      public override abstract Obj call(List args);
      public override Obj call0() { throw tooFewArgs(0).val; }
      public override Obj call1(Obj a) { throw tooFewArgs(1).val; }
      public override Obj call2(Obj a, Obj b)  { throw tooFewArgs(2).val; }
      public override Obj call3(Obj a, Obj b, Obj c) { throw tooFewArgs(3).val; }
      public override Obj call4(Obj a, Obj b, Obj c, Obj d) { throw tooFewArgs(4).val; }
      public override Obj call5(Obj a, Obj b, Obj c, Obj d, Obj e) { throw tooFewArgs(5).val; }
      public override Obj call6(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f) { throw tooFewArgs(6).val; }
      public override Obj call7(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g) { throw tooFewArgs(7).val; }
      public override Obj call8(Obj a, Obj b, Obj c, Obj d, Obj e, Obj f, Obj g, Obj h) { throw tooFewArgs(8).val; }
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    static readonly int GENERIC = 0x01;  // is this a generic method
    static readonly object[] noArgs = new object[0];

    internal readonly Type m_returnType;
    internal readonly Type m_inheritedReturnType; // for covariance
    internal readonly List m_pars;
    internal readonly int m_mask;
    internal readonly Method m_generic;
    internal MethodInfo[] m_reflect;

  }
}