//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Dec 06  Andy Frank  Creation
//

using System.Text;

namespace Fan.Sys
{
  /// <summary>
  /// MethodType is a parameterized type for Methods.
  /// </summary>
  public class MethodType : GenericType
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public MethodType(Type[] pars, Type ret) : base(Sys.MethodType)
    {
      this.m_pars = pars;
      this.m_ret  = ret;

      // I am a generic parameter type if any my args or
      // return type are generic parameter types.
      this.m_genericParameterType |= ret.isGenericParameter();
      for (int i=0; i<pars.Length; i++)
        this.m_genericParameterType |= pars[i].isGenericParameter();
    }

  //////////////////////////////////////////////////////////////////////////
  // Type
  //////////////////////////////////////////////////////////////////////////

    public override Int hash()
    {
      return signature().hash();
    }

    public override Bool equals(Obj obj)
    {
      if (obj is MethodType)
      {
        MethodType x = (MethodType)obj;
        if (m_pars.Length != x.m_pars.Length) return Bool.False;
        for (int i=0; i<m_pars.Length; i++)
          if (!m_pars[i].equals(x.m_pars[i]).val) return Bool.False;
        return m_ret.equals(x.m_ret);
      }
      return Bool.False;
    }

    public override Type @base()
    {
      return Sys.MethodType;
    }

    public override Str signature()
    {
      if (m_sig == null)
      {
        StringBuilder s = new StringBuilder();
        s.Append('|');
        for (int i=0; i<m_pars.Length; i++)
        {
          if (i > 0) s.Append(',');
          s.Append(m_pars[i].signature().val);
        }
        s.Append('-').Append('>');
        s.Append(m_ret.signature().val);
        s.Append('|');
        m_sig = Str.make(s.ToString());
      }
      return m_sig;
    }

    public override bool @is(Type type)
    {
      if (this == type) return true;
      if (type is MethodType)
      {
        MethodType t = (MethodType)type;

        // match return type
        if (!m_ret.@is(t.m_ret)) return false;

        // match pars - it is ok for me to have less than
        // the type pars (if I want to ignore them), but I
        // must have no more
        if (m_pars.Length > t.m_pars.Length) return false;
        for (int i=0; i<m_pars.Length; i++)
          if (!t.m_pars[i].@is(m_pars[i])) return false;

        // this method works for the specified method type
        return true;
      }
      return @base().@is(type);
    }

    internal override Map makeParams()
    {
      Map map = new Map(Sys.StrType, Sys.TypeType);
      for (int i=0; i<m_pars.Length; i++)
        map.set(Str.m_ascii['A'+i], m_pars[i]);
      return map.set(Str.m_ascii['R'], m_ret).ro();
    }

  //////////////////////////////////////////////////////////////////////////
  // GenericType
  //////////////////////////////////////////////////////////////////////////

    public override Type getRawType()
    {
      return Sys.MethodType;
    }

    public override bool isGenericParameter()
    {
      return m_genericParameterType;
    }

    protected override Type doParameterize(Type t)
    {
      // return
      if (t == Sys.RType) return m_ret;

      // if A-H maps to avail pars
      int name = t.m_name.val[0] - 'A';
      if (name < m_pars.Length) return m_pars[name];

      // otherwise let anything be used
      return Sys.ObjType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Method Support
  //////////////////////////////////////////////////////////////////////////

    internal List toMethodParams()
    {
      MethodParam[] p = new MethodParam[m_pars.Length];
      for (int i=0; i<p.Length; i++)
        p[i] = new MethodParam(Str.make("p" + i), m_pars[i], 0);
      return new List(Sys.MethodParamType, p);
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public readonly Type[] m_pars;
    public readonly Type m_ret;
    private Str m_sig;
    private bool m_genericParameterType;
  }
}