//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   13 Sep 06  Andy Frank  Creation
//

using System;
using System.Collections;
using System.Text;
using System.Text.RegularExpressions;

namespace Fan.Sys
{
  ///
  /// Str is a string of Unicode characters.
  ///
  public sealed class Str : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Str make(string val)
    {
      if (val == null) return null;
      if (val == "")   return Empty;
      return new Str(val);
    }

    public static Str makeTrim(StringBuilder s)
    {
      int start = 0;
      int end = s.Length;
      while (start < end) if (Int.isSpace(s[start])) start++; else break;
      while (end > start) if (Int.isSpace(s[end-1])) end--; else break;
      return make(s.ToString(start, end-start));
    }

    private Str(string val)
    {
      this.val = val;
    }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override bool Equals(Object obj)
    {
      return val == (obj as Str).val;
    }

    public override Bool equals(Obj obj)
    {
      if (obj is Str)
        return val == ((Str)obj).val ? Bool.True : Bool.False;
      else
        return Bool.False;
    }

    public Bool equalsIgnoreCase(Str s)
    {
      if (s == this) return Bool.True;

      string a = this.val;
      string b = s.val;
      int an   = a.Length;
      int bn   = b.Length;
      if (an != bn) return Bool.False;

      for (int i=0; i<an; i++)
      {
        int ac = a[i];
        int bc = b[i];
        if ('A' <= ac && ac <= 'Z') ac |= 0x20;
        if ('A' <= bc && bc <= 'Z') bc |= 0x20;
        if (ac != bc) return Bool.False;
      }
      return Bool.True;
    }

    public override Int compare(Obj obj)
    {
      int cmp = String.CompareOrdinal(val, ((Str)obj).val);
      if (cmp < 0) return Int.LT;
      return cmp == 0 ? Int.EQ : Int.GT;
    }

    public Int compareIgnoreCase(Str s)
    {
      if (s == this) return Int.Zero;

      string a = this.val;
      string b = s.val;
      int an   = a.Length;
      int bn   = b.Length;

      for (int i=0; i<an && i<bn; i++)
      {
        int ac = a[i];
        int bc = b[i];
        if ('A' <= ac && ac <= 'Z') ac |= 0x20;
        if ('A' <= bc && bc <= 'Z') bc |= 0x20;
        if (ac != bc) return ac < bc ? Int.LT : Int.GT;
      }

      if (an == bn) return Int.Zero;
      return an < bn ? Int.LT : Int.GT;
    }

    public override int GetHashCode()
    {
      return val.GetHashCode();
    }

    public override Int hash()
    {
      return Int.make(val.GetHashCode());
    }

    public int caseInsensitiveHash()
    {
      string val = this.val;
      int n = val.Length;
      int hash = 0;

      for (int i=0; i<n; i++)
      {
        int c = val[i];
        if ('A' <= c && c <= 'Z') c |= 0x20;
        hash = 31*hash + c;
      }

      return hash;
    }

    public override Str toStr()
    {
      return this;
    }

    public override Type type()
    {
      return Sys.StrType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Operators
  //////////////////////////////////////////////////////////////////////////

    public Int get(Int index)
    {
      int i = (int)index.val;
      if (i < 0) i = val.Length+i;
      return Int.pos(val[i]);
    }

    public Str slice(Range r)
    {
      int size = val.Length;

      int s = r.start(size);
      int e = r.end(size);
      if (e+1 < s) throw IndexErr.make(r).val;

      return make(val.Substring(s, (e-s)+1));
    }

    public Str plus(Obj obj)
    {
      if (obj == null) return make(String.Concat(val, "null"));
      Str x = obj.toStr();
      if (x.val == "") return this;
      return make(String.Concat(val, x.val));
    }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public Str intern()
    {
      Str s = (Str)interns[val];
      if (s == null) interns[val] = s = this;
      return s;
    }

    public Bool isEmpty()
    {
      return val.Length == 0 ? Bool.True : Bool.False;
    }

    public Int size()
    {
      return Int.pos(val.Length);
    }

    public Bool startsWith(Str s)
    {
      return Bool.make(val.StartsWith(s.val));
    }

    public Bool endsWith(Str s)
    {
      return Bool.make(val.EndsWith(s.val));
    }

    public Bool contains(Str s)
    {
      return index(s, Int.Zero) != null ? Bool.True : Bool.False;
    }

    public Int index(Str s) { return index(s, Int.Zero); }
    public Int index(Str s, Int off)
    {
      int i = (int)off.val;
      if (i < 0) i = val.Length+i;

      int r;
      string sval = s.val;
      if (sval.Length == 1)
        r = val.IndexOf(sval[0], i);
      else
        r = val.IndexOf(sval, i);

      if (r < 0) return null;
      return Int.make(r);
    }

    public Int indexr(Str s) { return indexr(s, Int.NegOne); }
    public Int indexr(Str s, Int off)
    {
      int i = (int)off.val;
      if (i < 0) i = val.Length+i;

      int r;
      String sval = s.val;
      if (sval.Length == 1)
        r = val.LastIndexOf(sval[0], i);
      else
        r = val.LastIndexOf(sval, i);

      if (r < 0) return null;
      return Int.make(r);
    }

  //////////////////////////////////////////////////////////////////////////
  // Iterators
  //////////////////////////////////////////////////////////////////////////

    public void each(Method m)
    {
      string val = this.val;
      int len = val.Length;
      for (int i=0; i<len ; i++)
        m.call2(Int.pos(val[i]), Int.pos(i));
    }

    public void eachr(Method m)
    {
      string val = this.val;
      for (int i=val.Length-1; i>=0; --i)
        m.call2(Int.pos(val[i]), Int.pos(i));
    }

    public Bool any(Method m)
    {
      string val = this.val;
      int len = val.Length;
      for (int i=0; i<len ; i++)
        if (m.call2(Int.pos(val[i]), Int.pos(i)) == Bool.True)
          return Bool.True;
      return Bool.False;
    }

    public Bool all(Method m)
    {
      string val = this.val;
      int len = val.Length;
      for (int i=0; i<len ; i++)
        if (m.call2(Int.pos(val[i]), Int.pos(i)) == Bool.False)
          return Bool.False;
      return Bool.True;
    }

  //////////////////////////////////////////////////////////////////////////
  // Utils
  //////////////////////////////////////////////////////////////////////////

    public static Str spaces(Int n)
    {
      // do an array lookup for reasonable length
      // strings since that is the common case
      int count = (int)n.val;
      try { return m_spaces[count]; } catch (IndexOutOfRangeException) {}

      // otherwise we build a new one
      StringBuilder s = new StringBuilder(m_spaces[m_spaces.Length-1].val);
      for (int i=m_spaces.Length-1; i<count; i++)
        s.Append(' ');
      return Str.make(s.ToString());
    }

    public Str lower() { return Str.make(lower(val)); }
    public static string lower(string val)
    {
      StringBuilder s = new StringBuilder(val.Length);
      for (int i=0; i<val.Length; i++)
      {
        int ch = val[i];
        if ('A' <= ch && ch <= 'Z') ch |= 0x20;
        s.Append((char)ch);
      }
      return s.ToString();
    }

    public Str upper()
    {
      string val = this.val;
      StringBuilder s = new StringBuilder(val.Length);
      for (int i=0; i<val.Length; i++)
      {
        int ch = val[i];
        if ('a' <= ch && ch <= 'z') ch &= ~0x20;
        s.Append((char)ch);
      }
      return make(s.ToString());
    }

    public Str justl(Int width)
    {
      string val = this.val;
      int w = (int)width.val;
      if (val.Length >= w) return this;
      StringBuilder s = new StringBuilder(w);
      s.Append(val);
      for (int i=val.Length; i<w; i++)
        s.Append(' ');
      return make(s.ToString());
    }

    public Str justr(Int width)
    {
      string val = this.val;
      int w = (int)width.val;
      if (val.Length >= w) return this;
      StringBuilder s = new StringBuilder(w);
      for (int i=val.Length; i<w; i++)
        s.Append(' ');
      s.Append(val);
      return make(s.ToString());
    }

    public Str reverse()
    {
      string val = this.val;
      if (val.Length < 2) return this;
      StringBuilder s = new StringBuilder(val.Length);
      for (int i=val.Length-1; i>=0; i--)
        s.Append(val[i]);
      return make(s.ToString());
    }

    public Str trim()
    {
      string val = this.val;
      int len = val.Length;
      if (len == 0) return this;
      if (val[0] > ' ' && val[len-1] > ' ') return this;
      return make(val.Trim());
    }

    public List split(Str seps)
    {
      // TODO?
      string[] split = Regex.Split(val, seps.val);
      Str[] str = new Str[split.Length];
      for (int i=0; i<str.Length; i++) str[i] = Str.make(split[i]);
      return new List(Sys.StrType, str);
    }

    /*
    public Str replace(Str from, Str to)
    {
      return make(StrUtil.Replace(val, from.val, to.val));
    }
    */

    public Bool isSpace()
    {
      string val = this.val;
      int len = val.Length;
      for (int i=0; i<len; i++)
      {
        int ch = val[i];
        if (ch >= 128 || (Int.charMap[ch] & Int.SPACE) == 0)
          return Bool.False;
      }
      return Bool.True;
    }

  //////////////////////////////////////////////////////////////////////////
  // Conversion
  //////////////////////////////////////////////////////////////////////////

    public Bool toBool() { return Bool.parse(this, Bool.True); }
    public Bool toBool(Bool check) { return Bool.parse(this, check); }

    public Int toInt() { return Int.parse(this, Int.Ten, Bool.True); }
    public Int toInt(Int radix) { return Int.parse(this, radix, Bool.True); }
    public Int toInt(Int radix, Bool check) { return Int.parse(this, radix, check); }

    public Float toFloat() { return Float.parse(this, Bool.True); }
    public Float toFloat(Bool check) { return Float.parse(this, check); }

    public Uri toUri() { return Uri.parse(this); }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public static readonly Str Empty = new Str("");
    public static readonly Str Null = new Str("null");

    private static readonly Hashtable interns = new Hashtable();

    internal static readonly Str[] m_ascii = new Str[128];
    internal static Str[] m_spaces = new Str[20];

    static Str()
    {
      // ascii
      for (int i=0; i<m_ascii.Length; i++)
        m_ascii[i] = make("" + (char)i).intern();

      // spaces
      StringBuilder s = new StringBuilder();
      for (int i=0; i<m_spaces.Length; i++)
      {
        m_spaces[i] = Str.make(s.ToString());
        s.Append(' ');
      }
    }

    public readonly string val;
  }
}