//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   14 Sep 06  Andy Frank  Creation
//

using System;
using System.Collections;
using System.IO;
using Fanx.Fcode;
using Fanx.Util;

namespace Fan.Sys
{
  ///
  /// Sys provides static access to the system's environment.
  ///
  public class Sys : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    private Sys() {}

  //////////////////////////////////////////////////////////////////////////
  // Environment
  //////////////////////////////////////////////////////////////////////////

    public static List args() { return m_args.ro(); }

    public static Map env() { return m_env; }

    public static Str hostName() { return m_hostName; }

    public static Str userName() { return m_userName; }

    public static Fan.Sys.File homeDir() { return m_homeDir; }

    public static void exit() { exit(Int.Zero); }
    public static void exit(Int status) { System.Environment.Exit((int)status.val); }

    /*
    public static InStream  in()  { return StdIn; }
    public static OutStream out() { return StdOut; }
    public static OutStream err() { return StdErr; }
    */

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public static readonly string HomeDir;
    public static readonly string PodsDir;
    public static readonly Pod SysPod;

    // are we using the bootstrap stub versions of sys types
    public static bool isStub;

    // if true then we disable FanClassLoader and all classes
    // must be available precompiled into bytecode from the
    // system's classloader (or webapp, etc); default is false
    public static bool usePrecompiledOnly = false;  // TODO

  //////////////////////////////////////////////////////////////////////////
  // Boot
  //////////////////////////////////////////////////////////////////////////

    static Sys()
    {
      try
      {
        // get fan home directory
        HomeDir = Environment.GetEnvironmentVariable("fan_home");
        if (HomeDir == null)
          throw new Exception("Missing fan_home environment variable.");
        if (!Directory.Exists(HomeDir))
          throw new Exception("Invalid fan_home dir: " + HomeDir);
        PodsDir = HomeDir + "/lib/fan";

        // load sys pod
        SysPod = new Pod("sys");
        Pod.m_podsByName["sys"] = SysPod;
        try
        {
          SysPod.load(Pod.readFPod("sys"));
          isStub = false;
        }
        catch (Exception e)
        {
          Console.WriteLine("WARNING: Using stub for 'sys' pod");
          Console.WriteLine(e.StackTrace);
          Sys.isStub = true;
        }

        // types
        ObjType = builtin("Obj", null);

        // basic primitives
        NumType      = builtin("Num",      ObjType);
        BoolType     = builtin("Bool",     ObjType);
        CharsetType  = builtin("Charset",  ObjType);
        DurationType = builtin("Duration", ObjType);
        IntType      = builtin("Int",      NumType);
        EnumType     = builtin("Enum",     ObjType);
        FloatType    = builtin("Float",    NumType);
        ListType     = builtin("List",     ObjType);
        MapType      = builtin("Map",      ObjType);
        MonthType    = builtin("Month",    ObjType);
        PodType      = builtin("Pod",      ObjType);
        RangeType    = builtin("Range",    ObjType);
        StrType      = builtin("Str",      ObjType);
        StrBufType   = builtin("StrBuf",   ObjType);
        TestType     = builtin("Test",     ObjType);
        TimeType     = builtin("Time",     ObjType);
        TypeType     = builtin("Type",     ObjType);
        UriType      = builtin("Uri",      ObjType);
        VoidType     = builtin("Void",     ObjType);
        WeekdayType  = builtin("Weekday",  ObjType);

        // reflection
        SlotType        = builtin("Slot",        ObjType);
        FieldType       = builtin("Field",       SlotType);
        MethodType      = builtin("Method",      SlotType);
        MethodParamType = builtin("MethodParam", ObjType);
        FacetsType      = builtin("Facets",      ObjType);

        // IO
        InStreamType     = builtin("InStream",     ObjType);
        SysInStreamType  = builtin("SysInStream",  ObjType);
        OutStreamType    = builtin("OutStream",    ObjType);
        SysOutStreamType = builtin("SysOutStream", ObjType);
        BufType          = builtin("Buf",          ObjType);
        FileType         = builtin("File",         ObjType);
        LocalFileType    = builtin("LocalFile",    FileType);
        ZipEntryFileType = builtin("ZipEntryFile", FileType);
        ZipType          = builtin("Zip",          ObjType);

        // utils
        //ProcessType     = builtin("Process",  ObjType);
        DependType      = builtin("Depend",   ObjType);
        VersionType     = builtin("Version",  ObjType);

        // exceptions
        ErrType             = builtin("Err",             ObjType);
        ArgErrType          = builtin("ArgErr",          ErrType);
        CastErrType         = builtin("CastErr",         ErrType);
        IndexErrType        = builtin("IndexErr",        ErrType);
        IOErrType           = builtin("IOErr",           ErrType);
        NotImmutableErrType = builtin("NotImmutableErr", ErrType);
        NullErrType         = builtin("NullErr",         ErrType);
        ReadonlyErrType     = builtin("ReadonlyErr",     ErrType);
        UnknownPodErrType   = builtin("UnknownPodErr",   ErrType);
        UnknownSlotErrType  = builtin("UnknownSlotErr",  ErrType);
        UnknownTypeErrType  = builtin("UnknownTypeErr",  ErrType);
        UnsupportedErrType  = builtin("UnsupportedErr",  ErrType);

        // generic types
        GenericParameterTypes['A'] = AType = new Type(SysPod, "A", 0, null);  // A-H Params
        GenericParameterTypes['B'] = BType = new Type(SysPod, "B", 0, null);
        GenericParameterTypes['C'] = CType = new Type(SysPod, "C", 0, null);
        GenericParameterTypes['D'] = DType = new Type(SysPod, "D", 0, null);
        GenericParameterTypes['E'] = EType = new Type(SysPod, "E", 0, null);
        GenericParameterTypes['F'] = FType = new Type(SysPod, "F", 0, null);
        GenericParameterTypes['G'] = GType = new Type(SysPod, "G", 0, null);
        GenericParameterTypes['H'] = HType = new Type(SysPod, "H", 0, null);
        GenericParameterTypes['K'] = KType = new Type(SysPod, "K", 0, null);  // Key
        GenericParameterTypes['L'] = LType = new Type(SysPod, "L", 0, null);  // Variegated List
        GenericParameterTypes['M'] = MType = new Type(SysPod, "M", 0, null);  // Variegated Map
        GenericParameterTypes['R'] = RType = new Type(SysPod, "R", 0, null);  // Return
        GenericParameterTypes['V'] = VType = new Type(SysPod, "V", 0, null);  // Value

        List noMixins = new List(TypeType, 0).ro();
        for (int i=0; i<GenericParameterTypes.Length; i++)
        {
          Type gp = GenericParameterTypes[i];
          if (gp == null) continue;
          gp.m_base   = ObjType;
          gp.m_mixins = noMixins;
        }

        m_args = new List(StrType);
        m_homeDir = new LocalFile(new DirectoryInfo(HomeDir));
        m_hostName = Str.make(Environment.MachineName);
        m_userName = Str.make(Environment.UserName);

        m_env = new Map(StrType, StrType);
        try
        {
          // environment variables
          IDictionary getenv = Environment.GetEnvironmentVariables();
          foreach (DictionaryEntry de in getenv)
          {
            string key = (string)de.Key;
            string val = (string)de.Value;
            m_env.set(Str.make(key), Str.make(val));
          }

          // TODO - is there an equiv in C# for Java sys props?
          /*
          // Java system properties
          it = System.getProperties().keySet().iterator();
          while (it.hasNext())
          {
            String key = (String)it.next();
            String val = System.getProperty(key);
            env.set(Str.make(key), Str.make(val));
          }
          */

          // sys.properties
          LocalFile f = new LocalFile(new FileInfo(HomeDir + "/lib/sys.props"));
          if (f.exists().val)
          {
            try
            {
              Map props = f.readProps();
              m_env.setAll(props);
            }
            catch (Exception e)
            {
              System.Console.WriteLine("ERROR: Invalid props file: " + f);
              System.Console.WriteLine("  " + e);
            }
          }
        }
        catch (Exception e)
        {
          System.Console.WriteLine(e.StackTrace);
        }
        m_env = m_env.ro();
      }
      catch(Exception e)
      {
        Console.WriteLine("ERROR: Sys.static");
        Console.WriteLine(e.ToString());
        throw e;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Built-in Types
  //////////////////////////////////////////////////////////////////////////

    // the Eve of all types
    public static readonly Type ObjType;

    // basic primitives
    public static readonly Type NumType;
    public static readonly Type BoolType;
    public static readonly Type CharsetType;
    public static readonly Type DurationType;
    public static readonly Type IntType;
    public static readonly Type EnumType;
    public static readonly Type FloatType;
    public static readonly Type ListType;
    public static readonly Type MapType;
    public static readonly Type MonthType;
    public static readonly Type PodType;
    public static readonly Type RangeType;
    public static readonly Type StrType;
    public static readonly Type StrBufType;
    public static readonly Type TestType;
    public static readonly Type TimeType;
    public static readonly Type TypeType;
    public static readonly Type UriType;
    public static readonly Type VoidType;
    public static readonly Type WeekdayType;

    // reflection
    public static readonly Type SlotType;
    public static readonly Type FieldType;
    public static readonly Type MethodType;
    public static readonly Type MethodParamType;
    public static readonly Type FacetsType;

    // IO
    public static readonly Type InStreamType;
    public static readonly Type SysInStreamType;
    public static readonly Type OutStreamType;
    public static readonly Type SysOutStreamType;
    public static readonly Type BufType;
    public static readonly Type FileType;
    public static readonly Type LocalFileType;
    public static readonly Type ZipEntryFileType;
    public static readonly Type ZipType;

    // utils
    //public static readonly Type ProcessType;
    public static readonly Type DependType;
    public static readonly Type VersionType;

    // exceptions
    public static readonly Type ErrType;
    public static readonly Type ArgErrType;
    public static readonly Type CastErrType;
    public static readonly Type IndexErrType;
    public static readonly Type IOErrType;
    public static readonly Type NotImmutableErrType;
    public static readonly Type NullErrType;
    public static readonly Type ReadonlyErrType;
    public static readonly Type UnknownPodErrType;
    public static readonly Type UnknownSlotErrType;
    public static readonly Type UnknownTypeErrType;
    public static readonly Type UnsupportedErrType;

    // generic parameter types used with generic types List, Map, and Method
    public static readonly Type[] GenericParameterTypes = new Type[256];
    public static readonly Type AType, BType, CType, DType, EType, FType, GType,
                             HType, KType, LType, MType, RType, VType;

    public static Type genericParameterType(string name)
    {
      if (name.Length == 1 && name[0] < GenericParameterTypes.Length)
        return GenericParameterTypes[name[0]];
      else
        return null;
    }

    static Type builtin(string name, Type baseType)
    {
      return SysPod.findType(name, true);
      //if (SysPod.fpod != null) return SysPod.findType(name, true);
      //return SysPod.Stub(Type.Stub(SysPod, name, baseType));
    }

    public static readonly List m_args;
    public static readonly LocalFile m_homeDir;
    public static readonly Str m_hostName;
    public static readonly Str m_userName;
    private static Map m_env;
  }
}