//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   12 Mar 07  Andy Frank  Creation
//

using System;
using System.IO;
using System.Text;

namespace Fan.Sys
{
  /// <summary>
  /// SysInStream.
  /// </summary>
  public class SysInStream : FanObj, InStream
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public static SysInStream make(Stream inn, Int bufSize)
    {
      if (bufSize == null || bufSize.val == 0)
        return new SysInStream(inn);
      else
        return new SysInStream(new BufferedStream(inn, (int)bufSize.val));
    }

    public SysInStream(Stream inn)
    {
      m_in = inn;
      m_charset = Charset.utf8();
      m_charsetDecoder = m_charset.newDecoder();
      m_charsetEncoder = m_charset.newEncoder();
    }

  //////////////////////////////////////////////////////////////////////////
  // Obj
  //////////////////////////////////////////////////////////////////////////

    public override Type type() { return Sys.SysInStreamType; }

  //////////////////////////////////////////////////////////////////////////
  // InStream
  //////////////////////////////////////////////////////////////////////////

    public Int read() { int n = r(); return n < 0 ? null : Int.pos(n); }
    public virtual int r()
    {
      try
      {
        return m_in.ReadByte();
      }
      catch (IOException e)
      {
        throw IOErr.make(e).val;
      }
    }

    public virtual Int readBuf(Buf buf, Int n)
    {
      try
      {
        int len = (int)n.val;
        buf.grow(buf.m_pos+len);
        int read = m_in.Read(buf.m_buf, buf.m_pos, len);
        if (read <= 0) return null;
        buf.m_pos  += read;
        buf.m_size += read;
        return Int.pos(read);
      }
      catch (IOException e)
      {
        throw IOErr.make(e).val;
      }
    }

    public InStream unread(Int n) { return unread((int)n.val); }
    public virtual InStream unread(int n)
    {
      try
      {
        // don't take the hit until we know we need to wrap
        // the raw input stream with a pushback stream
        if (!(m_in is PushbackStream))
          m_in = new PushbackStream(m_in, 1024);
        ((PushbackStream)m_in).Unread(n);
        return this;
      }
      catch (IOException e)
      {
        throw IOErr.make(e).val;
      }
    }

    public Buf readAllBuf() { return InStream_.readAllBuf(this); }

    public Buf readBufFully(Buf buf, Int n) { return InStream_.readBufFully(this, buf, n); }

    public Int readU1() { return InStream_.readU1(this); }

    public Int readS1() { return InStream_.readS1(this); }

    public Int readU2() { return InStream_.readU2(this); }

    public Int readS2() { return InStream_.readS2(this); }

    public Int readU4() { return InStream_.readU4(this); }

    public Int readS4() { return InStream_.readS4(this); }

    public Int readS8() { return InStream_.readS8(this); }

    public Float readF4() { return InStream_.readF4(this); }

    public Float readF8() { return InStream_.readF8(this); }

    public Bool readBool() { return InStream_.readBool(this); }

    public Str readUtf() { return InStream_.readUtf(this); }

    public virtual int rChar()
    {
      return m_charsetDecoder.decode(this);
    }

    public virtual Int readChar()
    {
      int ch = m_charsetDecoder.decode(this);
      return ch < 0 ? null : Int.pos(ch);
    }

  /*
    public OutStream writeChars(Str s) { return writeChars(s.val, 0, s.val.length()); }
    public OutStream writeChars(Str s, Int off) { return writeChars(s.val, (int)off.val, s.val.length()-(int)off.val); }
    public OutStream writeChars(Str s, Int off, Int len) { return writeChars(s.val, (int)off.val, (int)len.val); }
    public OutStream writeChars(String s, int off, int len)
    {
      try
      {
        writer().write(s, off, len);
        return this;
      }
      catch (java.io.IOException e)
      {
        throw IOErr.make(e);
      }
    }
  */

    public virtual InStream unreadChar(Int c)
    {
      m_charsetEncoder.encode((char)c.val, this);
      return this;
    }

    public Str readLine() { return InStream_.readLine(this, Int.m_pos[4096]); }
    public Str readLine(Int max) { return InStream_.readLine(this, max); }

    public List readAllLines() { return InStream_.readAllLines(this); }

    public Str readAllStr() { return InStream_.readAllStr(this, Bool.True); }
    public Str readAllStr(Bool normalizeNewlines)  { return InStream_.readAllStr(this, normalizeNewlines); }

    public Map readProps() { return InStream_.readProps(this); }

    public Charset charset()
    {
      return m_charset;
    }

    public void charset(Charset charset)
    {
      m_charsetDecoder = charset.newDecoder();
      m_charsetEncoder = charset.newEncoder();
      m_charset = charset;
    }

    public Int pipe(OutStream ot) { return InStream_.pipe(this, ot, null, Bool.True); }
    public Int pipe(OutStream ot, Int n) { return InStream_.pipe(this, ot, n, Bool.True); }
    public Int pipe(OutStream ot, Int n, Bool close) { return InStream_.pipe(this, ot, n, close); }

    public virtual Bool close()
    {
      try
      {
        m_in.Close();
        return Bool.True;
      }
      catch (IOException)
      {
        return Bool.False;
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // PushbackStream
  //////////////////////////////////////////////////////////////////////////

    internal class PushbackStream : Stream
    {
      public PushbackStream(Stream parent, int size)
      {
        this.parent = parent;
        this.pushback = new byte[size];
      }

      //
      // Properties
      //

      public override bool CanRead  { get { return parent.CanRead;  } }
      public override bool CanWrite { get { return parent.CanWrite; } }
      public override bool CanSeek  { get { return false; } } // don't allow seeking
      public override long Length   { get { return parent.Length; } }
      public override long Position
      {
        get { return parent.Position;  }
        set { parent.Position = value; }
      }

      //
      // Methods
      //

      public void Unread(int b)
      {
        if (pos >= pushback.Length)
          throw new IOException("Not enough room in pushback buffer");
        pushback[pos++] = (byte)b;
      }

      public override int Read(byte[] buf, int offset, int count)
      {
        // check pushback first
        if (pos > 0)
        {
          int len = Math.Min(pos, count);
          Array.Copy(pushback, pos-len, buf, offset, len);
          pos -= len;
          offset += len;
          count -= len;

          // short-circuit if buf filled with pushback
          if (count == 0) return len;
        }

        // grab remaining bytes
        return parent.Read(buf, offset, count);
      }

      public override void Write(byte[] buf, int offset, int count)
      {
        // This is purely an input stream, so Write not allowed
        throw new IOException("PushbackStream does not allow Write()");
      }

      public override long Seek(long offset, SeekOrigin origin)
      {
        // We don't allow seeking
        throw new IOException("PushbackStream does not allow Seek()");
      }

      public override void SetLength(long val)
      {
        // We don't allow seeking
        throw new IOException("PushbackStream does not allow SetLength()");
      }

      public override void Close() { parent.Close(); }
      public override void Flush() { parent.Flush(); }

      //
      // Fields
      //

      private Stream parent;
      private byte[] pushback;
      private int pos = 0;
    }

  //////////////////////////////////////////////////////////////////////////
  // .NET Conversion
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Get a System.IO.Stream for the specified input stream.
    /// </summary>
    public static Stream net(InStream ins)
    {
      if (ins is SysInStream)
        return ((SysInStream)ins).m_in;
      else
        return new NetInputStream (ins);
    }

    internal class NetInputStream : Stream
    {
      public NetInputStream(InStream ins) { this.ins = ins; }

      // Properties
      public override bool CanRead  { get { return true;  } }
      public override bool CanWrite { get { return false; } }
      public override bool CanSeek  { get { return false; } }
      public override long Length   { get { return -1; } }
      public override long Position { get { return -1; } set {} }

      // Methods
      public override int Read(byte[] b, int off, int len)
      {
        buf.m_buf = b;
        buf.m_pos = off;
        buf.m_size = b.Length;
        Int n = ins.readBuf(buf, Int.make(len));
        if (n == null) return -1;
        return (int)n.val;
      }
      public override long Seek(long off, SeekOrigin origin) { return -1; }
      public override void SetLength(long val) {}
      public override void Flush() {}
      public override void Write(byte[] b, int off, int len) {}
      public override void Close() { ins.close(); }

      // Fields
      private InStream ins;
      private Buf buf = new Buf();
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    Stream m_in;
    Charset m_charset;
    Charset.Decoder m_charsetDecoder;
    Charset.Encoder m_charsetEncoder;

  }
}