//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   12 Mar 07  Andy Frank  Creation
//

using System.IO;
using System.Text;

namespace Fan.Sys
{
  /// <summary>
  /// SysOutStream.
  /// </summary>
  public class SysOutStream : FanObj, OutStream
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public static SysOutStream make(Stream ot, Int bufSize)
    {
      if (bufSize == null || bufSize.val == 0)
        return new SysOutStream(ot);
      else
        return new SysOutStream(new BufferedStream(ot, (int)bufSize.val));
    }

    public SysOutStream(Stream ot)
    {
      this.m_out = ot;
      m_charset = Charset.utf8();
      m_charsetEncoder = m_charset.newEncoder();
    }

  //////////////////////////////////////////////////////////////////////////
  // Obj
  //////////////////////////////////////////////////////////////////////////

    public override Type type() { return Sys.SysOutStreamType; }

  //////////////////////////////////////////////////////////////////////////
  // OutStream
  //////////////////////////////////////////////////////////////////////////

    public OutStream write(Int b) { return w((int)b.val); }
    public virtual OutStream w(int v)
    {
      try
      {
        m_out.WriteByte((byte)v);
        return this;
      }
      catch (IOException e)
      {
        throw IOErr.make(e).val;
      }
    }

    public OutStream writeBuf(Buf buf) { return writeBuf(buf, buf.remaining()); }
    public virtual OutStream writeBuf(Buf buf, Int n)
    {
      try
      {
        int len = (int)n.val;
        if (buf.m_pos + len > buf.m_size) throw IOErr.make("Not enough bytes to write").val;
        m_out.Write(buf.m_buf, buf.m_pos, len);
        buf.m_pos += len;
        return this;
      }
      catch (IOException e)
      {
        throw IOErr.make(e).val;
      }
    }

    public OutStream writeI2(Int x) { return OutStream_.writeI2(this, x); }

    public OutStream writeI4(Int x) { return OutStream_.writeI4(this, x); }

    public OutStream writeI8(Int x) { return OutStream_.writeI8(this, x); }

    public OutStream writeF4(Float x) { return OutStream_.writeF4(this, x); }

    public OutStream writeF8(Float x) { return OutStream_.writeF8(this, x); }

    public OutStream writeBool(Bool x) { return OutStream_.writeBool(this, x); }

    public OutStream writeUtf(Str x) { return OutStream_.writeUtf(this, x); }

    public virtual OutStream writeChar(Int c)
    {
      m_charsetEncoder.encode((char)c.val, this);
      return this;
    }

    public OutStream writeChars(Str s) { return writeChars(s.val, 0, s.val.Length); }
    public OutStream writeChars(Str s, Int off) { return writeChars(s.val, (int)off.val, s.val.Length-(int)off.val); }
    public OutStream writeChars(Str s, Int off, Int len) { return writeChars(s.val, (int)off.val, (int)len.val); }
    public virtual OutStream writeChars(string s, int off, int len)
    {
      int end = off+len;
      for (int i=off; i<end; ++i)
        m_charsetEncoder.encode(s[i], this);
      return this;
    }

    public OutStream print(Obj obj) { return OutStream_.print(this, obj); }

    public OutStream printLine() { return OutStream_.printLine(this, Str.Empty); }
    public OutStream printLine(Obj obj) { return OutStream_.printLine(this, obj); }

    public OutStream writeProps(Map props) { return OutStream_.writeProps(this, props, Bool.True); }
    public OutStream writeProps(Map props, Bool close) { return OutStream_.writeProps(this, props, close); }

    public virtual void flush()
    {
      try
      {
        m_out.Flush();
      }
      catch (IOException e)
      {
        throw IOErr.make(e).val;
      }
    }

    public virtual Bool close()
    {
      try
      {
        m_out.Close();
        return Bool.True;
      }
      catch (IOException)
      {
        return Bool.False;
      }
    }

    public Charset charset()
    {
      return m_charset;
    }

    public void charset(Charset charset)
    {
      m_charsetEncoder = charset.newEncoder();
      m_charset = charset;
    }

  //////////////////////////////////////////////////////////////////////////
  // .NET Conversion
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Get a System.IO.Stream for the specified output stream.
    /// </summary>
    public static Stream net(OutStream outs)
    {
      if (outs is SysOutStream)
        return ((SysOutStream)outs).m_out;
      else
        return new NetOutputStream(outs);
    }

    internal class NetOutputStream : Stream
    {
      public NetOutputStream(OutStream outs) { this.outs = outs; }

      // Properties
      public override bool CanRead  { get { return false;  } }
      public override bool CanWrite { get { return true; } }
      public override bool CanSeek  { get { return false; } }
      public override long Length   { get { return -1; } }
      public override long Position { get { return -1; } set {} }

      // Methods
      public override int Read(byte[] buf, int off, int len) { return -1; }
      public override long Seek(long off, SeekOrigin origin) { return -1; }
      public override void SetLength(long val) {}
      public override void Flush() {}
      public override void Write(byte[] b, int off, int len)
      {
        buf.m_buf = b;
        buf.m_pos = off;
        buf.m_size = b.Length;
        outs.writeBuf(buf, Int.make(len));
      }
      public override void Close() { outs.close(); }

      // Fields
      private OutStream outs;
      private Buf buf = new Buf();
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    Stream m_out;
    Charset m_charset;
    Charset.Encoder m_charsetEncoder;

  }
}