//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   16 Jan 07  Brian Frank  Creation
//

using DateTime = System.DateTime;

namespace Fan.Sys
{
  /// <summary>
  /// Time represents an absolute instance in time.  Fan time is normalized
  /// as nanosecond ticks since 1 Jan 2000  GMT.
  /// </summary>
  public sealed class Time : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Time now()  { return now(toleranceDefault); }
    public static Time now(Duration tolerance)
    {
      lock (nowLock)
      {
        long now = (DateTime.Now.Ticks - m_normalize) * 100;
        if (tolerance != null && now - cached.m_ticks <= tolerance.m_ticks)
          return cached;
        else
          return cached = new Time(now);
      }
    }

    public static Time make(Int ticks) { return make(ticks.val); }
    public static Time make(long ticks)
    {
      if (ticks == 0) return zero;
      return new Time(ticks);
    }

    private Time(long ticks)
    {
      this.m_ticks = ticks;
    }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override Bool equals(Obj obj)
    {
      if (obj is Time)
      {
        return m_ticks == ((Time)obj).m_ticks ? Bool.True : Bool.False;
      }
      return Bool.False;
    }

    public override Int compare(Obj obj)
    {
      long that = ((Time)obj).m_ticks;
      if (m_ticks < that) return Int.LT; return m_ticks  == that ? Int.EQ : Int.GT;
    }

    public override int GetHashCode()
    {
      //return (int)(ticks ^ (ticks >>> 32));
      return (int)(m_ticks ^ (m_ticks >> 32));
    }

    public override Int hash()
    {
      return Int.make(m_ticks);
    }

    public override Type type()
    {
      return Sys.TimeType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Access
  //////////////////////////////////////////////////////////////////////////

    public Int ticks()
    {
      return Int.make(m_ticks);
    }

  //////////////////////////////////////////////////////////////////////////
  // Operators
  //////////////////////////////////////////////////////////////////////////

    public Duration minus(Time time)
    {
      return Duration.make(m_ticks-time.m_ticks);
    }

    public Time plus(Duration duration)
    {
      long d = duration.m_ticks;
      if (d == 0) return this;
      return make(m_ticks+d);
    }

  //////////////////////////////////////////////////////////////////////////
  // Conversion
  //////////////////////////////////////////////////////////////////////////

    public override Str toStr()
    {
      if (m_str == null)
      {
        string s = new DateTime((m_ticks / 100) + m_normalize).ToString(format);
        m_str = Str.make(s);
      }
      return m_str;
    }

  //////////////////////////////////////////////////////////////////////////
  // C#
  //////////////////////////////////////////////////////////////////////////

    public static Time net(DateTime dt)
    {
      return (dt.Year < 1700) ? null : new Time((dt.Ticks - m_normalize) * 100);
    }
    public DateTime net() { return new DateTime((m_ticks / 100) + m_normalize); }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    private static readonly string format = "d-MMM-yyyy HH:mm:ss zzz";
    private static readonly Time zero = new Time(0);
    private static readonly Duration toleranceDefault = Duration.makeMillis(250);
    private static readonly object nowLock = new object();
    private static Time cached = zero;

    // .NET counts ticks from 1/1/0001, so we need to
    // normalize our representation to 1/1/1970
    private static long m_normalize = new DateTime(1970, 1, 1).Ticks;

    private readonly long m_ticks;
    private Str m_str;

  }
}