//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   19 Oct 06  Andy Frank  Creation
//

using System;
using System.Text;

namespace Fan.Sys
{
  /// <summary>
  /// Uri is used to immutably represent a Universal Resource Identifier.
  /// </summary>
  public sealed class Uri : FanObj
  {

  //////////////////////////////////////////////////////////////////////////
  // Construction
  //////////////////////////////////////////////////////////////////////////

    public static Uri parse(string s) { return parse(Str.make(s)); }
    public static Uri parse(Str s)
    {
      try
      {
        return new Uri(s);
      }
      catch (Exception)
      {
        throw ArgErr.make("Invalid Uri: `" + s + "`").val;
      }
    }

    public Uri(System.Uri uri) : this(Str.make(uri.ToString())) {}

    public Uri(Str s)
    {
      m_str      = s;
      m_scheme   = null;
      m_userInfo = null;
      m_host     = null;
      m_port     = null;
      m_pathStr  = null;
      m_queryStr = null;
      m_frag     = null;

      parseBasic(s.val);
      m_path  = parsePath(m_pathStr);
      m_query = parseQuery(m_queryStr);
    }

//////////////////////////////////////////////////////////////////////////
// Parsing
//////////////////////////////////////////////////////////////////////////

    private void parseBasic(string s)
    {
      StringBuilder buf = new StringBuilder();

      bool pastScheme = false;
      bool forceQuery = false;

      for (int i=0; i<s.Length; i++)
      {
        int last  = (i > 0) ? s[i-1] : -1;
        int curr  = s[i];
        int peek  = (i < s.Length-1) ? s[i+1] : -1;
        int peek2 = (i < s.Length-2) ? s[i+2] : -1;

        // found scheme
        if (m_scheme == null && !pastScheme && curr == ':')
        {
          m_scheme = Str.make(buf.ToString().ToLower());
          buf.Length = 0;

          // check if opaque
          if (peek == '/' && peek2 == '/')
          {
            isOpaque = false;
            continue;
          }
          else
          {
            isOpaque = true;
            break;
          }
        }

        // found path
        if (curr == '/')
        {
          // we're past the scheme if we find a slash
          pastScheme = true;

          // check for authority
          if (m_host == null && peek == '/')
          {
            i += 1;
            while (i+1 < s.Length &&
              s[i+1] != '/' &&
              s[i+1] != '?' &&
              s[i+1] != '#')
            {
              // we gotta userInfo, set it and keep looking
              // for the host name and/or port
              if (s[i+1] == '@')
              {
                m_userInfo = Str.make(buf.ToString());
                i++;
                buf.Length = 0;
                continue;
              }

              // we gotta port number, set the host name
              // and keep chugging along
              if (s[i+1] == ':')
              {
                m_host = Str.make(buf.ToString());
                i++;
                buf.Length = 0;
                continue;
              }

              buf.Append(s[++i]);
            }

            // if host already set, then this is the port
            if (m_host == null)
              m_host = Str.make(buf.ToString());
            else
              m_port = Str.make(buf.ToString()).toInt();

            buf.Length = 0;
            continue;
          }
        }

        // found query
        if (curr == '?')
        {
          // dump current contents into either path
          m_pathStr = Str.make(buf.ToString());
          buf.Length = 0;
          forceQuery = true; // force empty query strs to be set
          continue;
        }

        // found fragment
        if (curr == '#')
        {
          // dump current contents into either pathStr or queryStr
          if (m_pathStr == null)
          {
            m_pathStr = Str.make(buf.ToString());
            buf.Length = 0;
          }
          else
          {
            m_queryStr = Str.make(buf.ToString());
            buf.Length = 0;
          }

          // now get fragment
          for (i++; i<s.Length; i++)
            buf.Append((char)s[i]);
          m_frag = Str.make(buf.ToString());
          buf.Length = 0;

          break;
        }

        buf.Append((char)curr);
      }

      // Handle any dangling characters if they exist
      if (isOpaque)
      {
        // fragments are allowed on opaque uris, so check for
        // one since we bailed out of the main loop above
        int index = s.IndexOf("#");
        if (index == s.Length-1) m_frag = Str.make("");
        else if (index >= 0) m_frag = Str.make(s.Substring(index+1));
      }
      else
      {
        // if more content in buffer, dump it into the right place
        if (buf.Length > 0)
        {
          if (m_pathStr == null)
            m_pathStr = Str.make(buf.ToString());
          else
            m_queryStr = Str.make(buf.ToString());
        }

        // make sure the path is always set to at least an empty str
        if (m_pathStr == null)
          m_pathStr = Str.make("");

        // make sure an empty query string still gets set
        if (forceQuery && m_queryStr == null)
          m_queryStr = Str.make("");
      }
    }

    internal static List parsePath(Str pathStr)
    {
      if (pathStr == null) return null;
      List list = new List(Sys.StrType);

      try
      {
        string p = pathStr.val;
        int start = 0, len = p.Length;
        for (int i=0; i<len; i++)
        {
          int ch = p[i];
          if (ch != '/') continue;
          if (start < i)
            list.add(Str.make(p.Substring(start, (i-start))));
          start = i+1;
        }
        if (start < len)
          list.add(Str.make(p.Substring(start, (len-start))));
      }
      catch (Exception e)
      {
        // don't let internal error bring down whole uri
        System.Console.WriteLine(e);
      }

      return list.trim().ro();
    }

    internal static Map parseQuery(Str queryStr)
    {
      if (queryStr == null) return null;
      Map map = new Map(Sys.StrType, Sys.StrType);

      try
      {
        string q = queryStr.val;
        int start = 0, eq = 0, len = q.Length;
        for (int i=0; i<len; i++)
        {
          int ch = q[i];
          if (ch == '=') eq = i;
          if (ch != '&' && ch != ';') continue;

          if (start < i)
          {
            if (start == eq)
              map.set(Str.make(q.Substring(start, (i-start))), null);
            else
              map.set(Str.make(q.Substring(start, (eq-start))), Str.make(q.Substring(eq+1, (i-(eq+1)))));
          }

          start = eq = i+1;
        }

        if (start < len)
        {
          if (start == eq)
            map.set(Str.make(q.Substring(start, (len-start))), null);
          else
            map.set(Str.make(q.Substring(start, (eq-start))), Str.make(q.Substring(eq+1, (len-(eq+1)))));
        }
      }
      catch (Exception e)
      {
        // don't let internal error bring down whole uri
        System.Console.WriteLine(e);
      }

      return map.ro();
    }

  //////////////////////////////////////////////////////////////////////////
  // Identity
  //////////////////////////////////////////////////////////////////////////

    public override bool Equals(Object obj)
    {
      return m_str.val == (obj as Uri).m_str.val;
    }

    public override Bool equals(Obj obj)
    {
      if (obj is Uri)
      {
        return m_str.equals((obj as Uri).m_str);
      }
      return Bool.False;
    }

    public override int GetHashCode()
    {
      return m_str.GetHashCode();
    }

    public override Int hash()
    {
      return m_str.hash();
    }

    public override Str toStr()
    {
      return m_str;
    }

    public override Type type()
    {
      return Sys.UriType;
    }

  //////////////////////////////////////////////////////////////////////////
  // Components
  //////////////////////////////////////////////////////////////////////////

    public Bool isAbs()
    {
      return m_scheme != null ? Bool.True : Bool.False;
    }

    public Bool isRel()
    {
      return m_scheme == null ? Bool.True : Bool.False;
    }

    public Bool isDir()
    {
      if (m_pathStr != null)
      {
        string p = m_pathStr.val;
        int len = p.Length;
        if (len > 0 && p[len-1] == '/')
          return Bool.True;
      }
      return Bool.False;
    }

    public Str scheme()
    {
      return m_scheme;
    }

    public Str body()
    {
      if (m_scheme == null) return m_str;
      return Str.make(m_str.val.Substring(m_scheme.val.Length+1));
    }

    public Str auth()
    {
      if (m_host == null) return null;
      if (m_port == null)
      {
        if (m_userInfo == null) return m_host;
        else return Str.make(m_userInfo.val + '@' + m_host.val);
      }
      else
      {
        if (m_userInfo == null) return Str.make(m_host.val + ':' + m_port.val);
        else return Str.make(m_userInfo.val + '@' + m_host.val + ':' + m_port.val);
      }
    }

    public Str host()
    {
      return m_host;
    }

    public Str userInfo()
    {
      return m_userInfo;
    }

    public Int port()
    {
      return m_port;
    }

    public List path()
    {
      return m_path;
    }

    public Str pathStr()
    {
      return m_pathStr;
    }

    public Bool isPathAbs()
    {
      if (m_pathStr == null || m_pathStr.val.Length == 0)
        return Bool.False;
      else
        return m_pathStr.val[0] == '/' ? Bool.True : Bool.False;
    }

    public Str name()
    {
      if (m_path == null || m_path.sz() == 0) return null;
      return (Str)m_path.last();
    }

    public Str basename()
    {
      Str nme = name();
      if (nme == null) return null;
      string n = nme.val;
      int dot = n.LastIndexOf('.');
      if (dot < 2)
      {
        if (dot < 0) return nme;
        if (n == ".") return nme;
        if (n == "..") return nme;
      }
      return Str.make(n.Substring(0, dot));
    }

    public Str ext()
    {
      Str nme = name();
      if (nme == null) return null;
      string n = nme.val;
      int dot = n.LastIndexOf('.');
      if (dot < 2)
      {
        if (dot < 0) return null;
        if (n == ".") return null;
        if (n == "..") return null;
      }
      return Str.make(n.Substring(dot+1));
    }

    public Map query()
    {
      return m_query;
    }

    public Str queryStr()
    {
      return m_queryStr;
    }

    public Str frag()
    {
      return m_frag;
    }

  //////////////////////////////////////////////////////////////////////////
  // Manipulation
  //////////////////////////////////////////////////////////////////////////

    public Uri parent()
    {
      // if no path bail
      if (m_path == null || m_path.sz() == 0) return null;

      // if just a simple filename, then no parent
      string p = m_pathStr.val;
      if (m_path.sz() == 1 && !isPathAbs().val && !isDir().val) return null;

      // build parent uri string
      StringBuilder s = new StringBuilder();
      if (m_scheme != null) s.Append(m_scheme).Append(':');
      if (m_host   != null) s.Append("//").Append(auth());
      s.Append(p.Substring(0, p.LastIndexOf('/', p.Length-2)+1));

      return Uri.parse(s.ToString());
    }

    public Uri normalize()
    {
      // Cannot normalize opaque URI's
      if (isOpaque) return Uri.parse(m_str);

      StringBuilder s = new StringBuilder();
      List norm = m_path.rw();

      // normalize the path
      int i = 0;
      while (i >= 0 && i < norm.sz())
      {
        string pre  = (i > 0) ? (norm.get(i-1) as Str).val : null;
        string curr = (norm.get(i) as Str).val;

        if (curr == ".") norm.removeAt(Int.m_pos[i--]);
        else if (curr == ".." && (pre != null && pre != ".."))
        {
          norm.removeAt(Int.m_pos[i]);
          norm.removeAt(Int.m_pos[i-1]);
          i -= 2;
        }
        else i++;
      }

      if (m_scheme  != null) s.Append(m_scheme.val).Append(":");
      Str a = auth();
      if (a != null) s.Append("//").Append(a.val);
      if (norm.sz() == 0 && m_pathStr.val.Length > 0 && m_pathStr.val[0] == '/')
      {
        s.Append('/');
      }
      else
      {
        for (i=0; i<norm.sz(); i++)
        {
          Str curr = norm.get(i) as Str;
          if (i > 0 || (i == 0 && m_pathStr.val[0] == '/'))
            s.Append('/');
          s.Append(curr);
          if (i == norm.sz()-1)
          {
            bool ta = norm.sz() != m_path.sz() && curr.val != "..";
            Str test = m_path.get(m_path.sz()-1) as Str;
            bool tb = test.val == "..";
            bool tc = m_pathStr.val[m_pathStr.val.Length-1] == '/';
            if ((ta && tb || tc)) s.Append("/");
          }
        }
      }
      if (m_queryStr != null) s.Append('?').Append(m_queryStr.val);
      if (m_frag     != null) s.Append('#').Append(m_frag.val);
      return new Uri(Str.make(s.ToString()));
    }

    public Uri plus(Uri join)
    {
      if (join.isAbs().val) return join;
      if (join.isOpaque) return join;

      // TODO - check for solo frag

      StringBuilder s = new StringBuilder();
      Str myAuth   = auth();
      Str joinAuth = join.auth();
      string newPath = null;

      if (m_scheme != null) s.Append(m_scheme).Append(":");
      if (joinAuth != null) s.Append("//").Append(joinAuth).Append(join.m_pathStr);
      else
      {
        if (myAuth != null) s.Append("//").Append(myAuth);
        if (join.isPathAbs().val) s.Append(newPath = join.m_pathStr.val);
        else
        {
          if (isDir().val)
          {
            newPath = m_pathStr.val + join.m_pathStr.val;
            s.Append(newPath = new Uri(Str.make(newPath)).normalize().m_pathStr.val);
          }
          else
          {
            newPath = m_pathStr.val;
            int z = newPath.LastIndexOf("/");
            if (z < 0) newPath = "";
            else newPath = newPath.Substring(0, z+1);
            newPath += join.m_pathStr.val;

            if (myAuth != null && newPath.Length > 0 && newPath[0] != '/')
              s.Append('/');
            s.Append(newPath = new Uri(Str.make(newPath)).normalize().m_pathStr.val);
          }
        }
      }
      if (join.m_queryStr != null)
      {
        if (newPath.Length == 0) s.Append('/');
        s.Append('?').Append(join.m_queryStr);
      }
      if (join.m_frag     != null) s.Append('#').Append(join.m_frag);

      return new Uri(Str.make(s.ToString()));
    }

    public Uri minus(Uri bse)
    {
      // We use the reverse semantics from java.net.URI.relativize()

      if (isOpaque || bse.isOpaque) return this;

      if (m_scheme == null && bse.m_scheme != null) return this;
      if (m_scheme != null && bse.m_scheme == null) return this;
      if (m_scheme != null && bse.m_scheme != null && m_scheme.val != bse.m_scheme.val)
        return this;

      Str myAuth = auth();
      Str baseAuth = bse.auth();

      if (myAuth == null && baseAuth != null) return this;
      if (myAuth != null && baseAuth == null) return this;
      if (myAuth != null && baseAuth != null && myAuth.val != baseAuth.val)
        return this;

      StringBuilder s = new StringBuilder();

      string myPath   = m_pathStr.val;
      string basePath = bse.m_pathStr.val;

      if (myPath.IndexOf(basePath) == 0)
      {
        int off = basePath.Length;
        if (myPath.Length > basePath.Length && basePath[basePath.Length-1] != '/') off++;
        s.Append(myPath.Substring(off));
      }
      else
        s.Append(myPath);

      if (m_queryStr != null) s.Append('?').Append(m_queryStr);
      if (m_frag     != null) s.Append('#').Append(m_frag);

      return new Uri(Str.make(s.ToString()));
    }

    public Uri tail() { return tail(Int.One); }
    public Uri tail(Int levels)
    {
      int lvl = (int)levels.val;
      if (m_path == null || m_path.sz() < lvl)
        throw ArgErr.make("Cannot call tail(" + lvl + ") on `" + m_str + "`").val;
      StringBuilder s = new StringBuilder();
      for (int i=lvl; i<m_path.sz(); i++)
      {
        if (i > lvl) s.Append('/');
        s.Append(m_path.get(i));
      }
      return parse(Str.make(s.ToString()));
    }

  //////////////////////////////////////////////////////////////////////////
  // Resolution
  //////////////////////////////////////////////////////////////////////////

    public Obj get() { return get(null); }
    public Obj get(Obj baseUri)
    {
      if (m_scheme != null)
      {
        throw UnsupportedErr.make("Absolute Uris not supported yet: " + m_str).val;
      }

      if (baseUri == null)
        throw ArgErr.make("Must specify base for `" + m_str + "`").val;

      //return baseUri.trapUri(this);
      return null;
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    internal Str m_str;
    internal bool isOpaque;

    internal Str m_scheme;
    internal Str m_userInfo;
    internal Str m_host;
    internal Int m_port;
    internal List m_path;
    internal Str m_pathStr;
    internal Map m_query;
    internal Str m_queryStr;
    internal Str m_frag;

  }
}