//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   2 Oct 06  Andy Frank  Creation
//

using System.Collections;
using System.Reflection;
using Fan.Sys;
using Fanx.Fcode;

namespace Fanx.Emit
{
  /// <summary>
  /// FTypeEmit translates FType fcode to Java bytecode.
  /// </summary>
  public abstract class FTypeEmit : Emitter
  {

  //////////////////////////////////////////////////////////////////////////
  // Factory
  //////////////////////////////////////////////////////////////////////////

    public static System.Type emitAndLoad(FType ftype)
    {
      string className = ftype.m_pod.nname(ftype.m_self);
      return emitPod(ftype.m_pod).GetType(className);
    }

    private static Assembly emitPod(FPod pod)
    {
      string podName = pod.m_podName;
      Assembly assembly = (Assembly)assemblies[podName];
      if (assembly == null)
      {
        Emitter emitter = new Emitter(podName);

        // emit the pod class itself (which declares all constants)
        //FPodEmit.EmitAndLoad(emitter, pod);
        FPodEmit.Emit(emitter, pod);

        // emit all the rest of the types in this pod
        for (int i=0; i<pod.m_types.Length; i++)
        {
          FType ftype = pod.m_types[i];
          FTypeRef tref = ftype.m_pod.typeRef(ftype.m_self);
          Type parent = Type.find(pod.name(tref.podName), pod.name(tref.typeName), true);

          // make sure we have reflected to setup slots
          parent.reflect();

          // route based on type
          if ((ftype.m_flags & FConst.Mixin) != 0)
          {
            // interface
            FMixinInterfaceEmit iemit = new FMixinInterfaceEmit(emitter, parent, ftype);
            iemit.emit();

            // body class
            FMixinBodyEmit bemit = new FMixinBodyEmit(emitter, parent, ftype);
            bemit.emit();
          }
          else
          {
            // class
            new FClassEmit(emitter, parent, ftype).emit();
          }
        }

        byte[] buf = emitter.commit();
        assembly = Assembly.Load(buf);
        assemblies[podName] = assembly;

        // load $Pod type
        FPodEmit.Load(assembly, pod);
      }
      return assembly;
    }

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    protected FTypeEmit(Emitter emitter, Type parent, FType type)
    {
      this.emitter = emitter;
      this.parent  = parent;
      this.pod     = type.m_pod;
      this.type    = type;
    }

    /// <summary>
    /// Initialize an Emitter to generate a class with given name,
    /// super class, interfaces, and class level access flags.
    /// <summary>
    public void init(string thisClass, string baseClass, string[] interfaces, int flags)
    {
      this.className = thisClass;
      this.baseClassName = baseClass;
      this.interfaces = interfaces;
      //this.cp.add(new CpDummy());  // dummy entry since constant pool starts at 1
      //this.thisClassIndex  = cls(thisClass);
      //this.superClassIndex = cls(superClass);
      //this.interfaces = new int[interfaces.length];
      //for (int i=0; i<interfaces.length; ++i)
      //  this.interfaces[i] = cls(interfaces[i]);
      //this.flags = flags;
      this.isAbstract = (flags & FConst.Abstract) != 0;
    }

  //////////////////////////////////////////////////////////////////////////
  // Emit
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Emit to IL assembly.
    /// </summary>
    public virtual void emit()
    {
      init(nname(type.m_self), @base(), mixins(), type.m_flags);
      this.selfName = className;
      //preview();
      emitType();

      // emit class body
      for (int i=0; i<type.m_fields.Length;  i++) emit(type.m_fields[i]);
      for (int i=0; i<type.m_methods.Length; i++) emit(type.m_methods[i]);
      for (int i=0; i<type.m_facets.Length;  i++) emit(type.m_facets[i]);
      for (int i=0; i<parent.mixins().sz(); i++) emitMixinRouters((Type)parent.mixins().get(i));
      if (!hasInstanceInit) emitInstanceInit(null);
      if (!hasStaticInit) emitStaticInit(null);
      emitTypeConstFields();
    }

    /// <summary>
    /// Return the base type for this type.
    /// </summary>
    protected abstract string @base();

    protected virtual string[] mixins()
    {
      string[] mixins = new string[type.m_mixins.Length];
      for (int i=0; i<mixins.Length; i++)
        mixins[i] = nname(type.m_mixins[i]);
      return mixins;
    }

    /*
    private void Preview()
    {
      this.isNative = (type.flags & FConst.Native) != 0;
      if (!this.isNative)
      {
        for (int i=0; i<type.methods.length; ++i)
          if ((type.methods[i].flags & Native) != 0) { this.isNative = true; break; }
      }
    }
    */

    /// <summary>
    /// Emit the type information for this type.
    /// </summary>
    protected abstract void emitType();

    /// <summary>
    /// Emit a facet.
    /// </summary>
    private void emit(FFacet f)
    {
      //String name = pod.name(f.name);
      //if (name.equals(SourceFile))
      //{
      //  AttrEmit attr = emitAttr("SourceFile");
      //  attr.info.u2(utf(f.data.utf()));
      //}
    }

    /// <summary>
    /// Emit a field.
    /// </summary>
    protected void emit(FField f)
    {
      emitter.emitField(name(f.m_name), nname(f.m_type), fieldFlags(f.m_flags));
    }

    /// <summary>
    /// Emit a method.
    /// </summary>
    protected virtual void emit(FMethod m)
    {
      string n = name(m.m_name);
      bool isNative = (m.m_flags & FConst.Native) != 0;
      bool isCtor   = (m.m_flags & FConst.Ctor)   != 0;

      // static$init -> .cctor
      // instance$init -> .ctor
      if (n == "static$init")   { emitStaticInit(m); return; }
      if (n == "instance$init") { emitInstanceInit(m); return; }

      // handle native/constructor/normal method
      if (isNative)
      {
        //new FMethodEmit(this, m).emitNative();
        throw new System.Exception("(FTypeEmit.Emit() -> Don't handle this yet!!!!");
      }
      else if (isCtor)
      {
        new FMethodEmit(this, m).emitCtor();
      }
      else
      {
        new FMethodEmit(this, m).emitStandard();
      }
    }

    protected void emitInstanceInit(FMethod m)
    {
      hasInstanceInit = true;
      PERWAPI.CILInstructions code = ctor.CreateCodeBuffer();

      // initalize code to call super
      code.Inst(PERWAPI.Op.ldarg_0);

      // if closure, push MethodType static field
      if (methodType != null)
      {
      //  code.op2(GETSTATIC, typeField.ref());
      //  code.op2(CHECKCAST, cls("fan/sys/MethodType"));
      //  code.op2(INVOKESPECIAL, method(superClassName +".<init>(Lfan/sys/MethodType;)V"));
      }
      else
      {
        PERWAPI.Method baseCtor = emitter.findMethod(baseClassName, ".ctor",
          new string[0], "System.Void");
        baseCtor.AddCallConv(PERWAPI.CallConv.Instance); // if stub, make sure instance callconv
        code.MethInst(PERWAPI.MethodOp.call, baseCtor);
      }

      // make peer
      //if (isNative)
      //{
      //  code.op(ALOAD_0);  // for putfield
      //  code.op(DUP);      // for arg to make
      //  code.op2(INVOKESTATIC, method(selfName + "Peer.make(L" + className + ";)L" + className + "Peer;"));
      //  code.op2(PUTFIELD, peerField.ref());
      //}

      if (m == null)
        code.Inst(PERWAPI.Op.ret);
      else
        new FCodeEmit(this, m, code).emit();
    }

    internal void emitStaticInit(FMethod m)
    {
      hasStaticInit = true;
      PERWAPI.CILInstructions code = cctor.CreateCodeBuffer();

      // set $Type field with type (if we this is a closure,
      // then the MethodType will be the type exposed)
      if (!parent.isMixin().val)
      {
        Type t = parent;
        if (parent.@base() is MethodType) t = parent.@base();

        code.ldstr(t.signature().val);
        PERWAPI.Method findType = emitter.findMethod("Fan.Sys.Type", "find",
          new string[] { "System.String" }, "Fan.Sys.Type");
        code.MethInst(PERWAPI.MethodOp.call, findType);
        code.FieldInst(PERWAPI.FieldOp.stsfld, typeField);
      }

      if (m == null)
        code.Inst(PERWAPI.Op.ret);
      else
        new FCodeEmit(this, m, code).emit();
    }

    internal void emitTypeConstFields()
    {
      // if during the emitting of all the methods we ran across a non-sys
      // LoadType opcode, then we need to generate a static field called
      // type${pod}${name} we can use to cache the type once it is looked up
//      if (typeLiteralFields == null) return;
//      Iterator it = typeLiteralFields.values().iterator();
//      while (it.hasNext())
//      {
//        String fieldName = (String)it.next();
//        emitField(fieldName, "Lfan/sys/Type;", EmitConst.PRIVATE|EmitConst.STATIC);
//      }
    }

    private void emitMixinRouters(Type type)
    {
      // generate router method for each concrete instance method
      List methods = type.methods();
      for (int i=0; i<methods.sz(); i++)
      {
        Method m = (Method)methods.get(i);

        // only emit router for non-abstract instance methods
        if (m.isStatic().val || m.isAbstract().val) continue;

        // only emit the router unless this is the exact one I inherit
        string name = m.name().val;
        if (parent.slot(name, true).parent() != type) continue;

        // do it
        new FMethodEmit(this).emitMixinRouter(m);
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Utils
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Map Fan flags to .NET field flags.  Note we emit protected as
    /// public and internal/private as package-private so that we don't
    /// need to deal with scope issues for accessors like closures and
    /// helper classes.
    /// </summary>
    internal static PERWAPI.FieldAttr fieldFlags(int fflags)
    {
      PERWAPI.FieldAttr nflags = 0;
      if ((fflags & FConst.Private)   != 0) nflags |= PERWAPI.FieldAttr.Assembly;
      if ((fflags & FConst.Protected) != 0) nflags |= PERWAPI.FieldAttr.Public;
      if ((fflags & FConst.Public)    != 0) nflags |= PERWAPI.FieldAttr.Public;
      if ((fflags & FConst.Internal)  != 0) nflags |= PERWAPI.FieldAttr.Assembly;
      if ((fflags & FConst.Static)    != 0) nflags |= PERWAPI.FieldAttr.Static;
      return nflags;
    }

    /// <summary>
    /// Given a Fan qname index, map to a .NET type name: sys/Bool
    /// </summary>
    internal string nname(int index)
    {
      return pod.nname(index);
    }

    /// <summary>
    /// Map a simple name index to it's string value
    /// </summary>
    internal string name(int index)
    {
      return pod.name(index);
    }

    /*
    /// <summary>
    /// Map a simple name index to it's string value
    /// </summary>
    internal string Name(int index)
    {
      return pod.Name(index);
    }

    /// <summary>
    /// Get method ref to sys::Sys.type(String, bool)
    /// </summary>
    internal int SysFindType()
    {
      if (sysFindType == 0)
        sysFindType = method("fan/sys/Sys.findType(Ljava/lang/String;Z)Lfan/sys/Type;");
      return sysFindType ;
    }
    private int sysFindType;
    */

  //////////////////////////////////////////////////////////////////////////
  // Cached CpInfo
  //////////////////////////////////////////////////////////////////////////

    internal PERWAPI.Method CompareEQ, CompareEQz;
    internal PERWAPI.Method CompareNE, CompareNEz;
    internal PERWAPI.Method Compare;
    internal PERWAPI.Method CompareLT, CompareLTz;
    internal PERWAPI.Method CompareLE, CompareLEz;
    internal PERWAPI.Method CompareGE, CompareGEz;
    internal PERWAPI.Method CompareGT, CompareGTz;
    internal PERWAPI.Method CompareSame;
    internal PERWAPI.Method CompareNotSame;
    internal PERWAPI.Method CompareNull;
    internal PERWAPI.Method CompareNotNull;
    internal PERWAPI.Method IsViaType;
    internal PERWAPI.Method AsViaType;
    internal PERWAPI.Method BoolMake;
    internal PERWAPI.Field BoolTrue;
    internal PERWAPI.Field BoolFalse;
    internal PERWAPI.Field BoolVal;
    internal PERWAPI.Field IntVal;
    internal PERWAPI.Method ErrMake;
    internal PERWAPI.Field ErrVal;

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public string className;
    public string baseClassName;
    public string[] interfaces;

    internal PERWAPI.MethodDef ctor;   // cached .ctor emit object
    internal PERWAPI.MethodDef cctor;  // cahced .cctor emit object

    internal Emitter emitter;              // The emitter for the assembly for this type
    internal Type parent;
    internal FPod pod;
    internal FType type;
    internal string selfName;              // type name to use as self (for mixin body - this is interface)
    internal PERWAPI.Field typeField;      // private static final Type $Type
    //internal FieldEmit peerField;          // public static final TypePeer peer
    internal bool hasInstanceInit;         // true if we already emitted <init>
    internal bool hasStaticInit;           // true if we already emitted <clinit>
    internal MethodType methodType;        // if type is a closure
    internal Hashtable typeLiteralFields;  // signature Strings we need to turn into cached fields
    internal bool isNative = false;        // do we have any native methods requiring a peer
    internal bool isAbstract;              // are we emitting an abstract method

    private static Hashtable assemblies = new Hashtable();  // assembly cache

  }
}