//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   11 Sep 06  Andy Frank  Creation
//

using System;
using System.Text;
using Fan.Sys;
using Uri = Fan.Sys.Uri;
using Fanx.Util;

namespace Fanx.Fcode
{
  ///
  /// FPod is the read/write fcode representation of sys::Pod.  It's main job in
  /// life is to manage all the pod-wide constant tables for names, literals,
  /// type/slot references and type/slot definitions.
  ///
  public sealed class FPod
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    public FPod(string podName, FStore store)
    {
      this.m_podName    = podName;
      this.m_store      = store;
      this.m_names      = new FTable.Names(this);
      this.m_typeRefs   = new FTable.TypeRefs(this);
      this.m_fieldRefs  = new FTable.FieldRefs(this);
      this.m_methodRefs = new FTable.MethodRefs(this);
      this.m_ints       = new FTable.Ints(this);
      this.m_floats     = new FTable.Floats(this);
      this.m_strs       = new FTable.Strs(this);
      this.m_durations  = new FTable.Durations(this);
      this.m_uris       = new FTable.Uris(this);
    }


  //////////////////////////////////////////////////////////////////////////
  // Lookup
  //////////////////////////////////////////////////////////////////////////

    public FType type(string nameToFind)
    {
      for (int i=0; i<m_types.Length; i++)
        if (name(typeRef(m_types[i].m_self).typeName) == nameToFind)
          return m_types[i];
      throw UnknownTypeErr.make(nameToFind).val;
    }

  //////////////////////////////////////////////////////////////////////////
  // Convenience
  //////////////////////////////////////////////////////////////////////////

    public string name(int index)       { return (string)m_names.get(index); }
    public FTypeRef typeRef(int index)  { return (FTypeRef)m_typeRefs.get(index); }
    public FTuple fieldRef(int index)   { return (FTuple)m_fieldRefs.get(index); }
    public FTuple methodRef(int index)  { return (FTuple)m_methodRefs.get(index); }
    public Int integer(int index)       { return (Int)m_ints.get(index); }
    public Float floats(int index)      { return (Float)m_floats.get(index); }
    public Str str(int index)           { return (Str)m_strs.get(index); }
    public Duration duration(int index) { return (Duration)m_durations.get(index); }
    public Uri uri(int index)           { return (Uri)m_uris.get(index); }

  //////////////////////////////////////////////////////////////////////////
  // .NET Emit Utils
  //////////////////////////////////////////////////////////////////////////

    public class NMethod
    {
      public string parentType;
      public string methodName;
      public string returnType;
      public string[] paramTypes;
      public bool isStatic = false;
    }

    /// <summary>
    /// Map a fcode method signature to a .NET method emit signature.
    /// </summary>
    public NMethod ncall(int index, int opcode)
    {
      if (m_ncalls == null) m_ncalls = new NMethod[m_methodRefs.size()];
      NMethod x = m_ncalls[index];
      if (x == null)
      {
        int[] m = methodRef(index).val;
        string parent = nname(m[0]);
        string mName = /*NameUtil.Upper(*/name(m[1]);/*);*/
        bool onObj = parent == "Fan.Sys.Obj";
        string[] pars = new string[m.Length-3];
        for (int i=0; i<pars.Length; i++) pars[i] = nname(m[i+3]);

        // static methods on sys::Obj are really FanObj
        if (onObj && (opcode == FConst.CallStatic || opcode == FConst.CallNonVirtual))
        {
          parent = "Fan.Sys.FanObj";
        }

        string ret = nname(m[2]);
        if (opcode == FConst.CallNew) ret = parent; // factory

        // Handle static mixin calls
        if (opcode == FConst.CallMixinStatic) parent += "_";

        x = new NMethod();
        x.parentType = parent;
        x.methodName = mName;
        x.returnType = ret;
        x.paramTypes = pars;
        if (!onObj) m_ncalls[index] = x;
      }
      return x;
    }

    public class NField
    {
      public string parentType;
      public string fieldName;
      public string fieldType;
    }

    /// <summary>
    /// Map a fcode field signature to a .NET field emit signature.
    /// </summary>
    public NField nfield(int index, bool mixin)
    {
      if (m_nfields == null) m_nfields = new NField[m_fieldRefs.size()];
      NField nfield = m_nfields[index];
      if (nfield == null)
      {
        int[] v = fieldRef(index).val;
        nfield = new NField();
        nfield.parentType = nname(v[0]);
        nfield.fieldName  = name(v[1]);
        nfield.fieldType  = nname(v[2]);
        // TODO - put in map???
      }
      return nfield;
    }

    /// <summary>
    /// Map a .NET type name (:: replaced wit .) via typeRefs table.
    /// <summary>
    public string nname(int index)
    {
      if (m_nnames == null) m_nnames = new string[m_typeRefs.size()];
      string n = m_nnames[index];
      if (n == null)
      {
        FTypeRef refer = typeRef(index);
        n = m_nnames[index] = "Fan." + NameUtil.upper(name(refer.podName)) + '.' + name(refer.typeName);
      }
      return n;
    }

  //////////////////////////////////////////////////////////////////////////
  // Read
  //////////////////////////////////////////////////////////////////////////

    ///
    /// Read from a FStore which provides random access.
    ///
    public void read()
    {
      m_names.read(m_store.read("names.def", true));
      m_typeRefs.read(m_store.read("typeRefs.def"));
      m_fieldRefs.read(m_store.read("fieldRefs.def"));
      m_methodRefs.read(m_store.read("methodRefs.def"));
      m_ints.read(m_store.read("ints.def"));
      m_floats.read(m_store.read("floats.def"));
      m_strs.read(m_store.read("strs.def"));
      m_durations.read(m_store.read("durations.def"));
      m_uris.read(m_store.read("uris.def"));

      // pod meta
      readPodMeta(m_store.read("pod.def", true));

      // type meta
      readTypeMeta(m_store.read("types.def", true));

      // full fcode always lazy loaded in Type.reflect()
    }

    private void readPodMeta(FStore.Input input)
    {
      int first2 = input.u2();
      if (first2 != 0x0FC0)
      {
        System.Console.WriteLine("FPod.readPodMeta: old version of " + m_podName);
        m_version = "0";
        m_depends = new Depend[0];
        m_facets  = new FFacet[0];
        return;
      }
      if (input.u2() != 0xDE05)
        throw new System.IO.IOException("Invalid magic");

      int ver = input.u4();
      if (ver != FConst.FCodeVersion)
        throw new System.IO.IOException("Invalid version 0x" + ver.ToString("x"));

      m_podName = input.utf();
      m_version = input.utf();
      m_depends = new Depend[input.u1()];
      for (int i=0; i<m_depends.Length; i++)
        m_depends[i] = Depend.parse(Fan.Sys.Str.make(input.utf()));

      m_facets = new FFacet[input.u2()];
      for (int i=0; i<m_facets.Length; i++)
        m_facets[i] = new FFacet().read(input);

      input.Close();
    }

    private void readTypeMeta(FStore.Input input)
    {
      m_types = new FType[input.u2()];
      for (int i=0; i<m_types.Length; i++)
      {
        m_types[i] = new FType(this).readMeta(input);
        m_types[i].m_hollow = true;
      }
      input.Close();
    }

    /*
    private void ReadType(String name, FStore.Input input)
    {
      if (types == null || types.length == 0)
        throw new IOException("types.def must be defined first");

      String typeName = name.substring(0, name.length()-".fcode".length());
      for (int i=0; i<types.length; ++i)
      {
        String n = name(typeRef(types[i].self).typeName);
        if (n.equals(typeName )) { types[i].read(input); return; }
      }

      throw new IOException("Unexpected fcode file: " + name);
    }
    */

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    public string m_podName;     // pod's unique name
    public string m_version;     // pod version
    public Depend[] m_depends;   // pod dependencies
    public FFacet[] m_facets;    // pod facets
    public FStore m_store;       // store we using to read/write
    public FType[] m_types;      // pod's declared types
    public FTable m_names;       // identifier names: foo
    public FTable m_typeRefs;    // types refs:   [pod,type,variances*]
    public FTable m_fieldRefs;   // fields refs:  [parent,name,type]
    public FTable m_methodRefs;  // methods refs: [parent,name,ret,params*]
    public FTable m_ints;        // Int literals
    public FTable m_floats;      // Float literals
    public FTable m_strs;        // Str literals
    public FTable m_durations;   // Duration literals
    public FTable m_uris;        // Uri literals
    private string[] m_nnames;   // cached fan typeRef   -> .net name
    private NMethod[] m_ncalls;  // cached fan methodRef -> .net method signatures
    private NField[] m_nfields;  // cached fan fieldRef  -> .net field signatures
  }
}