//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   11 Sep 06  Andy Frank  Creation
//

using System;
using System.Collections;
using System.IO;
using System.Text;
using ICSharpCode.SharpZipLib.Zip;
using Fan.Sys;

namespace Fanx.Fcode
{
  /// <summary>
  /// FStore models IO streams to use for reading and writing pod files.
  /// </summary>
  public sealed class FStore
  {

  //////////////////////////////////////////////////////////////////////////
  // Constructor
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// Construct a FStore to read.
    /// </summary>
    public static FStore openToRead(string path)
    {
      return new FStore(new ZipFile(path));
    }

    private FStore(ZipFile zipFile) { this.zipFile = zipFile; }

  //////////////////////////////////////////////////////////////////////////
  // File Access
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// List all the files available.
    /// </summary>
    public string[] list()
    {
      ArrayList acc = new ArrayList();
      IEnumerator en = zipFile.GetEnumerator();
      while (en.MoveNext())
        acc.Add(((ZipEntry)en.Current).Name);
      return (String[])acc.ToArray(System.Type.GetType("System.String"));
    }

    /// <summary>
    /// Return a map to use for Pod.files()
    /// </summary>
    public Map podFiles()
    {
      Map map = new Map(Sys.UriType, Sys.FileType);
      IEnumerator en = zipFile.GetEnumerator();
      while (en.MoveNext())
      {
        ZipEntry entry = (ZipEntry)en.Current;
        string name = entry.Name;
        if (name.EndsWith(".fcode")) continue;
        if (name.EndsWith(".class")) continue;
        Fan.Sys.ZipEntryFile file = new Fan.Sys.ZipEntryFile(zipFile, entry);
        map.set(file.uri(), file);
      }
      return map;
    }

    /// <summary>
    /// Convenience for read(path, false).
    /// </summary>
    public FStore.Input read(string path)
    {
      return read(path, false);
    }

    /// <summary>
    /// Open an input stream for the specified logical path.
    /// Return null if not found.
    /// </summary>
    public FStore.Input read(string path, bool required)
    {
      if (zipFile == null) throw new IOException("FStore not opened to read");
      ZipEntry entry = zipFile.GetEntry(path);
      if (entry == null)
      {
        if (required)
          throw new Exception("Missing required file \"" + path + "\" in pod zup");
        else
          return null;
      }
      return new FStore.Input(new BufferedStream(zipFile.GetInputStream(entry)));
    }

    /// <summary>
    /// Close this FStore, which should release all file locks
    /// on the pod file. This method exists for testing purposes,
    /// and should not otherwise be used.
    /// </summary>
    public void close()
    {
      zipFile.Close();
    }

  //////////////////////////////////////////////////////////////////////////
  // FStore.Input
  //////////////////////////////////////////////////////////////////////////

    /// <summary>
    /// FStore.Input is used to read from a FStore file.
    /// </summary>
    public class Input : BinaryReader
    {
      public Input(Stream baseStream) : base(baseStream) {}

      public int u1()  { return (int)ReadByte(); }
      public int u2()
      {
        int i = (int)ReadByte() << 8;
        i |= (int)ReadByte();
        return i;
      }
      public int u4()
      {
        int i = (int)ReadByte() << 24;
        i |= (int)ReadByte() << 16;
        i |= (int)ReadByte() << 8;
        i |= (int)ReadByte();
        return i;
      }
      public long u8()
      {
        long i = (long)ReadByte() << 56;
        i |= (long)ReadByte() << 48;
        i |= (long)ReadByte() << 40;
        i |= (long)ReadByte() << 32;
        i |= (long)ReadByte() << 24;
        i |= (long)ReadByte() << 16;
        i |= (long)ReadByte() << 8;
        i |= (long)ReadByte();
        return i;
      }
      public double f8()
      {
        byte[] b = new byte[8];
        b[7] = ReadByte();
        b[6] = ReadByte();
        b[5] = ReadByte();
        b[4] = ReadByte();
        b[3] = ReadByte();
        b[2] = ReadByte();
        b[1] = ReadByte();
        b[0] = ReadByte();
        return BitConverter.ToDouble(b, 0);
      }
      public string utf()
      {
        // TODO - more efficient way?
        int len = u2();
        if (len == 0) return "";
        byte[] buf = new byte[len];
        Read(buf, 0, len);
        return new UTF8Encoding().GetString(buf);
      }
    }

  //////////////////////////////////////////////////////////////////////////
  // Fields
  //////////////////////////////////////////////////////////////////////////

    private ZipFile zipFile;
  }
}