//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   16 Mar 06  Brian Frank  Creation
//

**
** Buf is used to model a buffer of bytes with random access.  Buf is
** typically backed by a file or a block of memory.  Buf implements the
** InStream and OutStream mixins to read and write into the buffer using
** a configurable position accessed via pos() and seek().
**
** When used as an InStream, bytes are read starting at pos with pos
** advanced after each read.  The end of stream is reached when pos reaches
** size.  When used as an OutStream, bytes are written starting at pos
** with pos advanced after each write.  If pos is less then size then
** the existing bytes are rewritten and size is not advanced, otherwise
** we automatically grow the buffer and advance size as bytes are appended.
** It is common to write bytes into the buffer as an OutStream, then
** call flip() to prepare the buffer to be used for reading.
**
class Buf : InStream, OutStream
{

//////////////////////////////////////////////////////////////////////////
// Constructors
//////////////////////////////////////////////////////////////////////////

  **
  ** Make a byte buffer with the initial given capacity.
  **
  new make(Int capacity := 1024)

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  **
  ** Two Bufs are equal if they have string of bytes.
  **
  override Bool equals(Obj that)

  **
  ** Return hash code based on string of bytes.
  **
  override Int hash()

  **
  ** Return hex string of bytes.
  **
  override Str toStr()

//////////////////////////////////////////////////////////////////////////
// Access
//////////////////////////////////////////////////////////////////////////

  **
  ** Return if size() == 0.
  **
  Bool empty()

  **
  ** Return the total number of bytes in the buffer.  If the size is
  ** set greater than capacity then the buffer's capacity is automatically
  ** grown, otherwise capacity remains the same.  Setting size does not
  ** actually change any bytes in the buffer.
  **
  virtual Int size

  **
  ** The number of bytes this buffer can hold without allocating more
  ** memory.  Capacity is always greater or equal to size.  If adding a
  ** large number of bytes, it may be more efficient to manually set
  ** capacity.  See the trim() method to automatically set capacity to
  ** size.  Throw ArgErr if attempting to set capacity less than size.
  **
  virtual Int capacity

  **
  ** Return the current position for the next read or write.  The
  ** position is always between 0 and size().  If pos is less then
  ** size then future writes will rewrite the existing bytes without
  ** growing size.
  **
  virtual Int pos()

  **
  ** Return the remaining number of bytes to read: size-pos.
  **
  virtual Int remaining()

  **
  ** Return if more bytes are available to read: remaining() > 0.
  **
  virtual Bool more()

  **
  ** Set the current position to the specified byte offset.  A
  ** negative index may be used to access from the end of the buffer.
  ** For example seek(-1) is translated into seek(size-1).
  ** Return this.
  **
  virtual Buf seek(Int pos)

  **
  ** Flip a buffer from write-mode to read-mode.  This method sets
  ** total size to current position, and position to 0.  Return this.
  **
  virtual Buf flip()

  **
  ** Get the byte at the specified absolute index.  A negative index
  ** may be used to access from the end of the buffer.  For example
  ** get(-1)  is translated into get(size()-1).  This method accesses
  ** the buffer absolutely independent of current position.  The get
  ** method is accessed via the [] shortcut operator.  Throw IndexErr
  ** if index out of range.
  **
  virtual Int get(Int index)

  **
  ** Return a new buffer containing the bytes in the specified absolute
  ** range.  Negative indexes may be used to access from the end of
  ** the buf.  This method accesses the buffer absolutely independent
  ** of current position.  This method is accessed via the [] operator.
  ** Throw IndexErr if range illegal.
  **
  ** Examples:
  **   buf := Buf.make
  **   buf.write(0xaa).write(0xbb).write(0xcc).write(0xdd)
  **   buf[0..2]   -> 0x[aabbcc]
  **   buf[3..3]   -> 0x[dd]
  **   buf[-2..-1] -> 0x[ccdd]
  **   buf[0...2]  -> 0x[aabb]
  **   buf[1..-2]  -> 0x[bbcc]
  **
  virtual Buf slice(Range range)

//////////////////////////////////////////////////////////////////////////
// Modification
//////////////////////////////////////////////////////////////////////////

  **
  ** Set is used to overwrite the byte at the specified the index.  A
  ** negative index may be used to access an index from the end of the
  ** buffer.  The set method is accessed via the []= shortcut operator.
  ** Return this.  Throw IndexErr if index is out of range.
  **
  virtual Buf set(Int index, Int byte)

  **
  ** Read the buffer for a fresh read by reseting the buffer's pos
  ** and size to zero.  The buffer's capacity remains the same.
  **
  Buf clear()

  **
  ** Trim the capacity such that the underlying storage is optimized
  ** for the current size.  Return this.
  **
  Buf trim()

//////////////////////////////////////////////////////////////////////////
// Both
//////////////////////////////////////////////////////////////////////////

  override Bool close()
  override Charset charset

//////////////////////////////////////////////////////////////////////////
// OutStream
//////////////////////////////////////////////////////////////////////////

  override OutStream write(Int byte)
  override OutStream writeBuf(Buf buf, Int n := buf.remaining)
  override OutStream writeChar(Int char)
  override OutStream writeChars(Str str, Int off := 0, Int len := str.size-off)

//////////////////////////////////////////////////////////////////////////
// InStream
//////////////////////////////////////////////////////////////////////////

  override Int read()
  override InStream unread(Int n)
  override Int readBuf(Buf buf, Int n)
  override Int readChar()
  override InStream unreadChar(Int c)

}