//
// Copyright (c) 2007, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   30 Jan 07  Brian Frank  Creation
//

**
** ThreadTest
**
class ThreadTest : Test
{

//////////////////////////////////////////////////////////////////////////
// Identity
//////////////////////////////////////////////////////////////////////////

  Void testIdentity()
  {
    // make
    a := Thread.make("sysTest.foo")
    b := Thread.make("sysTest.bar")
    c := Thread.make
    d := Thread.make
    verifyErr(ArgErr.type) |,| { Thread.make("sysTest.foo") }
    verifyErr(ArgErr.type) |,| { Thread.make(c.name) }

    // name
    verifyEq(a.name, "sysTest.foo")
    verifyEq(b.name, "sysTest.bar")
    verify(c.name != d.name)

    // equals
    verify(a == a)
    verify(a != b)

    // hash
    verifyEq(a.hash, a.name.hash)
    verifyEq(b.hash, b.name.hash)

    // toStr
    verifyEq(a.toStr, "sysTest.foo")
    verifyEq(d.name, d.toStr)

    // find
    verifySame(Thread.find("sysTest.foo"), a)
    verifySame(Thread.find(d.name), d)
    verifyEq(Thread.find("sysTest.forget it", false), null)
    verifyErr(Err.type) |,| { Thread.find("sysTest.forget it") }
    verifyErr(Err.type) |,| { Thread.find("sysTest.forget it", true) }

    // list
    verify(Thread.list.isRO)
    verifyEq(Thread.list.type, Thread[].type)
    verify(Thread.list.contains(a))
    verify(Thread.list.contains(b))
    verify(Thread.list.contains(c))
    verify(Thread.list.contains(d))

    // current
    verifyEq(Thread.current.name, "main")
    verifyEq(Thread.current.isNew, false)
    verifyEq(Thread.current.isRunning, true)
    verifyEq(Thread.current.isDead, false)
  }

//////////////////////////////////////////////////////////////////////////
// Count
//////////////////////////////////////////////////////////////////////////

  Void testCount()
  {
    // create it
    t := ThreadCount.make
    verifyEq(t.name, "ThreadCount")
    verifyEq(t.isNew, true)
    verifyEq(t.isRunning, false)
    verifyEq(t.isDead, false)
    verifySame(Thread.find("ThreadCount"), t)
    verifyErr(ArgErr.type) |,| { ThreadCount.make }

    // stop it
    t.stop
    verifyEq(t.isNew, false)
    verifyEq(t.isRunning, false)
    verifyEq(t.isDead, true)
    verifyEq(Thread.find("ThreadCount", false), null)
    old := t

    // create new one
    t = ThreadCount.make
    verify(t !== old)
    verifyEq(t.name, "ThreadCount")
    verifyEq(t.isNew, true)

    // start it
    t.start
    verifyEq(t.isNew, false)
    verifyEq(t.isRunning, true)
    verifyEq(t.isDead, false)
    verifySame(Thread.find("ThreadCount"), t)
    verifyErr(Err.type) |,| { t.start }

    // join until counter done
    Thread.sleep(500ms)
    verifyEq(t.isNew, false)
    verifyEq(t.isRunning, false)
    verifyEq(t.isDead, true)
    verifyEq(Thread.find("ThreadCount", false), null)
    verifyErr(Err.type) |,| { t.start }
    t.join  // ignored
    t.stop  // ignored
  }

//////////////////////////////////////////////////////////////////////////
// Messaging
//////////////////////////////////////////////////////////////////////////

  Void testMessaging()
  {
    // verify run is checked for immutable
    verifyErr(NotImmutableErr.type) |,|
    {
      Thread.make("sysTest.testMessaging") |Thread thread| { fail }
    }

    // create thread which takes an Int msg and doubles it
    t := Thread.make("sysTest.testMessaging", &runMessaging)
    t.start

    // verify this thread can't call t's loop
    verifyErr(Err.type) |,| { t.loop |Obj m->Obj| { return m } }

    // send some sync messages
    10.times |Int i|
    {
      r := (Int)t.sendSync(i)
      verifyEq(r, i*2)
    }

    // verify exception raised to me
    verifyErr(IOErr.type) |,| { t.sendSync("throw") }

    // verify messages must immutable
    verifyErr(NotImmutableErr.type) |,| { t.sendSync(this) }
    verifyErr(NotImmutableErr.type) |,| { t.sendAsync(this) }
    verifyErr(NotImmutableErr.type) |,| { t.sendSync("mutable") }
    //t.sendAsync("mutable")
    //t.sendAsync("throw")
    t.sendSync("die")

    verify(t.isDead)
  }

  static Void runMessaging(Thread t)
  {
    t.loop |Obj msg->Obj|
    {
      if (msg is Int) return (Int)msg * 2

      switch (msg)
      {
        case "throw":    throw IOErr.make()
        case "mutable":  return [0, 1, 2]
        case "die":      t.stop; return null
      }
      return null
    }
  }

//////////////////////////////////////////////////////////////////////////
// Join
//////////////////////////////////////////////////////////////////////////

  Void testJoin()
  {
    t := Thread.make(null, &runJoin("six")).start
    verifyEq(t.join, 6)
    verifyEq(t.join, null)

    t = Thread.make(null, &runJoin("mutable")).start
    verifyEq(t.join, Buf.make.print("abc"))
    verifyEq(t.join, null)
  }

  static Obj runJoin(Str mode)
  {
    switch (mode)
    {
      case "six":     return 6
      case "mutable": return Buf.make.print("abc")
      default:        throw UnsupportedErr.make
    }
  }

//////////////////////////////////////////////////////////////////////////
// Sync Stop
//////////////////////////////////////////////////////////////////////////

  Void testSyncStop()
  {
    verifySyncStop(true)
    verifySyncStop(false)
  }

  Void verifySyncStop(Bool explicitStop)
  {
    x := Thread.make(null, &runSyncStopMain).start
    a := Thread.make(null, &runSyncStop(x)).start
    b := Thread.make(null, &runSyncStop(x)).start
    c := Thread.make(null, &runSyncStop(x)).start

    if (explicitStop)
    {
      Thread.sleep(50ms)
      x.stop
    }
    else
    {
      x.join
    }

    t1 := Duration.now
    verify(a.join is InterruptedErr)
    verify(b.join is InterruptedErr)
    verify(c.join is InterruptedErr)
    t2 := Duration.now
    verify(t2-t1 < 50ms)
  }

  static Void runSyncStopMain(Thread t)
  {
    try
    {
      Thread.sleep(100ms)
    }
    catch (InterruptedErr e)
    {
    }
  }

  static Obj runSyncStop(Thread main, Thread t)
  {
    try
    {
      main.sendSync("x")
      return null
    }
    catch (Err e)
    {
      return e
    }
  }

//////////////////////////////////////////////////////////////////////////
// Thread Locals
//////////////////////////////////////////////////////////////////////////

  Void testLocals()
  {
    Thread.locals["sysTest.x"] = "main"
    verifyEq(Thread.locals["sysTest.x"], "main")

    a := Thread.make(null, &runLocal("alpha")).start
    b := Thread.make(null, &runLocal("beta")).start
    c := Thread.make(null, &runLocal("gamma")).start

    verifyEq(a.join, "alpha")
    verifyEq(b.join, "beta")
    verifyEq(c.join, "gamma")
    verifyEq(Thread.locals["sysTest.x"], "main")
  }

  static Obj runLocal(Str val)
  {
    r := val
    3.times |,|
    {
      Thread.locals["sysTest.x"] = val
      if (Thread.locals["sysTest.x"] != val)
      {
        Err.make.trace
        r = "bad bad bad"
      }
      Thread.sleep(10ms)
    }
    return r
  }

//////////////////////////////////////////////////////////////////////////
// Trace
//////////////////////////////////////////////////////////////////////////

  Void testTrace()
  {
    buf := Buf.make

    // trace new and dead thread
    t := Thread.make(null)
    t.trace(buf)
    t.stop
    t.trace(buf)
    verifyEq(buf.size, 0)

    // trace current
    traceA(buf)

    lines := buf.flip.readAllLines
    // lines.each |Str line| { echo(line) }

    verify(lines[0].contains("Thread.trace("))
    verify(lines[1].contains("ThreadTest.traceB("))
    verify(lines[2].contains("ThreadTest.traceA("))
    verify(lines[3].contains("ThreadTest.testTrace("))
  }

  Void traceA(OutStream out) { traceB(out) }
  Void traceB(OutStream out) { Thread.current.trace(out) }

}

**************************************************************************
** ThreadCount
**************************************************************************

const class ThreadCount : Thread
{
  new make() : super("ThreadCount") {}

  override Obj run()
  {
    3.times |Int c| { sleep(100ms) }
    return null
  }
}