//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   24 Mar 06  Brian Frank  Creation
//

**
** TypeTest
**
class TypeTest : Test
{

//////////////////////////////////////////////////////////////////////////
// Find
//////////////////////////////////////////////////////////////////////////

  Void testFind()
  {
    verifySame(Type.find("sys::Int"), Int.type)
    verifySame(Type.find("sys::Str[]"), Str[].type)
    verifySame(Type.find("sys::notHereFoo", false), null)
    verifyErr(UnknownTypeErr.type) |,| { Type.find("sys::notHereFoo") }
    verifyErr(UnknownPodErr.type) |,| { Type.find("notHereFoo::Duh") }
    verifyErr(ArgErr.type) |,| { Type.find("sys") }
    verifyErr(ArgErr.type) |,| { Type.find("sys::") }
    verifyErr(ArgErr.type) |,| { Type.find("::sys") }
    verifyErr(ArgErr.type) |,| { Type.find("[]") }
  }

//////////////////////////////////////////////////////////////////////////
// Flags
//////////////////////////////////////////////////////////////////////////

  Void testFlags()
  {
    // isAbstract
    verifyEq(Test.type.isAbstract, true)
    verifyEq(type.isAbstract, false)

    // isClass
    verifyEq(type.isClass, true)
    verifyEq(EnumAbc.type.isClass, false)
    verifyEq(OutStream.type.isClass, false)

    // isEnum
    verifyEq(type.isEnum, false)
    verifyEq(EnumAbc.type.isEnum, true)
    verifyEq(OutStream.type.isEnum, false)

    // isFinal
    verifyEq(Bool.type.isFinal, true)
    verifyEq(Test.type.isFinal, false)

    // isInternal
    verifyEq(type.isInternal, false)
    verifyEq(EnumAbc.type.isInternal, true)

    // isMixin
    verifyEq(type.isMixin, false)
    verifyEq(EnumAbc.type.isMixin, false)
    verifyEq(OutStream.type.isMixin, true)

    // isPublic
    verifyEq(type.isPublic, true)
    verifyEq(EnumAbc.type.isPublic, false)

    // isSynthetic
    // test below in testSynthetic()
  }

//////////////////////////////////////////////////////////////////////////
// Mixins
//////////////////////////////////////////////////////////////////////////

  Void testMixins()
  {
    verifyEq(Obj.type.mixins, Type[,])
    verifyEq(Obj.type.mixins.isRO, true)

    verifyEq(Buf.type.mixins, [InStream.type, OutStream.type])
    verifyEq(Buf.type.mixins.isRO, true)
  }

//////////////////////////////////////////////////////////////////////////
// Inheritance
//////////////////////////////////////////////////////////////////////////

  Void testInheritance()
  {
    // technically the list order shouldn't be required, but I think we
    // should maintain consistent order; I place Obj 1st always for
    // performance in fits() method
    verifyEq(Obj.type.inheritance, [Obj.type])
    verifyEq(Num.type.inheritance, [Obj.type, Num.type])
    verifyEq(Int.type.inheritance, [Obj.type, Int.type, Num.type])

    // TODO: add mixin types
  }

//////////////////////////////////////////////////////////////////////////
// Fits
//////////////////////////////////////////////////////////////////////////

  Void testFits()
  {
    verify(Float.type.fits(Float.type))
    verify(Float.type.fits(Num.type))
    verify(Float.type.fits(Obj.type))
    verifyFalse(Float.type.fits(Str.type))
    verifyFalse(Obj.type.fits(Float.type))

    // TODO: add mixin types
  }

//////////////////////////////////////////////////////////////////////////
// Generics
//////////////////////////////////////////////////////////////////////////

  Void testIsGeneric()
  {
    verifyEq(Obj.type.isGeneric,    false)
    verifyEq(Str.type.isGeneric,    false)
    verifyEq(Str[].type.isGeneric,  false)
    verifyEq(List.type.isGeneric,   true)
    verifyEq(Map.type.isGeneric,    true)
    verifyEq(Method.type.isGeneric, true)
  }

  Void testParams()
  {
    verifyEq(Str.type.params, Str:Type[:])
    verifyEq(Str.type.params.isRO, true)

    verifyEq(Str[].type.params, ["V":Str.type, "L":Str[].type])
    verifyEq(Str[].type.params.isRO, true)

    verifyEq(Int:Slot[].type.params, ["K":Int.type, "V":Slot[].type, "M":Int:Slot[].type])
    verifyEq(Int:Slot[].type.params.isRO, true)

    verifyEq(|Int a, Float b->Bool|.type.params, ["A":Int.type, "B":Float.type, "R":Bool.type])
    verifyEq(|Int a, Float b->Bool|.type.params.isRO, true)
  }

  Void testParameterization()
  {
    verifyEq(List.type.parameterize(["V":Bool.type]), Bool[].type)
    verifyEq(List.type.parameterize(["V":Bool[].type]), Bool[][].type)
    verifyErr(ArgErr.type) |,| { List.type.parameterize(["X":Bool[].type]) }

    verifyEq(Map.type.parameterize(["K":Str.type, "V":Slot.type]), Str:Slot.type)
    verifyEq(Map.type.parameterize(["K":Str.type, "V":Int[].type]), Str:Int[].type)
    verifyErr(ArgErr.type) |,| { Map.type.parameterize(["V":Bool[].type]) }
    verifyErr(ArgErr.type) |,| { Map.type.parameterize(["K":Bool[].type]) }

    verifyEq(Method.type.parameterize(["R":Void.type]), |,|.type)
    verifyEq(Method.type.parameterize(["A":Str.type, "R":Int.type]), |Str a->Int|.type)
    verifyEq(Method.type.parameterize(["A":Str.type, "B":Bool.type, "C":Int.type, "R":Float.type]),
      |Str a, Bool b, Int c->Float|.type)
    verifyErr(ArgErr.type) |,| { Method.type.parameterize(["A":Bool[].type]) }

    verifyErr(UnsupportedErr.type) |,| { Str.type.parameterize(["X":Void.type]) }
    verifyErr(UnsupportedErr.type) |,| { Str[].type.parameterize(["X":Void.type]) }
  }

  Void testToListOf()
  {
    verifyEq(Str.type.toListOf, Str[].type)
    verifyEq(Str[].type.toListOf, Str[][].type)
    verifyEq(Str[][].type.toListOf, Str[][][].type)
    verifyEq(Str:Buf.type.toListOf, [Str:Buf][].type)
  }

//////////////////////////////////////////////////////////////////////////
// Reflection
//////////////////////////////////////////////////////////////////////////

  Void testMake()
  {
    verify(File.type.make([`foo`]) is File)
    verifyEq(File.type.make([`foo`])->uri, `foo`)
    //verifyErr(Err.type) |,| { Bool.type.make }
  }

//////////////////////////////////////////////////////////////////////////
// Synthetic
//////////////////////////////////////////////////////////////////////////

  Void testSynthetic()
  {
    type.pod.types.each |Type t|
    {
      verifyEq(t.isSynthetic, t.name.index("\$") != null, t.toStr)
      verifySlotsSynthetic(t)
    }
  }

  Void verifySlotsSynthetic(Type t)
  {
    t.slots.each |Slot slot|
    {
      if (slot.parent.isSynthetic || slot.name.index("\$") != null)
        verify(slot.isSynthetic, slot.toStr)
    }
  }

//////////////////////////////////////////////////////////////////////////
// Generic Parameters
//////////////////////////////////////////////////////////////////////////

  Void testGenericParameters()
  {
    // TODO - what do we really want for these guys?
    v := List.type.method("get").returnType
    verifyEq(v.name, "V")
    verifyEq(v.qname, "sys::V")
    verifySame(v.pod, Obj.type.pod)
    verifySame(v.base, Obj.type)
    verifyEq(v.mixins, Type[,])
    verifyEq(v.mixins.ro, Type[,])
  }

}