//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   28 Jun 06  Brian Frank  Creation
//

**
** UriTest
**
class UriTest : Test
{

//////////////////////////////////////////////////////////////////////////
// Equals
//////////////////////////////////////////////////////////////////////////

  Void testEquals()
  {
    // equals
    verifyEq(Uri.parse("http://u@foo:88/a/b?q#f"), Uri.parse("http://u@foo:88/a/b?q#f"));
    verifyEq(`http://foo/`, Uri.parse("http://foo/"));
    verifyEq(`http://u@foo:88/a/b?q#f`, `http://u@foo:88/a/b?q#f`);
    verifySame(`http://u@foo:88/a/b?q#f`, `http://u@foo:88/a/b?q#f`);

    // case difference in each component
    verifyNotEq(`Http://u@foo:88/a/b?q#f`, `http://u@foo:88/a/b?q#f`);
    verifyNotEq(`http://u@foo:88/A/b?q#f`, `http://u@foo:88/a/b?q#f`);
    verifyNotEq(`http://u@foo:88/a/b?Q#f`, `http://u@foo:88/a/b?q#f`);
    verifyNotEq(`http://u@foo:88/a/b?q#F`, `http://u@foo:88/a/b?q#f`);
    verifyNotEq(`http://Foo/`, `http://foo/`);
  }

//////////////////////////////////////////////////////////////////////////
// Str
//////////////////////////////////////////////////////////////////////////

  Void testStr()
  {
    // for now at least, str returns the exact string passed to parse()
    verifyEq(`HtTp://OneBigHost/foo`.toStr, "HtTp://OneBigHost/foo");
  }

//////////////////////////////////////////////////////////////////////////
// Kitchen Sink
//////////////////////////////////////////////////////////////////////////

  Void testKitchenSink()
  {
    uri := `http://user@host:80/a/b/c.txt?a=b&c=d#frag`
    verifyEq(uri.scheme,   "http")
    verifyEq(uri.auth,     "user@host:80")
    verifyEq(uri.host,     "host")
    verifyEq(uri.port,     80)
    verifyEq(uri.userInfo, "user")
    verifyEq(uri.path,     ["a", "b", "c.txt"])
    verifyEq(uri.pathStr,  "/a/b/c.txt")
    verifyEq(uri.name,     "c.txt")
    verifyEq(uri.basename, "c")
    verifyEq(uri.ext,      "txt")
    verifyEq(uri.query,    ["a":"b", "c":"d"])
    verifyEq(uri.queryStr, "a=b&c=d")
    verifyEq(uri.frag,     "frag")
  }

//////////////////////////////////////////////////////////////////////////
// Abs Defaults
//////////////////////////////////////////////////////////////////////////

  Void testAbsDefaults()
  {
    uri := `http://host/a/b/c`
    verifyEq(uri.scheme,   "http")
    verifyEq(uri.auth,     "host")
    verifyEq(uri.host,     "host")
    verifyEq(uri.port,     null)
    verifyEq(uri.userInfo, null)
    verifyEq(uri.path,     ["a", "b", "c"])
    verifyEq(uri.pathStr,  "/a/b/c")
    verifyEq(uri.name,     "c")
    verifyEq(uri.basename, "c")
    verifyEq(uri.ext,      null)
    verifyEq(uri.query,    null)
    verifyEq(uri.queryStr, null)
    verifyEq(uri.frag,     null)
  }

//////////////////////////////////////////////////////////////////////////
// Scheme
//////////////////////////////////////////////////////////////////////////

  Void testScheme()
  {
    verifyScheme(`http://foo/`, "http");
    verifyScheme(`HTTP://foo/`, "http");
    verifyScheme(`mailto:who@there.com`, "mailto");
    verifyScheme(`MailTo:who@there.com`, "mailto");
    verifyScheme(`/a/b/c`, null);
    verifyScheme(`../a/b/c`, null);
    verifyScheme(`#frag`, null);
  }

  Void verifyScheme(Uri uri, Str scheme)
  {
    verifyEq(uri.scheme, scheme)
    verifyEq(uri.isAbs, scheme != null)
    verifyEq(uri.isRel, scheme == null)
  }

//////////////////////////////////////////////////////////////////////////
// Body
//////////////////////////////////////////////////////////////////////////

  Void testBody()
  {
    // abs
    verifyEq(`http://foo/a/b/c`.body, "//foo/a/b/c");
    verifyEq(`mailto:who@there.com`.body, "who@there.com");

    // relative
    verifyEq(`//foo/a/b/c`.body, "//foo/a/b/c");
    verifyEq(`/a/b/c`.body, "/a/b/c");
    verifyEq(`../a/b/c`.body, "../a/b/c");
    verifyEq(`#frag`.body, "#frag");
  }

//////////////////////////////////////////////////////////////////////////
// Auth (UserInfo, Host, Port)
//////////////////////////////////////////////////////////////////////////

  Void testAuth()
  {
    // combinations
    verifyAuth(`http://brian:pass@host:88/`, "brian:pass@host:88", "brian:pass", "host", 88)
    verifyAuth(`http://brian:pass@host/`, "brian:pass@host", "brian:pass", "host", null)
    verifyAuth(`http://host:88/`, "host:88", null, "host", 88)
    verifyAuth(`http://host/`, "host", null, "host", null)

    // host names
    verifyAuth(`http://HOST/#frag`, "HOST", null, "HOST", null)
    verifyAuth(`foo://www.foo.com:8080/a?q`, "www.foo.com:8080", null, "www.foo.com", 8080)
    verifyAuth(`http://www.foo-bar.com/index`, "www.foo-bar.com", null, "www.foo-bar.com", null)

    // IPv4 addresses
    verifyAuth(`http://10.89.255.0/#frag`, "10.89.255.0", null, "10.89.255.0", null)
    verifyAuth(`http://user@129.5.255.90:90/a/b`, "user@129.5.255.90:90", "user", "129.5.255.90", 90)
    // TODO - Java silently ignores invalid IP addresses like 256.256.256.256
    // verifyAuth(`http://256.256.256.256/`, "256.256.256.256", null, "256.256.256.256", null)

    // IPv6 addresses
    verifyAuth(`http://u@[2001:0db8:0000:0000:0000:0000:1428:57ab]:81/`, "u@[2001:0db8:0000:0000:0000:0000:1428:57ab]:81", "u", "[2001:0db8:0000:0000:0000:0000:1428:57ab]", 81)
    verifyAuth(`ftp://[2001:0db8:0000:0000:0000:0000:1428:57ab]/`, "[2001:0db8:0000:0000:0000:0000:1428:57ab]", null, "[2001:0db8:0000:0000:0000:0000:1428:57ab]", null)
    verifyAuth(`http://[::192.9.5.5]/`, "[::192.9.5.5]", null, "[::192.9.5.5]", null)

    // relative
    verifyAuth(`//u@foo:0/a/b/c`, "u@foo:0", "u", "foo", 0);
    verifyAuth(`//foo/a/b/c`, "foo", null, "foo", null);
    verifyAuth(`/a/b/c`, null, null, null, null);
    verifyAuth(`../a/b/c`, null, null, null, null);
    verifyAuth(`#frag`, null, null, null, null);
  }

  Void verifyAuth(Uri uri, Str auth, Str userInfo, Str host, Int port)
  {
    verifyEq(uri.auth, auth)
    verifyEq(uri.userInfo, userInfo)
    verifyEq(uri.host, host)
    verifyEq(uri.port, port)
  }

//////////////////////////////////////////////////////////////////////////
// Path
//////////////////////////////////////////////////////////////////////////

  Void testPath()
  {
    // none
    verifyPath(`mailto:me@there.com`, null, null)

    // absolute
    verifyPath(`http://host`, "", Str[,])
    verifyPath(`http://host/`, "/", Str[,])
    verifyPath(`http://host/a`, "/a", ["a"])
    verifyPath(`http://host/a/b`, "/a/b", ["a", "b"])
    verifyPath(`http://host/a/b/`, "/a/b/", ["a", "b"])
    verifyPath(`http://host/_a/_b`, "/_a/_b", ["_a", "_b"])

    // auth absolute
    verifyPath(`/`, "/", Str[,])
    verifyPath(`//host/`, "/", Str[,])
    verifyPath(`//host/a`, "/a", ["a"])
    verifyPath(`//host/a/`, "/a/", ["a"])
    verifyPath(`//host/a/b`, "/a/b", ["a", "b"])
    verifyPath(`//host/_a/_b`, "/_a/_b", ["_a", "_b"])

    // path absolute
    verifyPath(`/`, "/", Str[,])
    verifyPath(`/a`, "/a", ["a"])
    verifyPath(`/a/b`, "/a/b", ["a", "b"])

    // path relative
    verifyPath(``, "", Str[,])
    verifyPath(`a`, "a", ["a"])
    verifyPath(`a/b`, "a/b", ["a", "b"])

    // path dot
    verifyPath(`.`, ".", ["."])
    verifyPath(`./a`, "./a", [".", "a"])
    verifyPath(`./a/b`, "./a/b", [".", "a", "b"])

    // path backup
    verifyPath(`..`, "..", [".."])
    verifyPath(`../a`, "../a", ["..", "a"])
    verifyPath(`../a/b`, "../a/b", ["..", "a", "b"])

    // path backup
    verifyPath(`../..`, "../..", ["..", ".."])
    verifyPath(`../../a`, "../../a", ["..", "..", "a"])
    verifyPath(`../../a/b`, "../../a/b", ["..", "..", "a", "b"])
  }

  Void verifyPath(Uri uri, Str pathStr, Str[] path)
  {
    verifyEq(uri.pathStr, pathStr)
    verifyEq(uri.path, path)
    if (path == null)
    {
      verifyEq(uri.isPathAbs, false)
    }
    else
    {
      verifyEq(uri.isPathAbs, pathStr.startsWith("/"))
      verifyEq(uri.path.isRO, true)
      verifyEq(uri.isDir, pathStr.size > 0 && pathStr[-1] == '/')
    }
  }

//////////////////////////////////////////////////////////////////////////
// Name
//////////////////////////////////////////////////////////////////////////

  Void testName()
  {
    // none
    verifyName(`mailto:me@there.com`, null, null, null)

    // absolute roots
    verifyName(`http://host`, null, null, null)
    verifyName(`http://host/`, null, null, null)
    verifyName(`/`, null, null, null)

    // absolute names
    verifyName(`http://host/a`, "a", "a", null)
    verifyName(`http://host/a/`, "a", "a", null)
    verifyName(`http://host/a.`, "a.", "a", "")
    verifyName(`http://host/a.t`, "a.t", "a", "t")

    // relative paths
    verifyName(`alpha`, "alpha", "alpha", null)
    verifyName(`alpha.`, "alpha.", "alpha", "")
    verifyName(`alpha.txt`, "alpha.txt", "alpha", "txt")
    verifyName(`alpha.txt/`, "alpha.txt", "alpha", "txt")

    // dot
    verifyName(`.`, ".", ".", null)
    verifyName(`..`, "..", "..", null)
    verifyName(`../..`, "..", "..", null)
  }

  Void verifyName(Uri uri, Str name, Str basename, Str ext)
  {
    verifyEq(uri.name, name)
    verifyEq(uri.basename, basename)
    verifyEq(uri.ext, ext)
  }

//////////////////////////////////////////////////////////////////////////
// Query
//////////////////////////////////////////////////////////////////////////

  Void testQuery()
  {
    // none
    verifyQuery(`mailto:brian@there.com`, null, null)
    verifyQuery(`mailto:brian?there?com`, null, null)  // opaque
    verifyQuery(`http://foo/index`, null, null)
    verifyQuery(`http://foo/index#frag`, null, null)
    verifyQuery(`/index`, null, null)

    // various prefixes
    verifyQuery(`http://foo/index?a=b&c=d#frag`, "a=b&c=d", ["a":"b", "c":"d"])
    verifyQuery(`http://foo/index?a=b&c=d`, "a=b&c=d", ["a":"b", "c":"d"])
    verifyQuery(`/index?a=b;c=d`, "a=b;c=d", ["a":"b", "c":"d"])
    verifyQuery(`file.txt?a=b;c=d`, "a=b;c=d", ["a":"b", "c":"d"])
    verifyQuery(`?a=b&c=d`, "a=b&c=d", ["a":"b", "c":"d"])

    // various combinations
    verifyQuery(`?a`, "a", Str:Str["a":null])
    verifyQuery(`?a_b!`, "a_b!", Str:Str["a_b!":null])
    verifyQuery(`?a=b`, "a=b", ["a":"b"])
    verifyQuery(`?a=beta`, "a=beta", ["a":"beta"])
    verifyQuery(`?alpha=b`, "alpha=b", ["alpha":"b"])
    verifyQuery(`?alpha=b;`, "alpha=b;", ["alpha":"b"])
    verifyQuery(`?alpha=b&`, "alpha=b&", ["alpha":"b"])
    verifyQuery(`?alpha=b;;&;&`, "alpha=b;;&;&", ["alpha":"b"])
    verifyQuery(`?alpha=b&c`, "alpha=b&c", Str:Str["alpha":"b", "c":null])
    verifyQuery(`?a=b&;&charlie;`, "a=b&;&charlie;", Str:Str["a":"b", "charlie":null])
  }

  Void verifyQuery(Uri uri, Str queryStr, Str:Str query)
  {
    verifyEq(uri.queryStr, queryStr)
    verifyEq(uri.query, query)
    if (query != null) verify(uri.query.isRO())
  }

//////////////////////////////////////////////////////////////////////////
// Frag
//////////////////////////////////////////////////////////////////////////

  Void testFrag()
  {
    // none
    verifyFrag(`mailto:brian@there.com`, null)
    verifyFrag(`mailto:brian?there#com`, "com")
    verifyFrag(`http://foo/index`,  null)

    // combos
    verifyFrag(`http://foo/index?a=b&c=d#frag`, "frag")
    verifyFrag(`http://foo/index#frag`, "frag")
    verifyFrag(`//foo/index#f`, "f")
    verifyFrag(`file.txt#f_g`, "f_g")
    verifyFrag(`#h1`, "h1")
  }

  Void verifyFrag(Uri uri, Str frag)
  {
    verifyEq(uri.frag, frag)
  }

//////////////////////////////////////////////////////////////////////////
// Parent
//////////////////////////////////////////////////////////////////////////

  Void testParent()
  {
    verifyParent(`http://foo:81/a/b/c/?query#frag`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c?query#frag`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c/?query`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c?query`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c/#frag`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c#frag`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c/`, `http://foo:81/a/b/`)
    verifyParent(`http://foo:81/a/b/c`, `http://foo:81/a/b/`)

    verifyParent(`http://foo/a/b/c/`, `http://foo/a/b/`)
    verifyParent(`http://foo/a/b/c`, `http://foo/a/b/`)

    verifyParent(`/a/b/c/`, `/a/b/`)
    verifyParent(`/a/b/c`, `/a/b/`)

    verifyParent(`a/b/c/`, `a/b/`)
    verifyParent(`a/b/c`, `a/b/`)

    verifyParent(`../a/b/c/`, `../a/b/`)
    verifyParent(`../a/b/c#frag`, `../a/b/`)

    verifyParent(`/a/`, `/`)
    verifyParent(`/a`,  `/`)

    verifyParent(`/`, null)
    verifyParent(`a.txt`, null)
    verifyParent(`a`, null)
    verifyParent(``, null)
  }

  Void verifyParent(Uri uri, Uri parent)
  {
    verifyEq(uri.parent, parent)
  }

//////////////////////////////////////////////////////////////////////////
// Norm
//////////////////////////////////////////////////////////////////////////

  Void testNorm()
  {
    // abs
    verifyNorm(`http://foo:81/path?q#f`, `http://foo:81/path?q#f`)
    verifyNorm(`http://foo/.`, `http://foo/`)
    verifyNorm(`http://foo/..`, `http://foo/..`)
    verifyNorm(`http://foo/a/..`, `http://foo/`)
    verifyNorm(`http://foo/a/../..`, `http://foo/..`)
    verifyNorm(`http://foo/a/b/c/../..`, `http://foo/a/`)

    // relative
    verifyNorm(`..`, `..`)
    verifyNorm(`a/..`, ``)
    verifyNorm(`a/../..`, `..`)
    verifyNorm(`a/b/c/../..`, `a/`)
  }

  Void verifyNorm(Uri uri, Uri norm)
  {
    verifyEq(uri.normalize, norm)
    if (uri == norm) verify(uri.normalize == uri)
  }

//////////////////////////////////////////////////////////////////////////
// Plus
//////////////////////////////////////////////////////////////////////////

  Void testPlus()
  {
    // abs + abs
    verifyPlus(`http://foo:81/path?q#f`, `http://bar:82/p?q2#f2`, `http://bar:82/p?q2#f2`)
    verifyPlus(`http://foo:81/path?q#f`, `mailto:joe@there.com`, `mailto:joe@there.com`)
    verifyPlus(`http://foo/path`, `http://bar/`, `http://bar/`)

    // rel + abs
    verifyPlus(`/a/b/c`, `http://bar:82/p?q2#f2`, `http://bar:82/p?q2#f2`)
    verifyPlus(`/path?q#f`, `mailto:joe@there.com`, `mailto:joe@there.com`)

    // abs + rel /path
    verifyPlus(`http://foo:81/path?q#f`, `/newpath`, `http://foo:81/newpath`)
    verifyPlus(`http://foo:81/path`, `/a/b/c`, `http://foo:81/a/b/c`)
    verifyPlus(`http://foo/`, `/a?query`, `http://foo/a?query`)

    // abs + rel path
    verifyPlus(`http://foo/path?q#f`, `newpath`, `http://foo/newpath`)
    verifyPlus(`http://foo/path/?q#f`, `newpath`, `http://foo/path/newpath`)
    verifyPlus(`http://foo/`, `a/b/c`, `http://foo/a/b/c`)
    verifyPlus(`http://foo/root`, `a/b/c`, `http://foo/a/b/c`)
    verifyPlus(`http://foo/root/`, `a/b/c`, `http://foo/root/a/b/c`)
    verifyPlus(`http://foo/root/`, `../a`, `http://foo/a`)

    // abs + query
    verifyPlus(`http://foo/path?q#f`, `?newquery`, `http://foo/?newquery`)
    verifyPlus(`http://foo/path/?q#f`, `?newquery`, `http://foo/path/?newquery`)
    verifyPlus(`http://foo/path/`, `?newquery`, `http://foo/path/?newquery`)

    // abs + frag
    verifyPlus(`http://foo/path?q#f`, `#newfrag`, `http://foo/path?q#newfrag`)
    verifyPlus(`http://foo/`, `#newfrag`, `http://foo/#newfrag`)

    // rel + rel
    verifyPlus(`a/b/c`, `d`, `a/b/d`)
    verifyPlus(`a/b/c/`, `d`, `a/b/c/d`)
    verifyPlus(`a/b/c`, `./d`, `a/b/d`)
    verifyPlus(`a/b/c/`, `./d`, `a/b/c/d`)
    verifyPlus(`a/b/c`, `../d`, `a/d`)
    verifyPlus(`a/b/c/`, `../d`, `a/b/d`)
    verifyPlus(`a/b/c`, `../../d`, `d`)
    verifyPlus(`a/b/c/`, `../../d`, `a/d`)
    verifyPlus(`a/b/c`, `../../../d`, `../d`)
    verifyPlus(`a/b/c/`, `../../../d`, `d`)
  }

  Void verifyPlus(Uri a, Uri b, Uri r)
  {
    verifyEq(a + b, r)
  }

//////////////////////////////////////////////////////////////////////////
// Minus
//////////////////////////////////////////////////////////////////////////

  Void testMinus()
  {
    verifyMinus(`http://foo/path?q#f`, `http://bar/`, `http://foo/path?q#f`)
    verifyMinus(`http://foo/path?q#f`, `/bar`, `http://foo/path?q#f`)
    verifyMinus(`http://foo/path?q#f`, `bar`,  `http://foo/path?q#f`)

    verifyMinus(`http://bar/`, `http://foo/path?q#f`,  `http://bar/`)
    verifyMinus(`/bar`, `http://foo/path?q#f`, `/bar`)
    verifyMinus(`bar`, `http://foo/path?q#f`, `bar`)

    verifyMinus(`http://foo/a/b/c`,  `http://foo/a/b/c`,  ``)
    verifyMinus(`http://foo/a/b/c`,  `http://foo/a/b`,    `c`)
    verifyMinus(`http://foo/a/b/c`,  `http://foo/a/b/`,   `c`)
    verifyMinus(`http://foo/a/b/c`,  `http://foo/a`,      `b/c`)
    verifyMinus(`http://foo/a/b/c`,  `http://foo/`,       `a/b/c`)

    verifyMinus(`/a/b/c`, `/a/b/c`, ``)
    verifyMinus(`/a/b/c`, `/a/b`,   `c`)
    verifyMinus(`/a/b/c`, `/a`,     `b/c`)
    verifyMinus(`/a/b/c`, `/`,      `a/b/c`)
  }

  Void verifyMinus(Uri a, Uri b, Uri r)
  {
    verifyEq(a - b, r)
  }

//////////////////////////////////////////////////////////////////////////
// Tail
//////////////////////////////////////////////////////////////////////////

  Void testTail()
  {
    // various combinations
    verifyEq(`http://host:55/alpha/beta/charlie?query#frag`.tail, `beta/charlie`)
    verifyEq(`pod://sysTest/a/b.txt`.tail, `b.txt`)
    verifyErr(ArgErr.type) |,| { `mailto:bs@fan.org`.tail }

    // default 1
    verifyEq(`a/b/c/d`.tail, `b/c/d`)
    verifyEq(`a/b/c`.tail, `b/c`)
    verifyEq(`a/b`.tail, `b`)
    verifyEq(`a`.tail, ``)
    verifyErr(ArgErr.type) |,| { ``.tail }

    // tail(2)
    verifyEq(`a/b/c/d`.tail(2), `c/d`)
    verifyEq(`a/b/c`.tail(2), `c`)
    verifyEq(`a/b`.tail(2), ``)
    verifyErr(ArgErr.type) |,| { `a`.tail(2) }
    verifyErr(ArgErr.type) |,| { ``.tail(2) }

    // tail(3)
    verifyEq(`a/b/c/d`.tail(3), `d`)
    verifyEq(`a/b/c`.tail(3), ``)
    verifyErr(ArgErr.type) |,| { `a/b`.tail(3) }
    verifyErr(ArgErr.type) |,| { `a`.tail(3) }
    verifyErr(ArgErr.type) |,| { ``.tail(3) }
  }

//////////////////////////////////////////////////////////////////////////
// Obj.trapUri
//////////////////////////////////////////////////////////////////////////

  Void testTrapUri()
  {
    verifyEq(``.get(15), 15)
    verifyEq(`toStr`.get(15), "15")
    verifyEq(`toStr/size`.get(15), 2)
    verifyEq(`toStr/size/negate`.get(15), -2)
  }

//////////////////////////////////////////////////////////////////////////
// Dump
//////////////////////////////////////////////////////////////////////////

  Void dump(Uri uri)
  {
    echo("uri:        $uri")
    echo("  scheme:   $uri.scheme")
    echo("  auth:     $uri.auth")
    echo("  host:     $uri.host")
    echo("  port:     $uri.port")
    echo("  userInfo: $uri.userInfo")
    echo("  path:     $uri.path")
    echo("  pathStr:  $uri.pathStr")
    echo("  name:     $uri.name")
    echo("  basename: $uri.basename")
    echo("  ext:      $uri.ext")
    echo("  query:    $uri.query")
    echo("  queryStr: $uri.queryStr")
    echo("  frag:     $uri.frag")
  }

}