//
// Copyright (c) 2006, Brian Frank and Andy Frank
// Licensed under the Academic Free License version 3.0
//
// History:
//   27 Jul 06  Brian Frank  Creation
//

**
** VersionTest
**
class VersionTest : Test
{

//////////////////////////////////////////////////////////////////////////
// Parse
//////////////////////////////////////////////////////////////////////////

  Void testParse()
  {
    v := Version.parse("2")
    verifyEq(v.segments, [2])
    verifyEq(v.toStr, "2")

    v = Version.parse("1.2")
    verifyEq(v.segments, [1, 2])
    verifyEq(v.toStr, "1.2")

    v = Version.parse("12.0.33")
    verifyEq(v.segments, [12, 0, 33])
    verifyEq(v.toStr, "12.0.33")

    v = Version.parse("08.000.300")
    verifyEq(v.segments, [8, 0, 300])
    verifyEq(v.toStr, "8.0.300")

    verifyErr(ArgErr.type) |,| { Version.parse("") }
    verifyErr(ArgErr.type) |,| { Version.parse(".") }
    verifyErr(ArgErr.type) |,| { Version.parse("x") }
    verifyErr(ArgErr.type) |,| { Version.parse("3a") }
    verifyErr(ArgErr.type) |,| { Version.parse("3.x") }
    verifyErr(ArgErr.type) |,| { Version.parse("3..0") }
    verifyErr(ArgErr.type) |,| { Version.parse(".3.0") }
    verifyErr(ArgErr.type) |,| { Version.parse("3.0.") }
    verifyErr(ArgErr.type) |,| { Version.parse("1.0\n") }
  }

//////////////////////////////////////////////////////////////////////////
// Make
//////////////////////////////////////////////////////////////////////////

  Void testMake()
  {
    v := Version.make([1])
    verifyEq(v.segments, [1])
    verifyEq(v.segments.isRO, true)
    verifyEq(v.toStr, "1")

    v = Version.make([2, 5])
    verifyEq(v.segments, [2, 5])
    verifyEq(v.toStr, "2.5")

    v = Version.make([3, 0, 20060726])
    verifyEq(v.segments, [3, 0, 20060726])
    verifyEq(v.toStr, "3.0.20060726")

    // verify Version is independent of ints passed in
    ints := [1, 2]
    v = Version.make(ints)
    verifyEq(v.segments, [1, 2])
    verifyEq(v.toStr, "1.2")
    ints[1] = 5
    verifyEq(v.segments, [1, 2])
    verifyEq(v.toStr, "1.2")

    verifyErr(ArgErr.type) |,| { Version.make(Int[,]) }
    verifyErr(ArgErr.type) |,| { Version.make(Int[-2]) }
    verifyErr(ArgErr.type) |,| { Version.make(Int[1, -2]) }
    verifyErr(ReadonlyErr.type) |,| { Version.make([4, 8]).segments[0] = 9 }
  }

//////////////////////////////////////////////////////////////////////////
// Equals
//////////////////////////////////////////////////////////////////////////

  Void testEquals()
  {
    verifyEq(Version.parse("2"), Version.parse("2"))
    verifyEq(Version.parse("2"), Version.parse("02"))
    verifyEq(Version.parse("1.5.90"), Version.parse("1.5.90"))
    verifyEq(Version.parse("1.5.90"), Version.parse("1.05.90"))

    verifyNotEq(Version.parse("2"), Version.parse("3"))
    verifyNotEq(Version.parse("2"), Version.parse("2.0"))
    verifyNotEq(Version.parse("1.3"), Version.parse("1.2"))

    verifyEq(Version.parse("1.6").hash, "1.6".hash)
    verifyEq(Version.parse("1.006").hash, "1.6".hash)
  }

//////////////////////////////////////////////////////////////////////////
// Comparision
//////////////////////////////////////////////////////////////////////////

  Void testComparison()
  {
    verifyComparison("1",     "1",      0)
    verifyComparison("1.8",   "1.08",   0)
    verifyComparison("2",     "1",     +1)
    verifyComparison("10",    "7",     +1)
    verifyComparison("1.5",   "1",     +1)
    verifyComparison("1.5",   "1.3",   +1)
    verifyComparison("2.0",   "1.9",   +1)
    verifyComparison("1.2",   "1.2.3", -1)
    verifyComparison("1.2.3", "1.02",  +1)
    verifyComparison("1.11",  "1.9.3", +1)
  }

  Void verifyComparison(Str a, Str b, Int cmp)
  {
    verifyEq(Version.parse(a) <=> Version.parse(b), cmp)
    //verifyEq(Version.parse(a) <=> b, cmp) - removed support
  }

//////////////////////////////////////////////////////////////////////////
// Segments
//////////////////////////////////////////////////////////////////////////

  Void testSegments()
  {
    v := Version.parse("1")
    verifyEq(v.segments.isRO, true)
    verifyEq(v.major, 1)
    verifyEq(v.minor, null)
    verifyEq(v.build, null)
    verifyEq(v.patch, null)

    v = Version.parse("1.2")
    verifyEq(v.major, 1)
    verifyEq(v.minor, 2)
    verifyEq(v.build, null)
    verifyEq(v.patch, null)

    v = Version.parse("1.2.3")
    verifyEq(v.major, 1)
    verifyEq(v.minor, 2)
    verifyEq(v.build, 3)
    verifyEq(v.patch, null)

    v = Version.parse("1.2.3.4")
    verifyEq(v.major, 1)
    verifyEq(v.minor, 2)
    verifyEq(v.build, 3)
    verifyEq(v.patch, 4)
  }

}